

#include<cstdlib>
#include<cstdio>
#include<iostream>
#include<cassert>
#include <new>
#include<bitset>
#include<map>

#include "Explorer.h"

using namespace std;

#ifndef _CYCLE_HASH
#define _CYCLE_HASH

#define VERIFY(X)

_MemStruct::_MemStruct() {
    CurrLevelMem = new _DPState [INITALLOC];
    if(CurrLevelMem == 0) {
        cerr<<"MemStructInit: Fatal Error! Can not allocate memory\n";
        exit( 1);
    }
    nMemAvailable = INITALLOC;
    nMemUsage = 0;
    LevelMem[0] = CurrLevelMem;
    nMemLevel = 1;
}

_MemStruct::~_MemStruct() {
    int i;
    for(i = nMemLevel-1; i >= 0; i--)
        delete [] LevelMem[i];
}

_DPState* _MemStruct::DPStateAlloc() {
    if(nMemUsage < nMemAvailable)
        return &(CurrLevelMem[nMemUsage++]);
    else {
        //cout<<nMemAvailable<<" "<<nMemLevel<<" "<<nMemUsage<<endl;
        CurrLevelMem = 0;

        if(nMemLevel < MEMLEVELS) {
            CurrLevelMem = new _DPState [2*nMemAvailable];
            LevelMem[nMemLevel] = CurrLevelMem;
            nMemAvailable *= 2;
            nMemLevel++;
            nMemUsage = 0;
        }
        if(CurrLevelMem)
            return &(CurrLevelMem[nMemUsage++]);
        else {
            cerr<<"DPStateAlloc: Fatal Error! Can not allocate additional memory\n";
            return 0;
        }
    }
}

_HashTable::_HashTable() {
    int i;
    for(i = HASHTABLESIZE-1; i >= 0; i--)
        HashTable[i] = NULL;
}

void _HashTable::HashInsert(_DPState *pState) {
    unsigned long hashkey = 0;

    hashkey += pState->Set[0] + pState->Set[1] + pState->Set[2] + pState->Set[3];

    hashkey *= pState->i;
    hashkey %= HASHTABLESIZE;
    pState->pNextHash = HashTable[hashkey];
    HashTable[hashkey] = pState;
}

_DPState* _HashTable::HashFind(int i, unsigned long Set[], short stage) {
    unsigned long hashkey = 0;
    _DPState *Iter;

    hashkey += Set[0] + Set[1] + Set[2] + Set[3];

    hashkey *= i;


    if(hashkey < 0)
        hashkey = -hashkey;

    hashkey %= HASHTABLESIZE;



    Iter = HashTable[hashkey];

    for( ;
            Iter && ((Iter->i != i) || (Iter->Set[0] != Set[0]) || (Iter->Set[1] != Set[1])
                     || (Iter->Set[2] != Set[2]) || (Iter->Set[3] != Set[3])) && (Iter->Stage >= stage);
            Iter = Iter->pNextHash)
        ;

    if(Iter && Iter->Stage != stage)
        Iter = 0;
    return Iter;
}


//------------------------------------------------------------


ColVLSNImprovExplorer::ColVLSNImprovExplorer(ColoringManager* sm,
        Random *rnd, int state_limit,
        Graph* pin = NULL):
ColNeighborhoodExplorer(sm,rnd,pin) {
    STATE_LIMIT = state_limit;
};


ColVLSNImprovExplorer::ColVLSNImprovExplorer(const ColNeighborhoodExplorer* ne)
        :ColNeighborhoodExplorer(ne) {}
;

void ColVLSNImprovExplorer::SetParameters(int state_limit) {
    STATE_LIMIT = state_limit;
}

/*
int StateCmp(const void *p1, const void *p2)
{
  return (int)((*((struct _DPState **)p1))->fCost - (*((struct _DPState **)p2))->fCost);
}
*/

int StateCmp(const void *p1, const void *p2) {
    if ((*((_DPState **)p1))->fCost < (*((_DPState **)p2))->fCost)
        return -1;
    else if ((*((_DPState **)p1))->fCost == (*((_DPState **)p2))->fCost) {
        if ((*((_DPState **)p1))->Stage > (*((_DPState **)p2))->Stage)
            return -1;
        else if ((*((_DPState **)p1))->Stage == (*((_DPState **)p2))->Stage)
            return 0;
        else
            return 1;
    } else
        return 1;
}





void ColVLSNImprovExplorer::BestMove( const Coloring &current_sol, ColMove& mv) {
    _DPState *pDPState, *pDPState1;
    _DPState *poneex=NULL;
    _DPState **aStateTable1, **aStateTable2;
    int nStates1, nStates2, nAlloc;
    long num_states = 0;
    unsigned long Set[4];
    _DPState *pMin = NULL;
    fvalue fCostMin = MAX_FVALUE;
    _MemStruct Mem;
    fvalue temp = MAX_FVALUE;

    int Nodes = p_in->Vertices()+current_sol.CurrColors();
    int Real_Nodes = p_in->Vertices();

    _HashTable *StateHash = new _HashTable [Nodes];
    //  for(i = Nodes-1; i >= 0; i--)
    //      StateHash[i] = new _HashTable();

    aStateTable1 = new _DPState* [Nodes*Nodes];
    nStates1 = 0;


    for (VertexSet::iterator pt = current_sol.conflicts_set.begin();
            pt != current_sol.conflicts_set.end(); pt++) {
        int j=*pt;
        for(int i = 0; i < Real_Nodes; i++) {
            //for(int j = 0; j < Real_Nodes; j++)
            //{
            if((current_sol.Class(i) != current_sol.Class(j)) && (current_sol.ImpGraph(i,j) < 0))//Negative Sum Lemma
            {
                pDPState = Mem.DPStateAlloc();
                if(pDPState == NULL) {
                    delete [] aStateTable1;
                    delete [] StateHash;
                    cerr<<"Returning without cycle because of shortage of memory\n";
                    exit(1);
                }
                pDPState->Set[0] = pDPState->Set[1] = pDPState->Set[2] = pDPState->Set[3] = 0;

                if(current_sol.Class(i) < 32)
                    pDPState->Set[0] |= (1 << current_sol.Class(i));
                else if(current_sol.Class(i) < 64)
                    pDPState->Set[1] |= (1 << (current_sol.Class(i)-32));
                else if(current_sol.Class(i) < 96)
                    pDPState->Set[2] |= (1 << (current_sol.Class(i)-64));
                else
                    pDPState->Set[3] |= (1 << (current_sol.Class(i)-96));


                if(current_sol.Class(j) < 32)
                    pDPState->Set[0] |= (1 << current_sol.Class(j));
                else if(current_sol.Class(j) < 64)
                    pDPState->Set[1] |= (1 << (current_sol.Class(j)-32));
                else if(current_sol.Class(j) < 96)
                    pDPState->Set[2] |= (1 << (current_sol.Class(j)-64));
                else
                    pDPState->Set[3] |= (1 << (current_sol.Class(j)-96));


                pDPState->i = i;
                pDPState->j = j;
                pDPState->pred = 0;
                pDPState->fCost = current_sol.ImpGraph(i,j);
                pDPState->Stage = 0;

                /*------------------------------------------------------------*/
                StateHash[j].HashInsert(pDPState);
                aStateTable1[nStates1++] = pDPState;
                /*------------------------------------------------------------*/
                temp = current_sol.ImpGraph(i,j) + current_sol.ImpGraph(j,i);
                // temp here is the cost of a cycle of length 2, that is a one-exchange or a swap

                //cout<<temp<<" "<<i<<" "<<j<<endl;

                if(temp < fCostMin) {
                    VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<i<<"->"<<j<<endl;)
                    pMin = pDPState;
                    fCostMin = temp;
                }
            }
        }
        for (Color c=0 ; c < current_sol.CurrColors() ; c++) {
            if (current_sol.col[j]!=c) {
                temp = current_sol.Delta(j,c);
                if(temp < fCostMin) {
                    pDPState = Mem.DPStateAlloc();
                    if(pDPState == NULL) {
                        delete [] aStateTable1;
                        delete [] StateHash;
                        cerr<<"Returning without cycle because of shortage of memory\n";
                        exit(1);
                    }
                    pDPState->i = Real_Nodes+c;
                    pDPState->j = j;
                    pDPState->pred = 0;
                    pDPState->fCost = current_sol.ImpGraph(Real_Nodes+c,j);
                    pDPState->Stage = 0;
                    VERIFY(cout<<temp<<" ("<<pDPState->Stage<<"+) "
                           <<Real_Nodes+c<<"->"<<j<<endl;)
                    pMin = poneex = pDPState;
                    fCostMin = temp;
                }
            }
        }//close for(Color c....)
    }//close for(Vertex i....)



    if(nStates1 > 0)
        qsort( aStateTable1, nStates1, sizeof(struct _DPState *), StateCmp);

    if(nStates1 > STATE_LIMIT)
        nStates1 = STATE_LIMIT;

    /* -- Now that we have looked at all states of length 1, we look at bigger paths */
    while(nStates1 > 0) {
        num_states += nStates1;
        nAlloc = nStates1*STATE_LIMIT;
        aStateTable2 = new _DPState* [nAlloc];
        nStates2 = 0;
        for(int i = 0; i < nStates1; i++) {
            pDPState1 = aStateTable1[i];
            int j = pDPState1->j;
            Set[0] = pDPState1->Set[0];
            Set[1] = pDPState1->Set[1];
            Set[2] = pDPState1->Set[2];
            Set[3] = pDPState1->Set[3];

            for(int k = Nodes-1; k >= 0; k--) {
                if(pDPState1->fCost + current_sol.ImpGraph(j,k) < 0)//Negative Sum Lemma
                {
                    if(current_sol.Class(k) < 32) {
                        if(Set[0] & (1 << current_sol.Class(k)))
                            continue;
                        Set[0] |= (1 << current_sol.Class(k));
                    } else if(current_sol.Class(k) < 64) {
                        if(Set[1] & (1 << (current_sol.Class(k)-32)))
                            continue;
                        Set[1] |= (1 << (current_sol.Class(k)-32));
                    } else if(current_sol.Class(k) < 96) {
                        if(Set[2] & (1 << (current_sol.Class(k)-64)))
                            continue;
                        Set[2] |= (1 << (current_sol.Class(k)-64));
                    } else {
                        if(Set[3] & (1 << (current_sol.Class(k)-96)))
                            continue;
                        Set[3] |= (1 << (current_sol.Class(k)-96));
                    }

                    //check if there is another growing path that arrives in the class of k
                    //and substitute if this is better
                    //otherwise create
                    pDPState = StateHash[k].HashFind(pDPState1->i, Set,
                                                     (short)(pDPState1->Stage+1));
                    if(pDPState) {
                        if(pDPState->fCost > pDPState1->fCost + current_sol.ImpGraph(j,k)) {
                            pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                            pDPState->pred = pDPState1;
                        }
                    } else {
                        pDPState = Mem.DPStateAlloc();
                        if(pDPState == 0) {
                            delete [] aStateTable1;
                            delete [] aStateTable2;
                            delete [] StateHash;
                            cerr<<"Returning without cycle because of shortage of memory\n";
                            exit(1);
                        }
                        pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                        pDPState->i = pDPState1->i;
                        pDPState->j = k;

                        pDPState->Set[0] = Set[0];
                        pDPState->Set[1] = Set[1];
                        pDPState->Set[2] = Set[2];
                        pDPState->Set[3] = Set[3];

                        pDPState->pred = pDPState1;
                        pDPState->Stage = pDPState1->Stage + 1;

                        StateHash[k].HashInsert(pDPState);
                        if(nStates2 < nAlloc)
                            aStateTable2[nStates2++] = pDPState;
                        else {
                            _DPState **tmp = new _DPState* [2*nAlloc];
                            for(nStates2 = nAlloc-1; nStates2 >= 0; nStates2--)
                                tmp[nStates2] = aStateTable2[nStates2];

                            delete [] aStateTable2;
                            aStateTable2 = tmp;
                            nStates2 = nAlloc;
                            nAlloc *= 2;
                            aStateTable2[nStates2++] = pDPState;
                        }
                    }

                    /* -- Check for a cycle */
                    temp = pDPState->fCost + current_sol.ImpGraph(k,pDPState->i);
                    //cout<<temp<<" "<<k<<" "<<pDPState->i<<endl;
                    if(temp < fCostMin) {
                        VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<pDPState->i<<"->"<<k<<endl;)
                        pMin = pDPState;
                        fCostMin = temp;
                    }

                    if(current_sol.Class(k) < 32)
                        Set[0] &= ~(1 << current_sol.Class(k));
                    else if(current_sol.Class(k) < 64)
                        Set[1] &= ~(1 << (current_sol.Class(k)-32));
                    else if(current_sol.Class(k) < 96)
                        Set[2] &= ~(1 << (current_sol.Class(k)-64));
                    else
                        Set[3] &= ~(1 << (current_sol.Class(k)-96));
                } /*end if (..)*/
            }/*end k*/
        } /*end for i<nstates1*/
        nStates1 = nStates2;
        delete [] aStateTable1;
        aStateTable1 = aStateTable2;

        if(nStates1 > 0)
            qsort( aStateTable1, nStates1, sizeof(struct _DPState*), StateCmp);

        if(nStates1 > STATE_LIMIT)
            nStates1 = STATE_LIMIT;
    }

    delete [] aStateTable1;
    delete [] StateHash;

    if (fCostMin < 0) {
        mv.type = 2;
        int i = pMin->Stage+2;

        mv.seq.clear();
        mv.seq.resize(i+1,0);

        mv.seq[i] = pMin->i;
        for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
            mv.seq[--i] = pDPState->j;
        mv.seq[--i] = pDPState->j;
        mv.seq[0] = pDPState->i;

        mv.delta =  fCostMin;

    } else
        mv.type = 0;
    //if (poneex!=NULL)
    // delete poneex;
    return;
}









void ColVLSNExplorer::BestMove(const Coloring &current_sol, ColMove& mv)
{
    _DPState *pDPState, *pDPState1;
    _DPState *p_perturb=NULL;
    _DPState **aStateTable1, **aStateTable2;
    int nStates1, nStates2, nAlloc;
    long num_states = 0;
    unsigned long Set[4];
    _DPState *pMin = NULL;
    fvalue fCostMin = MAX_FVALUE;
    fvalue fCost_perturb = MAX_FVALUE;
    _MemStruct Mem;
    fvalue temp = MAX_FVALUE;


    int Nodes = p_in->Vertices()+current_sol.CurrColors();
    long int *ran_order=rg->generate_array(Nodes);

    _HashTable *StateHash = new _HashTable [Nodes];
    //  for(i = Nodes-1; i >= 0; i--)
    //      StateHash[i] = new _HashTable();

    aStateTable1 = new _DPState* [Nodes*Nodes];
    nStates1 = 0;

    for (VertexSet::iterator pt = current_sol.conflicts_set.begin();
            pt != current_sol.conflicts_set.end(); pt++) {
        int j=*pt;
        for(int i = 0; i < Nodes; i++) {
            //for(int j = 0; j < Real_Nodes; j++)
            //{
            //	  i=ran_order[l];
            if((current_sol.Class(i) != current_sol.Class(j)) && (current_sol.ImpGraph(i,j) < 0))//Negative Sum Lemma
            {
                pDPState = Mem.DPStateAlloc();
                if(pDPState == NULL) {
                    delete [] aStateTable1;
                    delete [] StateHash;
                    cerr<<"Returning without cycle because of shortage of memory\n";
                    exit(1);
                }
                pDPState->Set[0] = pDPState->Set[1] = pDPState->Set[2] = pDPState->Set[3] = 0;

                if(current_sol.Class(i) < 32)
                    pDPState->Set[0] |= (1 << current_sol.Class(i));
                else if(current_sol.Class(i) < 64)
                    pDPState->Set[1] |= (1 << (current_sol.Class(i)-32));
                else if(current_sol.Class(i) < 96)
                    pDPState->Set[2] |= (1 << (current_sol.Class(i)-64));
                else
                    pDPState->Set[3] |= (1 << (current_sol.Class(i)-96));


                if(current_sol.Class(j) < 32)
                    pDPState->Set[0] |= (1 << current_sol.Class(j));
                else if(current_sol.Class(j) < 64)
                    pDPState->Set[1] |= (1 << (current_sol.Class(j)-32));
                else if(current_sol.Class(j) < 96)
                    pDPState->Set[2] |= (1 << (current_sol.Class(j)-64));
                else
                    pDPState->Set[3] |= (1 << (current_sol.Class(j)-96));


                pDPState->i = i;
                pDPState->j = j;
                pDPState->pred = 0;
                pDPState->fCost = current_sol.ImpGraph(i,j);
                pDPState->Stage = 0;

                /*------------------------------------------------------------*/
                StateHash[j].HashInsert(pDPState);
                aStateTable1[nStates1++] = pDPState;
                /*------------------------------------------------------------*/
                temp = current_sol.ImpGraph(i,j) + current_sol.ImpGraph(j,i);
                // temp here is the cost of a cycle of length 2, that is a one-exchange or a swap

                //cout<<temp<<" "<<i<<" "<<j<<endl;
                // 	    isNotTabu = !tl.isTabu(i,current_sol.Class(j),par_iter) && !tl.isTabu(j,current_sol.Class(i),par_iter);
                // 	    if (((temp < fCostMin) && isNotTabu) ||
                // 		((temp == fCostMin) && isNotTabu && rg->ran01()<0.5) ||
                // 		current_sol.fitness + temp < best_ever)
                if (temp < fCostMin) {
                    VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<i<<"->"<<j<<endl;)
                    pMin =p_perturb= pDPState;
                    fCostMin =fCost_perturb= temp;
                }
            }
        }
        //      for (Color c=0 ; c < current_sol.CurrColors() ; c++) {
        //	if (current_sol.col[j]!=c)
        //	  {
        //	    temp = current_sol.Delta(j,c);
        //	    // if (((temp < fCostMin) && !tl.isTabu(i,c,par_iter)) ||
        //// 		//((temp == fCostMin) && !tl.isTabu(i,c,par_iter) && rg->ran01()<0.5) ||
        //// 		(current_sol.fitness + temp < best_ever)
        //// 		)
        //	    if (temp < fCostMin)
        //	      {
        //		pDPState = Mem.DPStateAlloc();
        //		if(pDPState == NULL)
        //		  {
        //		    delete [] aStateTable1;
        //		    delete [] StateHash;
        //		    cerr<<"Returning without cycle because of shortage of memory\n";
        //		    exit(1);
        //		  }
        //		pDPState->i = Real_Nodes+c;
        //		pDPState->j = j;
        //		pDPState->pred = 0;
        //		pDPState->fCost = current_sol.ImpGraph(Real_Nodes+c,j);
        //		pDPState->Stage = 0;
        //		VERIFY(cout<<temp<<" ("<<pDPState->Stage<<"+) "
        //		       <<Real_Nodes+c<<"->"<<j<<endl;)
        //		pMin = poneex = pDPState;
        //		fCostMin = temp;
        //	      }
        //	  }
        //      }//close for(Color c....)
    }//close for(Vertex i....)




    if(nStates1 > 0) {
        //  if (fCostMin < 0)
        qsort( aStateTable1, nStates1, sizeof(struct _DPState *), StateCmp);
    } else {
        cerr<<"nStates1 == 0!"<<endl;
        exit(EXIT_FAILURE);
    }


    //cerr<<nStates1<<" "<<STATE_LIMIT<<endl;
    if(nStates1 > STATE_LIMIT)
        nStates1 = STATE_LIMIT;

    /* -- Now that we have looked at all states of length 1, we look at bigger paths */
    while(nStates1 > 0) {
        num_states += nStates1;
        nAlloc = nStates1*STATE_LIMIT;
        aStateTable2 = new _DPState* [nAlloc];
        nStates2 = 0;
        for(int i = 0; i < nStates1; i++) {
            pDPState1 = aStateTable1[i];
            int j = pDPState1->j;
            Set[0] = pDPState1->Set[0];
            Set[1] = pDPState1->Set[1];
            Set[2] = pDPState1->Set[2];
            Set[3] = pDPState1->Set[3];

            for(int l = Nodes-1; l >= 0; l--) {
                int k=(int) ran_order[l];
                if(pDPState1->fCost + current_sol.ImpGraph(j,k) < 0)//Negative Sum Lemma
                {
                    if(current_sol.Class(k) < 32) {
                        if(Set[0] & (1 << current_sol.Class(k)))
                            continue;
                        Set[0] |= (1 << current_sol.Class(k));
                    } else if(current_sol.Class(k) < 64) {
                        if(Set[1] & (1 << (current_sol.Class(k)-32)))
                            continue;
                        Set[1] |= (1 << (current_sol.Class(k)-32));
                    } else if(current_sol.Class(k) < 96) {
                        if(Set[2] & (1 << (current_sol.Class(k)-64)))
                            continue;
                        Set[2] |= (1 << (current_sol.Class(k)-64));
                    } else {
                        if(Set[3] & (1 << (current_sol.Class(k)-96)))
                            continue;
                        Set[3] |= (1 << (current_sol.Class(k)-96));
                    }

                    //check if there is another growing path that arrives in the class of k
                    //and substitute if this is better
                    //otherwise create
                    pDPState = StateHash[k].HashFind(pDPState1->i, Set,
                                                     (short)(pDPState1->Stage+1));
                    if(pDPState) {
                        if(pDPState->fCost > pDPState1->fCost + current_sol.ImpGraph(j,k)) {
                            pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                            pDPState->pred = pDPState1;
                        }
                    } else {
                        pDPState = Mem.DPStateAlloc();
                        if(pDPState == 0) {
                            delete [] aStateTable1;
                            delete [] aStateTable2;
                            delete [] StateHash;
                            cerr<<"Returning without cycle because of shortage of memory\n";
                            exit(1);
                        }
                        pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                        pDPState->i = pDPState1->i;
                        pDPState->j = k;

                        pDPState->Set[0] = Set[0];
                        pDPState->Set[1] = Set[1];
                        pDPState->Set[2] = Set[2];
                        pDPState->Set[3] = Set[3];

                        pDPState->pred = pDPState1;
                        pDPState->Stage = pDPState1->Stage + 1;

                        StateHash[k].HashInsert(pDPState);
                        if(nStates2 < nAlloc)
                            aStateTable2[nStates2++] = pDPState;
                        else {
                            _DPState **tmp = new _DPState* [2*nAlloc];
                            for(nStates2 = nAlloc-1; nStates2 >= 0; nStates2--)
                                tmp[nStates2] = aStateTable2[nStates2];

                            delete [] aStateTable2;
                            aStateTable2 = tmp;
                            nStates2 = nAlloc;
                            nAlloc *= 2;
                            aStateTable2[nStates2++] = pDPState;
                        }
                    }

                    /* -- Check for a cycle */
                    temp = pDPState->fCost + current_sol.ImpGraph(k,pDPState->i);
                    //cout<<temp<<" "<<k<<" "<<pDPState->i<<endl;
                    if(temp < fCostMin) {
                        VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<pDPState->i<<"->"<<k<<endl;)
                        pMin = pDPState;
                        fCostMin = temp;
                    } else if (temp==0) {
                        if ((pDPState->Stage > p_perturb->Stage) ||
                                ((pDPState->Stage == p_perturb->Stage) && rg->next()<=0.5)) {
                            VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<pDPState->i<<"->"<<k<<endl;)
                            p_perturb = pDPState;
                            fCost_perturb = temp;
                        }
                    }

                    if(current_sol.Class(k) < 32)
                        Set[0] &= ~(1 << current_sol.Class(k));
                    else if(current_sol.Class(k) < 64)
                        Set[1] &= ~(1 << (current_sol.Class(k)-32));
                    else if(current_sol.Class(k) < 96)
                        Set[2] &= ~(1 << (current_sol.Class(k)-64));
                    else
                        Set[3] &= ~(1 << (current_sol.Class(k)-96));
                } /*end if (..)*/
            }/*end k*/
        } /*end for i<nstates1*/
        nStates1 = nStates2;
        delete [] aStateTable1;
        aStateTable1 = aStateTable2;

        if(nStates1 > 0)
            qsort( aStateTable1, nStates1, sizeof(struct _DPState*), StateCmp);

        if(nStates1 > STATE_LIMIT)
            nStates1 = STATE_LIMIT;
    }

    delete [] aStateTable1;
    delete [] StateHash;


    if (fCostMin>=0) {
        pMin=p_perturb;
        fCostMin=fCost_perturb;
    }

    mv.type = 2;
    int i = pMin->Stage+2;

    mv.seq.clear();
    mv.seq.resize(i+1,0);

    mv.seq[i] = pMin->i;
    for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
        mv.seq[--i] = pDPState->j;
    mv.seq[--i] = pDPState->j;
    mv.seq[0] = pDPState->i;

    mv.delta =  fCostMin;

    //if (poneex!=NULL)
    // delete poneex;
    return;
}





void ColVLSNExplorer::BestMove(const Coloring &current_sol, ColMove& mv,
                               TabuList &tl, unsigned long par_iter,
                               fvalue best_ever) {
    _DPState *pDPState, *pDPState1;
    _DPState *p_perturb=NULL;
    _DPState **aStateTable1, **aStateTable2;
    int nStates1, nStates2, nAlloc;
    long num_states = 0;
    unsigned long Set[4];
    _DPState *pMin = NULL;
    fvalue fCostMin = MAX_FVALUE;
    fvalue fCost_perturb = MAX_FVALUE;
    _MemStruct Mem;
    fvalue temp = MAX_FVALUE;

    bool isTabu =false;


    int Nodes = p_in->Vertices()+current_sol.CurrColors();
    long int *ran_order=rg->generate_array(Nodes);

    _HashTable *StateHash = new _HashTable [Nodes];
    //  for(i = Nodes-1; i >= 0; i--)
    //      StateHash[i] = new _HashTable();

    aStateTable1 = new _DPState* [Nodes*Nodes];
    nStates1 = 0;

    for (VertexSet::iterator pt = current_sol.conflicts_set.begin();
            pt != current_sol.conflicts_set.end(); pt++) {
        int j=*pt;
        for(int i = 0; i < Nodes; i++) {
            //for(int j = 0; j < Real_Nodes; j++)
            //{
            //	  i=ran_order[l];

            if((current_sol.Class(i) != current_sol.Class(j)) && (current_sol.ImpGraph(i,j) < 0))//Negative Sum Lemma
            {
                pDPState = Mem.DPStateAlloc();
                if(pDPState == NULL) {
                    delete [] aStateTable1;
                    delete [] StateHash;
                    cerr<<"Returning without cycle because of shortage of memory\n";
                    exit(1);
                }
                pDPState->Set[0] = pDPState->Set[1] = pDPState->Set[2] = pDPState->Set[3] = 0;

                if(current_sol.Class(i) < 32)
                    pDPState->Set[0] |= (1 << current_sol.Class(i));
                else if(current_sol.Class(i) < 64)
                    pDPState->Set[1] |= (1 << (current_sol.Class(i)-32));
                else if(current_sol.Class(i) < 96)
                    pDPState->Set[2] |= (1 << (current_sol.Class(i)-64));
                else
                    pDPState->Set[3] |= (1 << (current_sol.Class(i)-96));


                if(current_sol.Class(j) < 32)
                    pDPState->Set[0] |= (1 << current_sol.Class(j));
                else if(current_sol.Class(j) < 64)
                    pDPState->Set[1] |= (1 << (current_sol.Class(j)-32));
                else if(current_sol.Class(j) < 96)
                    pDPState->Set[2] |= (1 << (current_sol.Class(j)-64));
                else
                    pDPState->Set[3] |= (1 << (current_sol.Class(j)-96));


                pDPState->i = i;
                pDPState->j = j;
                pDPState->pred = 0;
                pDPState->fCost = current_sol.ImpGraph(i,j);
                pDPState->Stage = 0;

                /*------------------------------------------------------------*/
                StateHash[j].HashInsert(pDPState);
                aStateTable1[nStates1++] = pDPState;
                /*------------------------------------------------------------*/
                temp = current_sol.ImpGraph(i,j) + current_sol.ImpGraph(j,i);
                // temp here is the cost of a cycle of length 2, that is a one-exchange or a swap

                //cout<<temp<<" "<<i<<" "<<j<<endl;
                isTabu = tl.isTabu(j,current_sol.Class(j),par_iter)
                         || tl.isTabu(i,current_sol.Class(i),par_iter);
                if (((temp < fCostMin) && !isTabu) ||
                        ((temp == fCostMin) && !isTabu && rg->next()<0.5) ||
                        current_sol.fitness + temp < best_ever) {
                    VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<i<<"->"<<j<<endl;)
                    pMin =p_perturb= pDPState;
                    fCostMin =fCost_perturb= temp;
                }
            }
        }
        //      for (Color c=0 ; c < current_sol.CurrColors() ; c++) {
        //	if (current_sol.col[j]!=c)
        //	  {
        //	    temp = current_sol.Delta(j,c);
        //	    // if (((temp < fCostMin) && !tl.isTabu(i,c,par_iter)) ||
        //// 		//((temp == fCostMin) && !tl.isTabu(i,c,par_iter) && rg->next()<0.5) ||
        //// 		(current_sol.fitness + temp < best_ever)
        //// 		)
        //	    if (temp < fCostMin)
        //	      {
        //		pDPState = Mem.DPStateAlloc();
        //		if(pDPState == NULL)
        //		  {
        //		    delete [] aStateTable1;
        //		    delete [] StateHash;
        //		    cerr<<"Returning without cycle because of shortage of memory\n";
        //		    exit(1);
        //		  }
        //		pDPState->i = Real_Nodes+c;
        //		pDPState->j = j;
        //		pDPState->pred = 0;
        //		pDPState->fCost = current_sol.ImpGraph(Real_Nodes+c,j);
        //		pDPState->Stage = 0;
        //		VERIFY(cout<<temp<<" ("<<pDPState->Stage<<"+) "
        //		       <<Real_Nodes+c<<"->"<<j<<endl;)
        //		pMin = poneex = pDPState;
        //		fCostMin = temp;
        //	      }
        //	  }
        //      }//close for(Color c....)
    }//close for(Vertex i....)




    if(nStates1 > 0) {
        //  if (fCostMin < 0)
        qsort( aStateTable1, nStates1, sizeof(struct _DPState *), StateCmp);
    } else {
        cerr<<"nStates1 == 0!"<<endl;
        exit(EXIT_FAILURE);
    }


    //cerr<<nStates1<<" "<<STATE_LIMIT<<endl;
    if(nStates1 > STATE_LIMIT)
        nStates1 = STATE_LIMIT;

    /* -- Now that we have looked at all states of length 1, we look at bigger paths */
    while(nStates1 > 0) {
        num_states += nStates1;
        nAlloc = nStates1*STATE_LIMIT;
        aStateTable2 = new _DPState* [nAlloc];
        nStates2 = 0;
        for(int i = 0; i < nStates1; i++) {
            pDPState1 = aStateTable1[i];
            int j = pDPState1->j;
            Set[0] = pDPState1->Set[0];
            Set[1] = pDPState1->Set[1];
            Set[2] = pDPState1->Set[2];
            Set[3] = pDPState1->Set[3];

            for(int l = Nodes-1; l >= 0; l--) {
                int k=(int) ran_order[l];
                if(pDPState1->fCost + current_sol.ImpGraph(j,k) < 0)//Negative Sum Lemma
                {
                    if(current_sol.Class(k) < 32) {
                        if(Set[0] & (1 << current_sol.Class(k)))
                            continue;
                        Set[0] |= (1 << current_sol.Class(k));
                    } else if(current_sol.Class(k) < 64) {
                        if(Set[1] & (1 << (current_sol.Class(k)-32)))
                            continue;
                        Set[1] |= (1 << (current_sol.Class(k)-32));
                    } else if(current_sol.Class(k) < 96) {
                        if(Set[2] & (1 << (current_sol.Class(k)-64)))
                            continue;
                        Set[2] |= (1 << (current_sol.Class(k)-64));
                    } else {
                        if(Set[3] & (1 << (current_sol.Class(k)-96)))
                            continue;
                        Set[3] |= (1 << (current_sol.Class(k)-96));
                    }

                    //check if there is another growing path that arrives in the class of k
                    //and substitute if this is better
                    //otherwise create
                    pDPState = StateHash[k].HashFind(pDPState1->i, Set,
                                                     (short)(pDPState1->Stage+1));
                    if(pDPState) {
                        if(pDPState->fCost > pDPState1->fCost + current_sol.ImpGraph(j,k)) {
                            pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                            pDPState->pred = pDPState1;
                        }
                    } else {
                        pDPState = Mem.DPStateAlloc();
                        if(pDPState == 0) {
                            delete [] aStateTable1;
                            delete [] aStateTable2;
                            delete [] StateHash;
                            cerr<<"Returning without cycle because of shortage of memory\n";
                            exit(1);
                        }
                        pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                        pDPState->i = pDPState1->i;
                        pDPState->j = k;

                        pDPState->Set[0] = Set[0];
                        pDPState->Set[1] = Set[1];
                        pDPState->Set[2] = Set[2];
                        pDPState->Set[3] = Set[3];

                        pDPState->pred = pDPState1;
                        pDPState->Stage = pDPState1->Stage + 1;

                        StateHash[k].HashInsert(pDPState);
                        if(nStates2 < nAlloc)
                            aStateTable2[nStates2++] = pDPState;
                        else {
                            _DPState **tmp = new _DPState* [2*nAlloc];
                            for(nStates2 = nAlloc-1; nStates2 >= 0; nStates2--)
                                tmp[nStates2] = aStateTable2[nStates2];

                            delete [] aStateTable2;
                            aStateTable2 = tmp;
                            nStates2 = nAlloc;
                            nAlloc *= 2;
                            aStateTable2[nStates2++] = pDPState;
                        }
                    }

                    /* -- Check for a cycle */
                    temp = pDPState->fCost + current_sol.ImpGraph(k,pDPState->i);
                    //cout<<temp<<" "<<k<<" "<<pDPState->i<<endl;
                    isTabu = tl.isTabu(k,current_sol.Class(k),par_iter)
                             || tl.isTabu(i,current_sol.Class(pDPState->i),par_iter);
                    if (((temp < fCostMin) && !isTabu) ||
                            ((temp == fCostMin) && !isTabu && rg->next()<0.5) ||
                            current_sol.fitness + temp < best_ever) {
                        VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<pDPState->i<<"->"<<k<<endl;)
                        pMin = pDPState;
                        fCostMin = temp;
                    }
                    //		  else if (temp==0)
                    //		    {
                    //		      if ((pDPState->Stage > p_perturb->Stage) ||
                    //			  ((pDPState->Stage == p_perturb->Stage) && rg->next()<=0.5))
                    //			{
                    //			  VERIFY(cout<<temp<<" ("<<pDPState->Stage<<") "<<pDPState->i<<"->"<<k<<endl;)
                    //			    p_perturb = pDPState;
                    //			  fCost_perturb = temp;
                    //			}
                    //		    }

                    if(current_sol.Class(k) < 32)
                        Set[0] &= ~(1 << current_sol.Class(k));
                    else if(current_sol.Class(k) < 64)
                        Set[1] &= ~(1 << (current_sol.Class(k)-32));
                    else if(current_sol.Class(k) < 96)
                        Set[2] &= ~(1 << (current_sol.Class(k)-64));
                    else
                        Set[3] &= ~(1 << (current_sol.Class(k)-96));
                } /*end if (..)*/
            }/*end k*/
        } /*end for i<nstates1*/
        nStates1 = nStates2;
        delete [] aStateTable1;
        aStateTable1 = aStateTable2;

        if(nStates1 > 0)
            qsort( aStateTable1, nStates1, sizeof(struct _DPState*), StateCmp);

        if(nStates1 > STATE_LIMIT)
            nStates1 = STATE_LIMIT;
    }

    delete [] aStateTable1;
    delete [] StateHash;

    //
    //  if (fCostMin>=0)
    //    {
    //      pMin=p_perturb;
    //      fCostMin=fCost_perturb;
    //    }
    //
    mv.type = 2;
    int i = pMin->Stage+2;

    mv.seq.clear();
    mv.seq.resize(i+1,0);

    mv.seq[i] = pMin->i;
    for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
        mv.seq[--i] = pDPState->j;
    mv.seq[--i] = pDPState->j;
    mv.seq[0] = pDPState->i;

    mv.delta =  fCostMin;

    //if (poneex!=NULL)
    // delete poneex;
    return;
}




void ColReducedMinimumHashCycleOrPathExplorer::BestMove( const Coloring &current_sol,
        ColMove& mv,
        TabuList &tl, unsigned long par_iter,
        fvalue best_ever) {
    _DPState *pDPState, *pDPState1;
    _DPState **aStateTable1, **aStateTable2;
    int nStates1, nStates2, nAlloc;
    long num_states = 0;
    unsigned long Set[4];
    _DPState *pMin = NULL;
    fvalue fCostMin = MAX_FVALUE;
    _MemStruct Mem;
    bool path = false;
    fvalue temp = MAX_FVALUE;

    int Nodes = p_in->Vertices();

    _HashTable *StateHash = new _HashTable [Nodes];
    //  for(i = Nodes-1; i >= 0; i--)
    //      StateHash[i] = new _HashTable();

    aStateTable1 = new _DPState* [Nodes*Nodes];
    nStates1 = 0;


    for(int i = 0; i < Nodes; i++)
        for(int j = 0; j < Nodes; j++)
            if (current_sol.conflicts[i]>0 || current_sol.conflicts[j]>0) {
                if((current_sol.Class(i) != current_sol.Class(j))) {
                    pDPState = Mem.DPStateAlloc();
                    if(pDPState == NULL) {
                        delete [] aStateTable1;
                        delete [] StateHash;
                        cerr<<"Returning without cycle because of shortage of memory\n";
                        exit(1);
                    }
                    pDPState->Set[0] = pDPState->Set[1] = pDPState->Set[2] =pDPState->Set[3] = 0;
                    if(current_sol.Class(i) < 32)
                        pDPState->Set[0] |= (1 << current_sol.Class(i));
                    else if(current_sol.Class(i) < 64)
                        pDPState->Set[1] |= (1 << (current_sol.Class(i)-32));
                    else if(current_sol.Class(i) < 96)
                        pDPState->Set[2] |= (1 << (current_sol.Class(i)-64));
                    else
                        pDPState->Set[3] |= (1 << (current_sol.Class(i)-96));


                    if(current_sol.Class(j) < 32)
                        pDPState->Set[0] |= (1 << current_sol.Class(j));
                    else if(current_sol.Class(j) < 64)
                        pDPState->Set[1] |= (1 << (current_sol.Class(j)-32));
                    else if(current_sol.Class(j) < 96)
                        pDPState->Set[2] |= (1 << (current_sol.Class(j)-64));
                    else
                        pDPState->Set[3] |= (1 << (current_sol.Class(j)-96));

                    pDPState->i = i;
                    pDPState->j = j;
                    pDPState->pred = 0;
                    pDPState->fCost = current_sol.ImpGraph(i,j);
                    pDPState->Stage = 0;

                    /*------------------------------------------------------------*/
                    StateHash[j].HashInsert(pDPState);
                    aStateTable1[nStates1++] = pDPState;
                    /*------------------------------------------------------------*/
                    /* -- Check for a cycle */
                    temp = current_sol.ImpGraph(i,j) + current_sol.ImpGraph(j,i);

                    if(temp < fCostMin) {
                        if ((!tl.isTabu(i,current_sol.Class(j),par_iter)
                                && !tl.isTabu(j,current_sol.Class(i),par_iter))
                                ||  current_sol.fitness + temp < best_ever) {
                            pMin = pDPState;
                            fCostMin = temp;
                            path = false;
                        }
                    }
                    /* -- Check for a path */
                    temp = current_sol.ImpGraph(i,j)+current_sol.conflicts[j]
                           - current_sol.conflicts[i]+(fvalue) p_in->Edge(i,j);
                    if(temp < fCostMin) {
                        if (!tl.isTabu(i,current_sol.Class(j),par_iter)
                                || current_sol.fitness + temp < best_ever) {
                            pMin = pDPState;
                            fCostMin = temp;
                            path = true;
                        }
                    }
                }
            }

    if(nStates1 > 0)
        qsort( aStateTable1, nStates1, sizeof(struct _DPState *), StateCmp);

    if(nStates1 > STATE_LIMIT)
        nStates1 = STATE_LIMIT;

    /* -- Now that we have looked at all states of length 1, we look at bigger paths */
    while(nStates1 > 0) {
        num_states += nStates1;
        nAlloc = nStates1*STATE_LIMIT;
        aStateTable2 = new _DPState* [nAlloc];
        nStates2 = 0;
        for(int i = 0; i < nStates1; i++) {
            pDPState1 = aStateTable1[i];
            int j = pDPState1->j;
            Set[0] = pDPState1->Set[0];
            Set[1] = pDPState1->Set[1];
            Set[2] = pDPState1->Set[2];
            Set[3] = pDPState1->Set[3];
            for(int k = Nodes-1; k >= 0; k--) {
                if(pDPState1->fCost + current_sol.ImpGraph(j,k) < 0) {
                    if(current_sol.Class(k) < 32) {
                        if(Set[0] & (1 << current_sol.Class(k)))
                            continue;
                        Set[0] |= (1 << current_sol.Class(k));
                    } else if(current_sol.Class(k) < 64) {
                        if(Set[1] & (1 << (current_sol.Class(k)-32)))
                            continue;
                        Set[1] |= (1 << (current_sol.Class(k)-32));
                    } else if(current_sol.Class(k) < 96) {
                        if(Set[2] & (1 << (current_sol.Class(k)-64)))
                            continue;
                        Set[2] |= (1 << (current_sol.Class(k)-64));
                    } else {
                        if(Set[3] & (1 << (current_sol.Class(k)-96)))
                            continue;
                        Set[3] |= (1 << (current_sol.Class(k)-96));
                    }

                    pDPState = StateHash[k].HashFind(pDPState1->i, Set,
                                                     (short)(pDPState1->Stage+1));
                    if(pDPState) {
                        if(pDPState->fCost > pDPState1->fCost + current_sol.ImpGraph(j,k)) {
                            pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                            pDPState->pred = pDPState1;
                        }
                    } else {
                        pDPState = Mem.DPStateAlloc();
                        if(pDPState == 0) {
                            delete [] aStateTable1;
                            delete [] aStateTable2;
                            delete [] StateHash;
                            cerr<<"Returning without cycle because of shortage of memory\n";
                            exit(1);
                        }
                        pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                        pDPState->i = pDPState1->i;
                        pDPState->j = k;

                        pDPState->Set[0] = Set[0];
                        pDPState->Set[1] = Set[1];
                        pDPState->Set[2] = Set[2];
                        pDPState->Set[3] = Set[3];

                        pDPState->pred = pDPState1;
                        pDPState->Stage = pDPState1->Stage + 1;

                        StateHash[k].HashInsert(pDPState);
                        if(nStates2 < nAlloc)
                            aStateTable2[nStates2++] = pDPState;
                        else {
                            _DPState **tmp = new _DPState* [2*nAlloc];
                            for(nStates2 = nAlloc-1; nStates2 >= 0; nStates2--)
                                tmp[nStates2] = aStateTable2[nStates2];

                            delete [] aStateTable2;
                            aStateTable2 = tmp;
                            nStates2 = nAlloc;
                            nAlloc *= 2;
                            aStateTable2[nStates2++] = pDPState;
                        }
                    }

                    if (pDPState->pred == pDPState1) {
                        /* -- Check for a cycle */
                        temp = pDPState->fCost + current_sol.ImpGraph(k,pDPState->i);
                        if(temp < fCostMin) {
                            if ((!tl.isTabu(k,current_sol.Class(pDPState->i),par_iter)
                                    && !tl.isTabu(j,current_sol.Class(k),par_iter))
                                    || current_sol.fitness + temp < best_ever) {
                                pMin = pDPState;
                                fCostMin = temp;
                                path = false;
                            }
                        }
                        /* -- Check for a path */
                        temp = pDPState->fCost + current_sol.conflicts[k]
                               - current_sol.conflicts[pDPState->i] + (fvalue) p_in->Edge(j,k);
                        if(temp < fCostMin) {
                            if (!tl.isTabu(j,current_sol.Class(k),par_iter)
                                    || current_sol.fitness + temp < best_ever) {
                                pMin = pDPState;
                                fCostMin = temp;
                                path = true;
                            }
                        }
                    }

                    if(current_sol.Class(k) < 32)
                        Set[0] &= ~(1 << current_sol.Class(k));
                    else if(current_sol.Class(k) < 64)
                        Set[1] &= ~(1 << (current_sol.Class(k)-32));
                    else if(current_sol.Class(k) < 96)
                        Set[2] &= ~(1 << (current_sol.Class(k)-64));
                    else
                        Set[3] &= ~(1 << (current_sol.Class(k)-96));
                } /*end if (..)*/
            }/*end k*/
        } /*end for i<nstates1*/
        nStates1 = nStates2;
        delete [] aStateTable1;
        aStateTable1 = aStateTable2;

        if(nStates1 > 0)
            qsort( aStateTable1, nStates1, sizeof(struct _DPState*), StateCmp);

        if(nStates1 > STATE_LIMIT)
            nStates1 = STATE_LIMIT;
    }

    delete [] aStateTable1;
    delete [] StateHash;

    if (path) {
        mv.type = 3;
        int i = pMin->Stage+2;

        mv.seq.clear();
        mv.seq.resize(i,0);

        mv.seq[0] = pMin->i;
        for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
            mv.seq[--i] = pDPState->j;
        mv.seq[--i] = pDPState->j;

        mv.delta =  fCostMin;
    } else {
        if (pMin != NULL) {
            mv.type = 2;
            int i = pMin->Stage+2;

            mv.seq.clear();
            mv.seq.resize(i+1,0);

            mv.seq[i] = pMin->i;
            for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
                mv.seq[--i] = pDPState->j;
            mv.seq[--i] = pDPState->j;
            mv.seq[0] = pDPState->i;

            mv.delta =  fCostMin;
        } else {
            mv.type = 1;
            mv.seq.clear();
            mv.seq.resize(2,0);
            do {
                mv.seq[0]=rg->ran0N(p_in->Vertices()-1);
                mv.seq[1]=rg->ran0N(current_sol.CurrColors()-1);
            } while (mv.seq[1]==(Vertex) current_sol.Class(mv.seq[0]));
            mv.delta = current_sol.Delta(mv.seq[0],mv.seq[1]);
        }
    }
    return;
}





//for best improvement small graphs. One Ex Union Path Union Cycles
void ColRevRedMinOneExOrNegHashCycleOrPathExplorerII::BestMove(const Coloring &current_sol,
        ColMove& mv) {

    _DPState *pMin = NULL;
    _MemStruct Mem;
    bool path = false;
    bool one_ex = true;
    fvalue temp = MAX_FVALUE;
    fvalue global_best = MAX_FVALUE;
    vector < pair<Vertex, Color> > candidates;
    Vertex Nodes = p_in->Vertices();
    //cout<<current_sol.CurrColors()<<endl;
    for(Vertex i = 0; i < Nodes; i++) {
        for (Color c=0 ; c < current_sol.CurrColors() ; c++) {
            if (current_sol.col[i]!=c) {
                {
                    temp = current_sol.Delta(i,c);
                    if (temp == global_best) {
                        pair<Vertex, Color> initPair;
                        initPair.first = i ;
                        initPair.second = c ;
                        candidates.push_back(initPair);
                    } else if (temp < global_best)  {
                        candidates.clear();
                        pair<Vertex, Color> initPair;
                        initPair.first = i ;
                        initPair.second = c ;
                        candidates.push_back(initPair);
                        global_best = temp;
                    }
                }
            }
        }//close for(Color c....)
    } //close for(Vertex i=...)

    if (global_best <= 0) //= there is a possible one_ex improvement
    {

        _DPState *pDPState, *pDPState1;
        _DPState **aStateTable1, **aStateTable2;
        int nStates1, nStates2, nAlloc;
        long num_states = 0;
        unsigned long Set[4];

        _HashTable *StateHash = new _HashTable [Nodes];

        aStateTable1 = new _DPState* [Nodes*Nodes];
        nStates1 = 0;

        for(Vertex i = 0; i < Nodes; i++) {
            for(Vertex j = 0; j < Nodes; j++) {
                if (current_sol.conflicts[i]>0 || current_sol.conflicts[j]>0) {
                    if((current_sol.col[i] != current_sol.col[j])
                            && (current_sol.ImpGraph(i,j) < 0)) {
                        //possible paths of length have been already considered above
                        {
                            pDPState = Mem.DPStateAlloc();
                            if(pDPState == NULL) {
                                delete [] aStateTable1;
                                delete [] StateHash;
                                cerr<<"Returning without cycle because of shortage of memory\n";
                                exit(1);
                            }
                            pDPState->Set[0] = pDPState->Set[1] = pDPState->Set[2]
                                                                  = pDPState->Set[3] = 0;
                            if(current_sol.col[i] < 32)
                                pDPState->Set[0] |= (1 << current_sol.col[i]);
                            else if(current_sol.col[i] < 64)
                                pDPState->Set[1] |= (1 << (current_sol.col[i]-32));
                            else if(current_sol.col[i] < 96)
                                pDPState->Set[2] |= (1 << (current_sol.col[i]-64));
                            else
                                pDPState->Set[3] |= (1 << (current_sol.col[i]-96));

                            if(current_sol.col[j] < 32)
                                pDPState->Set[0] |= (1 << current_sol.col[j]);
                            else if(current_sol.col[j] < 64)
                                pDPState->Set[1] |= (1 << (current_sol.col[j]-32));
                            else if(current_sol.col[j] < 96)
                                pDPState->Set[2] |= (1 << (current_sol.col[j]-64));
                            else
                                pDPState->Set[3] |= (1 << (current_sol.col[j]-96));

                            pDPState->i = i;
                            pDPState->j = j;
                            pDPState->pred = 0;
                            pDPState->fCost = current_sol.ImpGraph(i,j);
                            pDPState->Stage = 0;

                            //------------------------------------------------------------
                            StateHash[j].HashInsert(pDPState);
                            aStateTable1[nStates1++] = pDPState;
                            //------------------------------------------------------------
                            // -- Check for a cycle
                            temp = current_sol.ImpGraph(i,j) + current_sol.ImpGraph(j,i);

                            if(temp < global_best) {
                                pMin = pDPState;
                                global_best = temp;
                                path = false;
                                one_ex=false;
                            }
                        }
                    }
                }
            }//close for(Vertex j=...)
        }//close for(Vertex i=...)


        if(nStates1 > 0)
            qsort( aStateTable1, nStates1, sizeof(struct _DPState *), StateCmp);

        if(nStates1 > STATE_LIMIT)
            nStates1 = STATE_LIMIT;

        // -- Now that we have looked at all states of length 1, we look at bigger paths
        while(nStates1 > 0) {
            num_states += nStates1;
            nAlloc = nStates1*STATE_LIMIT;
            aStateTable2 = new _DPState* [nAlloc];
            nStates2 = 0;
            for(int i = 0; i < nStates1; i++) {
                pDPState1 = aStateTable1[i];
                int j = pDPState1->j;
                Set[0] = pDPState1->Set[0];
                Set[1] = pDPState1->Set[1];
                Set[2] = pDPState1->Set[2];
                Set[3] = pDPState1->Set[3];

                for(int k = Nodes-1; k >= 0; k--) {
                    if(pDPState1->fCost + current_sol.ImpGraph(j,k) < 0) {
                        {
                            if(current_sol.col[k] < 32) {
                                if(Set[0] & (1 << current_sol.col[k]))
                                    continue;
                                Set[0] |= (1 << current_sol.col[k]);
                            } else if(current_sol.col[k] < 64) {
                                if(Set[1] & (1 << (current_sol.col[k]-32)))
                                    continue;
                                Set[1] |= (1 << (current_sol.col[k]-32));
                            } else if(current_sol.col[k] < 96) {
                                if(Set[2] & (1 << (current_sol.col[k]-64)))
                                    continue;
                                Set[2] |= (1 << (current_sol.col[k]-64));
                            } else {
                                if(Set[3] & (1 << (current_sol.col[k]-96)))
                                    continue;
                                Set[3] |= (1 << (current_sol.col[k]-96));
                            }

                            pDPState = StateHash[k].HashFind(pDPState1->i, Set,
                                                             (short)(pDPState1->Stage+1));
                            if(pDPState) {
                                if(pDPState->fCost > pDPState1->fCost + current_sol.ImpGraph(j,k)) {
                                    pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                                    pDPState->pred = pDPState1;
                                }
                            } else {
                                pDPState = Mem.DPStateAlloc();
                                if(pDPState == 0) {
                                    delete [] aStateTable1;
                                    delete [] aStateTable2;
                                    delete [] StateHash;
                                    cerr<<"Returning without cycle because of shortage of memory\n";
                                    exit(1);
                                }
                                pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                                pDPState->i = pDPState1->i;
                                pDPState->j = k;

                                pDPState->Set[0] = Set[0];
                                pDPState->Set[1] = Set[1];
                                pDPState->Set[2] = Set[2];
                                pDPState->Set[3] = Set[3];

                                pDPState->pred = pDPState1;
                                pDPState->Stage = pDPState1->Stage + 1;

                                StateHash[k].HashInsert(pDPState);
                                if(nStates2 < nAlloc)
                                    aStateTable2[nStates2++] = pDPState;
                                else {
                                    _DPState **tmp = new _DPState* [2*nAlloc];
                                    for(nStates2 = nAlloc-1; nStates2 >= 0; nStates2--)
                                        tmp[nStates2] = aStateTable2[nStates2];

                                    delete [] aStateTable2;
                                    aStateTable2 = tmp;
                                    nStates2 = nAlloc;
                                    nAlloc *= 2;
                                    aStateTable2[nStates2++] = pDPState;
                                }
                            }

                            if (pDPState->pred == pDPState1) {
                                // -- Check for a cycle
                                temp = pDPState->fCost + current_sol.ImpGraph(k,pDPState->i);
                                if(temp < global_best) {
                                    //if (!tl.isTabu(k,current_sol.col[pDPState->i],par_iter))
                                    //checking if the starting node is tabu is not needed since
                                    //it is already checked when look for length 1
                                    pMin = pDPState;
                                    global_best = temp;
                                    path = false;
                                    one_ex=false;
                                }
                                // -- Check for a path
                                temp = pDPState->fCost + current_sol.conflicts[k]
                                       - current_sol.conflicts[pDPState->i] + (fvalue) p_in->Edge(j,k);
                                if(temp < global_best) {
                                    pMin = pDPState;
                                    global_best = temp;
                                    path = true;
                                    one_ex=false;
                                }
                            }

                            if(current_sol.col[k] < 32)
                                Set[0] &= ~(1 << current_sol.col[k]);
                            else if(current_sol.col[k] < 64)
                                Set[1] &= ~(1 << (current_sol.col[k]-32));
                            else if(current_sol.col[k] < 96)
                                Set[2] &= ~(1 << (current_sol.col[k]-64));
                            else
                                Set[3] &= ~(1 << (current_sol.col[k]-96));
                        }//close if(!is.Tabu..)
                    } //ends if (..)
                }//ends k
            } //ends for(i<nstates1)
            nStates1 = nStates2;
            delete [] aStateTable1;
            aStateTable1 = aStateTable2;

            if(nStates1 > 0)
                qsort( aStateTable1, nStates1, sizeof(struct _DPState*), StateCmp);

            if(nStates1 > STATE_LIMIT)
                nStates1 = STATE_LIMIT;
        }


        delete [] aStateTable1;
        delete [] StateHash;

    }//close if (global_best <= 0)

    if (one_ex) {
        mv.seq.clear();
        mv.seq.resize(2);
        mv.type= 1;
        if (candidates.size()>0) {
            int rnd = rg->ran0N(candidates.size()-1);
            mv.seq[0] = candidates[rnd].first;
            mv.seq[1] = candidates[rnd].second;
            mv.delta = global_best;
            //rc.seq.push_back(coloring.col[rc.seq[0]]);
        } else {
            do {
                mv.seq[0]=rg->ran0N(p_in->Vertices()-1);
                mv.seq[1]=rg->ran0N(current_sol.CurrColors()-1);
            } while (mv.seq[1]==(Vertex) current_sol.col[mv.seq[0]]);
            mv.delta = current_sol.Delta(mv.seq[0],mv.seq[1]);
        }
    } else {
        if (path) {
            mv.type = 3;
            int i = pMin->Stage+2;

            mv.seq.clear();
            mv.seq.resize(i,0);

            mv.seq[0] = pMin->i;
            _DPState *pDPState;
            for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
                mv.seq[--i] = pDPState->j;
            mv.seq[--i] = pDPState->j;

            mv.delta =  global_best;
        } else {
            mv.type = 2;
            int i = pMin->Stage+2;

            mv.seq.clear();
            mv.seq.resize(i+1,0);

            mv.seq[i] = pMin->i;
            _DPState *pDPState;
            for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
                mv.seq[--i] = pDPState->j;
            mv.seq[--i] = pDPState->j;
            mv.seq[0] = pDPState->i;

            mv.delta =  global_best;
        }
    }
    return;
}




/*used in first version paper!!!!*/
void ColRevRedMinOneExOrNegHashCycleOrPathExplorerII::BestMove(const Coloring &current_sol,
        ColMove& mv,
        TabuList &tl, unsigned long par_iter) {
    _DPState *pMin = NULL;
    _MemStruct Mem;
    bool path = false;
    bool one_ex = true;
    fvalue temp = MAX_FVALUE;
    fvalue global_best = MAX_FVALUE;
    vector < pair<Vertex, Color> > candidates;
    Vertex Nodes = p_in->Vertices();

    for(Vertex i = 0; i < Nodes; i++) {
        for (Color c=0 ; c < current_sol.CurrColors() ; c++) {
            if (current_sol.col[i]!=c) {
                if (!tl.isTabu(i,c,par_iter)) {
                    temp = current_sol.Delta(i,c);
                    if (temp == global_best) {
                        pair<Vertex, Color> initPair;
                        initPair.first = i ;
                        initPair.second = c ;
                        candidates.push_back(initPair);
                    } else if (temp < global_best)  {
                        candidates.clear();
                        pair<Vertex, Color> initPair;
                        initPair.first = i ;
                        initPair.second = c ;
                        candidates.push_back(initPair);
                        global_best = temp;
                    }
                }
            }
        }//close for(Color c....)
    } //close for(Vertex i=...)

    if (global_best <= 0) //= there is a possible one_ex improvement
    {
        //cout<<"ok"<<endl;
        _DPState *pDPState, *pDPState1;
        _DPState **aStateTable1, **aStateTable2;
        int nStates1, nStates2, nAlloc;
        long num_states = 0;
        unsigned long Set[4];

        _HashTable *StateHash = new _HashTable [Nodes];

        aStateTable1 = new _DPState* [Nodes*Nodes];
        nStates1 = 0;

        for(Vertex i = 0; i < Nodes; i++) {
            for(Vertex j = 0; j < Nodes; j++) {
                if (current_sol.conflicts[i]>0 || current_sol.conflicts[j]>0) {
                    if((current_sol.col[i] != current_sol.col[j])
                            && (current_sol.ImpGraph(i,j) < 0)) {
                        if (!tl.isTabu(i,current_sol.col[j],par_iter)
                                && !tl.isTabu(j,current_sol.col[i],par_iter))
                        {
                        	//possible paths of length have been already considered above
                        	                            pDPState = Mem.DPStateAlloc();
                        if(pDPState == NULL) {
                            delete [] aStateTable1;
                            delete [] StateHash;
                            cerr<<"Returning without cycle because of shortage of memory\n";
                            exit(1);
                        }
                        pDPState->Set[0] = pDPState->Set[1] = pDPState->Set[2]
                                                              = pDPState->Set[3] = 0;
                        if(current_sol.col[i] < 32)
                            pDPState->Set[0] |= (1 << current_sol.col[i]);
                        else if(current_sol.col[i] < 64)
                            pDPState->Set[1] |= (1 << (current_sol.col[i]-32));
                        else if(current_sol.col[i] < 96)
                            pDPState->Set[2] |= (1 << (current_sol.col[i]-64));
                        else
                            pDPState->Set[3] |= (1 << (current_sol.col[i]-96));

                        if(current_sol.col[j] < 32)
                            pDPState->Set[0] |= (1 << current_sol.col[j]);
                        else if(current_sol.col[j] < 64)
                            pDPState->Set[1] |= (1 << (current_sol.col[j]-32));
                        else if(current_sol.col[j] < 96)
                            pDPState->Set[2] |= (1 << (current_sol.col[j]-64));
                        else
                            pDPState->Set[3] |= (1 << (current_sol.col[j]-96));

                        pDPState->i = i;
                        pDPState->j = j;
                        pDPState->pred = 0;
                        pDPState->fCost = current_sol.ImpGraph(i,j);
                        pDPState->Stage = 0;

                        //------------------------------------------------------------
                        StateHash[j].HashInsert(pDPState);
                        aStateTable1[nStates1++] = pDPState;
                        //------------------------------------------------------------
                        // -- Check for a cycle
                        temp = current_sol.ImpGraph(i,j) + current_sol.ImpGraph(j,i);

                        if(temp < global_best) {
                            pMin = pDPState;
                            global_best = temp;
                            path = false;
                            one_ex=false;
                        }
                    }
                }
            }
        }//close for(Vertex j=...)
    }//close for(Vertex i=...)


    if(nStates1 > 0)
        qsort( aStateTable1, nStates1, sizeof(struct _DPState *), StateCmp);

    if(nStates1 > STATE_LIMIT)
        nStates1 = STATE_LIMIT;

    // -- Now that we have looked at all states of length 1, we look at bigger paths
    while(nStates1 > 0) {
        num_states += nStates1;
        nAlloc = nStates1*50;
        aStateTable2 = new _DPState* [nAlloc];
        nStates2 = 0;
        for(int i = 0; i < nStates1; i++) {
            pDPState1 = aStateTable1[i];
            int j = pDPState1->j;
            Set[0] = pDPState1->Set[0];
            Set[1] = pDPState1->Set[1];
            Set[2] = pDPState1->Set[2];
            Set[3] = pDPState1->Set[3];

            for(int k = Nodes-1; k >= 0; k--) {
                if(pDPState1->fCost + current_sol.ImpGraph(j,k) < 0) {
                    if (!tl.isTabu(j,current_sol.col[k],par_iter)) {
                        if(current_sol.col[k] < 32) {
                            if(Set[0] & (1 << current_sol.col[k]))
                                continue;
                            Set[0] |= (1 << current_sol.col[k]);
                        } else if(current_sol.col[k] < 64) {
                            if(Set[1] & (1 << (current_sol.col[k]-32)))
                                continue;
                            Set[1] |= (1 << (current_sol.col[k]-32));
                        } else if(current_sol.col[k] < 96) {
                            if(Set[2] & (1 << (current_sol.col[k]-64)))
                                continue;
                            Set[2] |= (1 << (current_sol.col[k]-64));
                        } else {
                            if(Set[3] & (1 << (current_sol.col[k]-96)))
                                continue;
                            Set[3] |= (1 << (current_sol.col[k]-96));
                        }

                        pDPState = StateHash[k].HashFind(pDPState1->i, Set,
                                                         (short)(pDPState1->Stage+1));
                        if(pDPState) {
                            if(pDPState->fCost > pDPState1->fCost + current_sol.ImpGraph(j,k)) {
                                pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                                pDPState->pred = pDPState1;
                            }
                        } else {
                            pDPState = Mem.DPStateAlloc();
                            if(pDPState == 0) {
                                delete [] aStateTable1;
                                delete [] aStateTable2;
                                delete [] StateHash;
                                cerr<<"Returning without cycle because of shortage of memory\n";
                                exit(1);
                            }
                            pDPState->fCost = pDPState1->fCost + current_sol.ImpGraph(j,k);
                            pDPState->i = pDPState1->i;
                            pDPState->j = k;

                            pDPState->Set[0] = Set[0];
                            pDPState->Set[1] = Set[1];
                            pDPState->Set[2] = Set[2];
                            pDPState->Set[3] = Set[3];

                            pDPState->pred = pDPState1;
                            pDPState->Stage = pDPState1->Stage + 1;

                            StateHash[k].HashInsert(pDPState);
                            if(nStates2 < nAlloc)
                                aStateTable2[nStates2++] = pDPState;
                            else {
                                _DPState **tmp = new _DPState* [2*nAlloc];
                                for(nStates2 = nAlloc-1; nStates2 >= 0; nStates2--)
                                    tmp[nStates2] = aStateTable2[nStates2];

                                delete [] aStateTable2;
                                aStateTable2 = tmp;
                                nStates2 = nAlloc;
                                nAlloc *= 2;
                                aStateTable2[nStates2++] = pDPState;
                            }
                        }

                        if (pDPState->pred == pDPState1) {
                            // -- Check for a cycle
                            temp = pDPState->fCost + current_sol.ImpGraph(k,pDPState->i);
                            if(temp < global_best) {
                                //if (!tl.isTabu(k,current_sol.col[pDPState->i],par_iter))
                                //checking if the starting node is tabu is not needed since
                                //it is already checked when look for length 1
                                pMin = pDPState;
                                global_best = temp;
                                path = false;
                                one_ex=false;
                            }
                            // -- Check for a path
                            temp = pDPState->fCost + current_sol.conflicts[k]
                                   - current_sol.conflicts[pDPState->i] + (fvalue) p_in->Edge(j,k);
                            if(temp < global_best) {
                                pMin = pDPState;
                                global_best = temp;
                                path = true;
                                one_ex=false;
                            }
                        }

                        if(current_sol.col[k] < 32)
                            Set[0] &= ~(1 << current_sol.col[k]);
                        else if(current_sol.col[k] < 64)
                            Set[1] &= ~(1 << (current_sol.col[k]-32));
                        else if(current_sol.col[k] < 96)
                            Set[2] &= ~(1 << (current_sol.col[k]-64));
                        else
                            Set[3] &= ~(1 << (current_sol.col[k]-96));
                    }//close if(!is.Tabu..)
                } //ends if (..)
            }//ends k
        } //ends for(i<nstates1)
        nStates1 = nStates2;
        delete [] aStateTable1;
        aStateTable1 = aStateTable2;

        if(nStates1 > 0)
            qsort( aStateTable1, nStates1, sizeof(struct _DPState*), StateCmp);

        if(nStates1 > STATE_LIMIT)
            nStates1 = STATE_LIMIT;
    }


    delete [] aStateTable1;
    delete [] StateHash;

}//close if (global_best <= 0)


if (one_ex) {
    mv.seq.clear();
    mv.seq.resize(2);
    mv.type= 1;
    if (candidates.size()>0) {
        int rnd = rg->ran0N(candidates.size()-1);
        mv.seq[0] = candidates[rnd].first;
        mv.seq[1] = candidates[rnd].second;
        mv.delta = global_best;
        //rc.seq.push_back(coloring.col[rc.seq[0]]);
    } else {
        do {
            mv.seq[0]=rg->ran0N(p_in->Vertices()-1);
            mv.seq[1]=rg->ran0N(current_sol.CurrColors()-1);
        } while (mv.seq[1]== (Vertex) current_sol.col[mv.seq[0]]);
        mv.delta = current_sol.Delta(mv.seq[0],mv.seq[1]);
    }
} else {
    if (path) {
        mv.type = 3;
        int i = pMin->Stage+2;

        mv.seq.clear();
        mv.seq.resize(i,0);

        mv.seq[0] = pMin->i;
        _DPState *pDPState;
        for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
            mv.seq[--i] = pDPState->j;
        mv.seq[--i] = pDPState->j;

        mv.delta =  global_best;
    } else {
        mv.type = 2;
        int i = pMin->Stage+2;

        mv.seq.clear();
        mv.seq.resize(i+1,0);

        mv.seq[i] = pMin->i;
        _DPState *pDPState;
        for(pDPState = pMin; pDPState->pred; pDPState = pDPState->pred)
            mv.seq[--i] = pDPState->j;
        mv.seq[--i] = pDPState->j;
        mv.seq[0] = pDPState->i;

        mv.delta =  global_best;
    }
}
return;
}

#endif

