/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <algorithm>
#include <cmath>
#include <iostream>
#include <sstream>
#include <list>

#include "Solution.h"

using namespace std;

#define CHARBITS  8

extern long int *array;




void ColoringManager::constructiveDSATUR(Coloring& ass)
{ 
  Color num_col=1;
  vector<Vertex> dsat(p_in->Vertices());
  Vertex node;
  
  VertexSet::iterator p;

  list<Vertex> toColor;
  list<Vertex>::iterator pt,pt_node;

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Vertex j = 0; j < (unsigned) (p_in->Vertices()); j++)
	Bitmap[i][j] = '\0';
    }

  array = rg->generate_array(p_in->Vertices());
  
  dsat = p_in->degree;
  
  for (Vertex i = 0; i < p_in->Vertices(); i++)
    {
      toColor.push_back(array[i]);
      ass.col[i] = -1;
    }

  
  bool first = true;

  while (!toColor.empty())
    {
    
      pt = pt_node = toColor.begin();
      node = *pt_node;
      if (first)   //select node with max degree
	{
	  while (pt!=toColor.end())
	    {
	      if (dsat[node] < dsat[*pt])
		{
		  node = *pt;
		  pt_node = pt;
		}
	      pt++;
	    }
	  first = false;
	  
	}
      else //select node with largets saturation and break ties by largest degree
	{
	  bool passed = false;
	  
	  while (pt!=toColor.end())
	    {
	      if (dsat[*pt] != p_in->degree[*pt])
		{
		  if  (dsat[node] < dsat[*pt])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		  else if  ((dsat[node]==dsat[*pt]) 
			    && (p_in->degree[*pt] > p_in->degree[node]))
		    { //second criterion
		      node = *pt;
		      pt_node = pt;
		    }			
		  passed = true;
		}
	      else if (!passed)
		{
		  if (p_in->degree[*pt] > p_in->degree[node])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		}
	      pt++;
	    }
	}
      
      //cout<<node<<endl;
      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	num_col++;

      ass.col[node] = col;
      toColor.erase(pt_node);
      //Update DSAT
      //Dsat has to be updated for the nodes w adjacent to "node" not yet coloured:
      //1. dsat has to be updated
      //2. the Bitmap with colors used must be updated
      for (p=p_in->adjacency_set[node].begin();
	   p!=p_in->adjacency_set[node].end(); p++)
	{
	  if (ass.col[*p] < 0)
	    {
	      if (dsat[*p]==p_in->degree[*p])
		dsat[*p]=0;
	      if (!(Bitmap[(col)/CHARBITS][(*p)] & (1<<((col)%CHARBITS))))
		{
		  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
		  dsat[*p]++;
		}
	      
	    }
	}
    } 
 
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap;
 
  ass.SetColors(num_col);
  updateColoringData(ass);  
}





void ColoringManager::constructiveDSATURBounded(Coloring& ass, Color num_col)
{ 
  vector<Vertex> dsat(p_in->Vertices());
  Vertex node;
  
  VertexSet::iterator p;

  list<Vertex> toColor;
  list<Vertex>::iterator pt,pt_node;

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[num_col];
      for (Color j = 0; j < (num_col); j++)
	Bitmap[i][j] = '\0';
    }

  array = rg->generate_array(p_in->Vertices());
  
  dsat = p_in->degree;
  
  for (Vertex i = 0; i < p_in->Vertices(); i++)
    {
      toColor.push_back(array[i]);
      ass.col[i] = -1;
    }

  
  bool first = true;

  while (!toColor.empty())
    {
    
      pt = pt_node = toColor.begin();
      node = *pt_node;
      if (first)   //select node with max degree
	{
	  while (pt!=toColor.end())
	    {
	      if (dsat[node] < dsat[*pt])
		{
		  node = *pt;
		  pt_node = pt;
		}
	      pt++;
	    }
	  first = false;
	}
      else //select node with largets saturation and break ties by largest degree
	{
	  bool passed = false;
	  
	  while (pt!=toColor.end())
	    {
	      if (dsat[*pt] != p_in->degree[*pt])
		{
		  if  (dsat[node] < dsat[*pt])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		  else if  ((dsat[node]==dsat[*pt]) 
			    && (p_in->degree[*pt] > p_in->degree[node]))
		    { //second criterion
		      node = *pt;
		      pt_node = pt;
		    }			
		  passed = true;
		}
	      else if (!passed)
		{
		  if (p_in->degree[*pt] > p_in->degree[node])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		}
	      pt++;
	    }
	}
      

      Color col = 0;
      while (col<num_col && (Bitmap[(node)/CHARBITS][(col)] & (1<<((node)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	{
	  col = rg->ran0N(num_col-1);
	}
      
      ass.col[node] = col;
      toColor.erase(pt_node);
      //Update DSAT
      //Dsat has to be updated for the nodes w adjacent to "node" not yet coloured:
      //1. dsat has to be updated
      //2. the Bitmap with colors used must be updated
      for (p=p_in->adjacency_set[node].begin();
	   p!=p_in->adjacency_set[node].end(); p++)
	{
	  if (ass.col[*p] < 0)
	    {
	      if (dsat[*p]==p_in->degree[*p])
		dsat[*p]=0;
	      if (!(Bitmap[(*p)/CHARBITS][(col)] & (1<<((*p)%CHARBITS))))
		{
		  Bitmap[(*p)/CHARBITS][(col)] |= (1 << ((*p) % CHARBITS));
		  dsat[*p]++;
		}
	      
	    }
	}
    } 
 
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap;
 
  ass.SetColors(num_col);
  updateColoringData(ass);  
}






void ColoringManager::constructiveDSATURPartial(Coloring& ass, Color used_c)
{ 
  Color num_col=used_c;
  vector<Vertex> dsat(p_in->Vertices());
  Vertex node;
  
  VertexSet::iterator p;

  list<Vertex> toColor;
  list<Vertex>::iterator pt,pt_node;

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Vertex j = 0; j < (unsigned) (p_in->Vertices()); j++)
	Bitmap[i][j] = '\0';
    }

  array = rg->generate_array(p_in->Vertices());
  
  dsat = p_in->degree;
  
  for (Vertex i = 0; i < p_in->Vertices(); i++)
    {
      node = array[i];
      if (ass.col[node] != (Color) p_in->Vertices())
	{
	  for (p=p_in->adjacency_set[node].begin();
	       p!=p_in->adjacency_set[node].end(); p++)
	    {
	      Color col = ass.col[node];
	       if (dsat[*p]==p_in->degree[*p])
		dsat[*p]=0;
	      if (!(Bitmap[(col)/CHARBITS][(*p)] & (1<<((col)%CHARBITS))))
		{
		  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
		  dsat[*p]++;
		}

	    }
	}
      else
	{
	  toColor.push_back(node);
	  ass.col[node] = -1;
	}
    }

  
  bool first = true;

  while (!toColor.empty())
    {
    
      pt = pt_node = toColor.begin();
      node = *pt_node;
      if (first)   //select node with max degree
	{
	  while (pt!=toColor.end())
	    {
	      if (dsat[node] < dsat[*pt])
		{
		  node = *pt;
		  pt_node = pt;
		}
	      pt++;
	    }
	  first = false;
	  
	}
      else //select node with largets saturation and break ties by largest degree
	{
	  bool passed = false;
	  
	  while (pt!=toColor.end())
	    {
	      if (dsat[*pt] != p_in->degree[*pt])
		{
		  if  (dsat[node] < dsat[*pt])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		  else if  ((dsat[node]==dsat[*pt]) 
			    && (p_in->degree[*pt] > p_in->degree[node]))
		    { //second criterion
		      node = *pt;
		      pt_node = pt;
		    }			
		  passed = true;
		}
	      else if (!passed)
		{
		  if (p_in->degree[*pt] > p_in->degree[node])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		}
	      pt++;
	    }
	}
      

      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	num_col++;

      ass.col[node] = col;
      toColor.erase(pt_node);
      //Update DSAT
      //Dsat has to be updated for the nodes w adjacent to "node" not yet coloured:
      //1. dsat has to be updated
      //2. the Bitmap with colors used must be updated
      for (p=p_in->adjacency_set[node].begin();
	   p!=p_in->adjacency_set[node].end(); p++)
	{
	  if (ass.col[*p] < 0)
	    {
	      if (dsat[*p]==p_in->degree[*p])
		dsat[*p]=0;
	      if (!(Bitmap[(col)/CHARBITS][(*p)] & (1<<((col)%CHARBITS))))
		{
		  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
		  dsat[*p]++;
		}
	      
	    }
	}
    } 
 
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap;
 
  ass.SetColors(num_col);
  updateColoringData(ass);  
}






void ColoringManager::constructiveDSATURPartialBounded(Coloring& ass, Color num_col)
{ 
  vector<Vertex> dsat(p_in->Vertices());
  Vertex node;
  
  VertexSet::iterator p;

  list<Vertex> toColor;
  list<Vertex>::iterator pt,pt_node;

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Color j = 0; j < (Color) p_in->Vertices(); j++)
	Bitmap[i][j] = '\0';
    }

  array = rg->generate_array(p_in->Vertices());
  
  dsat = p_in->degree;
  
  for (Vertex i = 0; i < p_in->Vertices(); i++)
    {
      node = array[i];
      if (ass.col[node] != (Color) p_in->Vertices())
	{
	  for (p=p_in->adjacency_set[node].begin();
	       p!=p_in->adjacency_set[node].end(); p++)
	    {
	      Color col = ass.col[node];
	       if (dsat[*p]==p_in->degree[*p])
		dsat[*p]=0;
	      if (!(Bitmap[(col)/CHARBITS][(*p)] & (1<<((col)%CHARBITS))))
		{
		  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
		  dsat[*p]++;
		}

	    }
	}
      else
	{
	  toColor.push_back(node);
	  ass.col[node] = -1;
	}
    }

  
  bool first = true;

  while (!toColor.empty())
    {
    
      pt = pt_node = toColor.begin();
      node = *pt_node;
      if (first)   //select node with max degree
	{
	  while (pt!=toColor.end())
	    {
	      if (dsat[node] < dsat[*pt])
		{
		  node = *pt;
		  pt_node = pt;
		}
	      pt++;
	    }
	  first = false;
	  
	}
      else //select node with largets saturation and break ties by largest degree
	{
	  bool passed = false;
	  
	  while (pt!=toColor.end())
	    {
	      if (dsat[*pt] != p_in->degree[*pt])
		{
		  if  (dsat[node] < dsat[*pt])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		  else if  ((dsat[node]==dsat[*pt]) 
			    && (p_in->degree[*pt] > p_in->degree[node]))
		    { //second criterion
		      node = *pt;
		      pt_node = pt;
		    }			
		  passed = true;
		}
	      else if (!passed)
		{
		  if (p_in->degree[*pt] > p_in->degree[node])
		    {
		      node = *pt;
		      pt_node = pt;
		    }
		}
	      pt++;
	    }
	}
      

      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
       if (col>=num_col)
	{
	  col = rg->ran0N(num_col-1);
	}

      ass.col[node] = col;
      toColor.erase(pt_node);
      //Update DSAT
      //Dsat has to be updated for the nodes w adjacent to "node" not yet coloured:
      //1. dsat has to be updated
      //2. the Bitmap with colors used must be updated
      for (p=p_in->adjacency_set[node].begin();
	   p!=p_in->adjacency_set[node].end(); p++)
	{
	  if (ass.col[*p] < 0)
	    {
	      if (dsat[*p]==p_in->degree[*p])
		dsat[*p]=0;
	      if (!(Bitmap[(col)/CHARBITS][(*p)] & (1<<((col)%CHARBITS))))
		{
		  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
		  dsat[*p]++;
		}
	      
	    }
	}
    } 
 
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap;
 
  ass.SetColors(num_col);
  updateColoringData(ass);  
}
