/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <algorithm>
#include <cmath>
#include <iostream>
#include <string>
#include<fstream>
#include <sstream>
#include <list>
#include "Solution.h"

#define CHARBITS 8

extern long int *array;

void ColoringManager::randomAssignment(Coloring& ass, Color num_col)
{
    Color rnd;
    ass.SetColors(num_col);
    for (Vertex v = 0; v < p_in->Vertices(); v++)
    {
	rnd = (unsigned int) rg->ran0N(num_col-1);
	ass.col[v]=rnd;
    }
    
    //cout<<ass;
    //exit(0);
    updateColoringData(ass);
}


void ColoringManager::readStartingSolution(Coloring& ass, Color num_col)
{
    ostringstream file; //string x = s.str();
  Color col;
  vector<Color> tmp_col;
 
  if (strlen(p_ctrl->getInputSolutionStream())>0)
  {
      file << p_ctrl->getInputSolutionStream();
  }
  else 
  {
      char *tmp = "";
      strncpy(tmp,p_ctrl->is,strlen(p_ctrl->is)-4);
      file << "starting_solutions/" << basename(tmp) << "." << num_col << "." << p_ctrl->getNumTry()-1 << ".col";
  }
  ifstream is(file.str().c_str());
  if (!is.is_open())
  {
      cout<<"file "<<file.str()<<" not open"<<endl;
      exit(0);
  }

  do 
  {
      is>>col;
      //assert(col<num_col);
      tmp_col.push_back(col);
  } while(!is.eof());//for (Vertex v = 0; v < p_in->Vertices(); v++)
  
  for (Vertex v = 0; v < p_in->Vertices(); v++)
  {
    ass.col[v] = tmp_col[v];
  }
 
  //cout<<num_col<<endl;
  ass.SetColors(num_col);
  updateColoringData(ass);
  is.close();
}


//----------------------------------------------------------------------
//----------------------------------------------------------------------
//----------------------------------------------------------------------


void ColoringManager::constructiveGreedy(Coloring &ass)
{
  Color num_col = 1;
  Vertex node;
  VertexSet::iterator pt;

  array = rg->generate_array(p_in->Vertices());

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Vertex j = 0; j < (unsigned) (p_in->Vertices()); j++)
	{
	  ass.col[i] = -1;
	  Bitmap[i][j] = '\0';
	}
    }

  
  for(Vertex i = 0; i < p_in->Vertices(); i++) 
    {
      node = array[i];

      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	num_col++;

      ass.col[node] = col;
      
      for (pt=p_in->adjacency_set[node].begin();
	   pt!=p_in->adjacency_set[node].end(); pt++)
	{
	  Bitmap[(col)/CHARBITS][(*pt)] |= (1 << ((col) % CHARBITS));
	}
    }
  
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap; 
  ass.SetColors(num_col);
  updateColoringData(ass);
}


void ColoringManager::constructiveGreedyBounded(Coloring &ass, Color num_col)
{
  Vertex node;
  VertexSet::iterator pt;

  array = rg->generate_array(p_in->Vertices());

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Vertex j = 0; j < (unsigned) (p_in->Vertices()); j++)
	{
	  ass.col[i] = -1;
	  Bitmap[i][j] = '\0';
	}
    }

  
  for(Vertex i = 0; i < p_in->Vertices(); i++) 
    {
      node = array[i];

      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	col = rg->ran0N(num_col-1);

      ass.col[node] = col;
      
      for (pt=p_in->adjacency_set[node].begin();
	   pt!=p_in->adjacency_set[node].end(); pt++)
	{
	  Bitmap[(col)/CHARBITS][(*pt)] |= (1 << ((col) % CHARBITS));
	}
    }
  
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap; 
  ass.SetColors(num_col);
  updateColoringData(ass);
}



void ColoringManager::constructiveGreedyPartialBounded(Coloring &ass, Color num_col)
{
  Vertex node;
  VertexSet::iterator p;

  array = rg->generate_array(p_in->Vertices());

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Color j = 0; j < num_col; j++)
	{
	
	  Bitmap[i][j] = '\0';
	}
    }
 
  for (Vertex i = 0; i < p_in->Vertices(); i++)
    {
      if (ass.col[i] != (Color) p_in->Vertices())
	{
	  for (p=p_in->adjacency_set[i].begin();
	       p!=p_in->adjacency_set[i].end(); p++)
	    {
	      Bitmap[(ass.col[i])/CHARBITS][(*p)] |= (1 << ((ass.col[i]) % CHARBITS));
	    }
	}
      else
	{
	  ass.col[i] = -1;
	}
    }
  
  for(Vertex i = 0; i < p_in->Vertices(); i++) 
    {
      node = array[i];
      if (ass.col[node] >= 0) continue;
      
      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	col = rg->ran0N(num_col-1);

      ass.col[node] = col;
      
      for (p=p_in->adjacency_set[node].begin();
	   p!=p_in->adjacency_set[node].end(); p++)
	{
	  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
	}
    }
  
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap; 
  ass.SetColors(num_col);
  updateColoringData(ass);
}





void ColoringManager::constructiveGreedyPartial(Coloring &ass, Color num_col)
{
  Vertex node;
  VertexSet::iterator p;

  array = rg->generate_array(p_in->Vertices());

  unsigned char **Bitmap;
  unsigned int x = (p_in->Vertices()/CHARBITS+1);
  Bitmap = new unsigned char *[x];
  for (Vertex i = 0; i < x; i++)
    {
      Bitmap[i] = new unsigned char[p_in->Vertices()];
      for (Vertex j = 0; j < (unsigned) (p_in->Vertices()); j++)
	{
	
	  Bitmap[i][j] = '\0';
	}
    }
 
  for (Vertex i = 0; i < p_in->Vertices(); i++)
    {
      if (ass.col[i] != (Color) p_in->Vertices())
	{
	  for (p=p_in->adjacency_set[i].begin();
	       p!=p_in->adjacency_set[i].end(); p++)
	    {
	      Bitmap[(ass.col[i])/CHARBITS][(*p)] |= (1 << ((ass.col[i]) % CHARBITS));
	    }
	}
      else
	{
	  ass.col[i] = -1;
	}
    }
  
  for(Vertex i = 0; i < p_in->Vertices(); i++) 
    {
      node = array[i];
      if (ass.col[node] >= 0) continue;
      
      Color col = 0;
      while (col<num_col && (Bitmap[(col)/CHARBITS][(node)] & (1<<((col)%CHARBITS))))
	col++;
      
      if (col>=num_col)
	num_col++;

      ass.col[node] = col;
      
      for (p=p_in->adjacency_set[node].begin();
	   p!=p_in->adjacency_set[node].end(); p++)
	{
	  Bitmap[(col)/CHARBITS][(*p)] |= (1 << ((col) % CHARBITS));
	}
    }
  
  delete [] array;        
  for (Vertex i = 0; i < (p_in->Vertices()/CHARBITS+1); i++)
    delete [] Bitmap[i];
  delete [] Bitmap; 
  ass.SetColors(num_col);
  updateColoringData(ass);
}

