
#include <cstdio>
#include <cmath>
#include <cassert>
#include "Random.h"
using namespace std;

#define VERBOSE(x) x
#define VERYVERBOSE(x)

/* pseudo-random number generator as proposed in numerical recipes in C 
   Input:   a long value; has to be the seed variable
   Output:  a pseudo-random number uniformly distributed in [0,1]
   Side effects: changes the value of the input variable, must be this way
*/
double Random::ran01()
{
  long k;
  double ans;
  k =(my_seed)/IQ;
  my_seed = IA * (my_seed - k * IQ) - IR * k;
  if (my_seed < 0 ) my_seed += IM;
  ans = AM * (my_seed);
  return ans;
}

double rint(double x)
{
  return (int)(x+0.5);
}

/*ran0N returns a value in [0,max_ran]*/
int Random::ran0N(int max_ran)  {
  double o;
  o = (double) rint(ran01()*max_ran);
  /*printf("ran: %ld seed: %ld maxrun: %ld\n",o,seed,max_ran);*/
  return (int)o;
}

long int* Random::generate_array(const int& size) {
   long int  i, j, help;
   long int  *v;

   v = new long int[size];

   for ( i = 0 ; i < size; i++ ) 
     v[i] = i;

   for ( i = 0 ; i < size-1 ; i++) {
     j = (long int) ( ran01( ) * (size - i)); 
     assert( i + j < size );
     help = v[i];
     v[i] = v[i+j];
     v[i+j] = help;
   }
   VERYVERBOSE ( printf("Random vector:\n");
   for (i = 0 ; i < size ; i++ ) 
     printf(" %ld ",v[i]);
   printf("\n"); )
   return v;
}

