/* GCP version 6.0 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Runners.h"

#include <algorithm>
#include <cmath>
#include <iostream>
using namespace std;


//------------------------------------------------------------
//--------------------COLORING RUNNERS------------------------
//------------------------------------------------------------
ColoringRunner::ColoringRunner(ColoringManager* s,
                               Random *rnd, Control *ctrl, Graph* in = NULL)
{
    p_sm = s;
    p_in = in;
    rg = rnd;
    p_ctrl = ctrl;
    number_of_iterations = 0;

}

ColoringRunner::ColoringRunner(ColoringManager* s,  ColNeighborhoodExplorer* e,
                               Random *rnd, Control *ctrl, Graph* in = NULL)
{
    p_sm = s;
    p_ne = e;
    p_in = in;
    rg = rnd;
    p_ctrl = ctrl;
    number_of_iterations = 0;

}

void ColoringRunner::InitializeRun()
{
    idle_iterations = 0;
    number_of_iterations = 0;
}

void ColoringRunner::initialization_procedure(Coloring &ass)
{
  
    switch (p_ctrl->init)
    {
    case 0:
        p_sm->randomAssignment(ass,p_ctrl->k);
        break;
    case 1:
        p_sm->constructiveGreedy(ass);
        break;
    case 2:

        p_sm->constructiveGreedyBounded(ass,p_ctrl->k);
        break;
    case 3:
        p_sm->constructiveRecursiveLargestFirst(ass);
        break;
    case 4:
        //cout<<p_ctrl->k<<" "<<p_ctrl->init <<endl;
        p_sm->constructiveRecursiveLargestFirstBounded(ass,p_ctrl->k);
        break;
    case 5:
        p_sm->constructiveDSATUR(ass);
        break;
    case 6:
        p_sm->constructiveDSATURBounded(ass,p_ctrl->k);
        break;
    case 7:
        p_sm->readStartingSolution(ass,p_ctrl->k);
        break;
    case 9:
        // cout<<p_ctrl->k<<" "<<p_ctrl->init <<endl;
        p_sm->constructiveRecursiveLargestFirstStatic(ass);
        break;
    case 11:
        p_sm->constructiveRecursiveLargestFirstReverse(ass);
        break;
    default:
        cerr<<"Errors in initialization\n";
        exit(1);
    }
}


void ColoringRunner::decreasing_procedure(Coloring &ass)
{
    switch (p_ctrl->decr)
    {
    case 0:
        p_sm->decrease_colors(ass);
        break;
    case 1:
        //decrease_colors_with_dsat(ass,k);
        //ass.col = --k;
        break;
    case 2:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->constructiveGreedyBounded(ass,ass.CurrColors());
        break;
    case 3:
        //decrease_colors_with_dsat(ass,k);
        //ass.col = --k;
        break;
    case 4:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->constructiveRecursiveLargestFirstBounded(ass,ass.CurrColors());
        break;
    case 5:
        //decrease_colors_with_dsat(ass,k);
        //ass.col = --k;
        break;
    case 6:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->constructiveDSATURBounded(ass,ass.CurrColors());
        break;

        //------------------------------------------------------------
    case 10:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->randomAssignment(ass,ass.CurrColors());
        break;

    case 12:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->constructiveGreedyBounded(ass,ass.CurrColors());
        break;

    case 14:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->constructiveRecursiveLargestFirstBounded(ass,ass.CurrColors());
        break;

    case 16:
        ass.SetColors(ass.CurrColors()-1);
        p_sm->constructiveDSATURBounded(ass,ass.CurrColors());
        break;
    default:
        cerr<<"something wrong!!\n";
        exit(1);
    }
}



void ColoringRunner::Run_procedure(Coloring& initial_ass)
{
    switch (p_ctrl->test_version)
    {
    case 0:
        Run(initial_ass);
        break;
    case 1:
        TestRun(initial_ass);
        break;
    default:
        fprintf(stderr,"something wrong in LS!!\n");
        exit(1);
    }
}




void ColoringRunner::TestRun(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;

    // cout<<current_state;

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
        p_ne->BestMove(current_state,mv);

        if (mv.type > 0 && mv.delta < 0)
        {
            Coloring test_state(current_state);

            if (!(test_state == current_state))
            {
                cerr<<"before"<<endl;
                exit(1);
            }

            cout<<mv<<endl;
            p_ne->MakeMove(current_state,mv);
            //cout << current_state<<endl;
            p_ne->MakeMove(test_state,mv);
            //test_state.col[mv.seq[0]] = mv.seq[1];

            //	  current_state.fitness += mv.delta;

            p_sm->updateColoringData(test_state);
            //cout << test_state<<endl;
            if (!(test_state == current_state))
            {
                exit(1);
            }

            //cout << current_state;

            if (current_state.fitness != 0)
                current_state.feasible = false;
            else
                current_state.feasible = true;

            if (current_state.fitness < initial_ass.fitness)
            {
                initial_ass = current_state;
                p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
                idle_iterations = 0;
            }

        }
        else
            break;
        idle_iterations++;
        UpdateIterationCounter();
    }
    TerminateRun();
}


void ColoringRunner::Run(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;

    while ((current_state.fitness != 0) && !MaxIterationExpired() &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {

        //cout<<idle_iterations<<endl;
        p_ne->BestMove(current_state,mv);

        if (mv.type > 0 && mv.delta < 0)
        {
            p_ne->MakeMove(current_state,mv);
            if (current_state.fitness != 0)
                current_state.feasible = false;
            else
            {
                current_state.feasible = true;
            }
            if (current_state.fitness < initial_ass.fitness)
            {
                initial_ass = current_state;

                p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
                idle_iterations = 0;
            }
        }
        else
            break;
        if (p_ctrl->verify)
            cout<<mv<<endl;
        idle_iterations++;
        UpdateIterationCounter();
    }

    TerminateRun();
}


void ColoringRunnerUpHill::Run(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
      UpdateIterationCounter();
      p_ne->BestMove(current_state,mv);
	if (mv.type > 0)
	  p_ne->MakeMove(current_state,mv);
	else
	  break;

        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;

        if (current_state.fitness < initial_ass.fitness)
        {
	  initial_ass = current_state;
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;

        }
        idle_iterations++;

    }

}


void ColoringRunnerUpHill::TestRun(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;

    TabuList tl(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
        UpdateIterationCounter();
        p_ne->BestMove(current_state,mv);

        Coloring test_state(current_state);

        if (!(test_state == current_state))
        {
            cerr<<"before"<<endl;
            exit(1);
        }
        cout<<current_state.fitness<<" "<<current_state.CurrColors()<<endl;
        cout<<mv<<endl;

	if (mv.type > 0)
	  {
	    p_ne->MakeMove(current_state,mv);
	    p_ne->MakeMove(test_state,mv);
	  }
	else
	  break;

        p_sm->updateColoringData(test_state);
        if (!(test_state == current_state))
        {
            exit(1);
        }

        //cout << current_state;

        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;

        if (current_state.fitness < initial_ass.fitness)
        {

            initial_ass = current_state;
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;

        }
        idle_iterations++;
    }
    TerminateRun();
}



void ColoringTabuSearch::Run(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;
    bool tl_created = false;

	if (p_tl == NULL) {
		p_tl = new TabuList();
		tl_created = true;
	}
    p_tl->Init(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
        UpdateIterationCounter();
        p_ne->BestMove(current_state,mv,*p_tl,
                       number_of_iterations,initial_ass.fitness);

        p_tl->UpdateTabuList(current_state,mv,number_of_iterations);
        p_ne->MakeMove(current_state,mv);

        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;
        if (current_state.fitness < initial_ass.fitness)
        {
            initial_ass = current_state;

            //	  ofstream ooo("tmp-sol.col");
            //	  ooo<<current_state;
            //	  ooo.close();
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;
        }

        idle_iterations++;
    }
    TerminateRun();
    if (tl_created) {
    		delete p_tl;
    		p_tl = NULL;
    }
}


void ColoringTabuSearch::TestRun(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;
    bool tl_created = false;

	if (p_tl == NULL) {
		p_tl = new TabuList();
		tl_created = true;
	}
    p_tl->Init(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {

        UpdateIterationCounter();
        p_ne->BestMove(current_state,mv,*p_tl,
                       number_of_iterations,initial_ass.fitness);
        Coloring test_state(current_state);
        //cout << test_state<<endl;
        if (!(test_state == current_state))
        {
            cerr<<"before"<<endl;
            exit(1);
        }
        cout<<mv<<endl;;
        p_tl->UpdateTabuList(current_state,mv,number_of_iterations);
        p_ne->MakeMove(current_state,mv);
        p_ne->MakeMove(test_state,mv);
        //test_state.col[mv.seq[0]] = mv.seq[1];

        //	  current_state.fitness += mv.delta;

        p_sm->updateColoringData(test_state);
        //cout << test_state<<endl;
        if (!(test_state == current_state))
        {
            exit(1);
        }
        //cerr<<mv.delta<<" "<<current_state.fitness<<" "<<number_of_iterations<<endl;
        //cout << current_state;

        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;

        if (current_state.fitness < initial_ass.fitness)
        {
            initial_ass = current_state;
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;

        }
        idle_iterations++;

    }
    TerminateRun();
      if (tl_created) {
    		delete p_tl;
    		p_tl = NULL;
    }
}




void ColoringTabuSearchWithoutAsp::Run(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;

    TabuList tl(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
        UpdateIterationCounter();
        p_ne->BestMove(current_state,mv,tl,
                       number_of_iterations);

        tl.UpdateTabuList(current_state,mv,number_of_iterations);
        p_ne->MakeMove(current_state,mv);

        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;

        if (current_state.fitness < initial_ass.fitness)
        {

            initial_ass = current_state;
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;

        }
        idle_iterations++;

    }

}


void ColoringTabuSearchWithoutAsp::TestRun(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;

    TabuList tl(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
        UpdateIterationCounter();
        p_ne->BestMove(current_state,mv,tl,
                       number_of_iterations);

        Coloring test_state(current_state);

        if (!(test_state == current_state))
        {
            cerr<<"before"<<endl;
            exit(1);
        }
        cout<<current_state.fitness<<" "<<current_state.CurrColors()<<endl;
        cout<<mv<<endl;

        tl.UpdateTabuList(current_state,mv,number_of_iterations);
        p_ne->MakeMove(current_state,mv);

        p_ne->MakeMove(test_state,mv);
        //test_state.col[mv.seq[0]] = mv.seq[1];

        //	  current_state.fitness += mv.delta;

        p_sm->updateColoringData(test_state);
        //cout << test_state<<endl;
        if (!(test_state == current_state))
        {
            exit(1);
        }

        //cout << current_state;

        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;

        if (current_state.fitness < initial_ass.fitness)
        {

            initial_ass = current_state;
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;

        }
        idle_iterations++;
    }
    TerminateRun();
}






