
/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef RUNNERS_H
#define RUNNERS_H


#include <stdio.h>
#include <stdlib.h>
#include <cfloat>
#include <vector>
#include <list>
#include "Problem.h"
#include "Control.h"
#include "Random.h"
#include "Solution.h"
#include "Explorer.h"



class ColoringRunner 
{
public:
  ColoringRunner(ColoringManager* s, ColNeighborhoodExplorer* e,
		 Random *rnd, Control *ctrl, Graph* in);

  ColoringRunner(ColoringManager* s,
		 Random *rnd, Control *ctrl, Graph* in);
		 
  virtual ~ColoringRunner() {};

  void SetCurrentState(const Coloring& s);

  Coloring GetCurrentState();
  fvalue CurrentStateCost();
  Coloring GetBestState();
  fvalue BestStateCost();

  void ComputeCost();

  //bool LowerBoundReached();
  unsigned long NumberOfIterations() const {return number_of_iterations;};
  //unsigned long MaxIteration() const;
  void SetMaxIteration(unsigned long max) {max_iteration = max;}
  void SetMaxIdleIteration(unsigned long max) {max_idle_iteration = max;}

  void initialization_procedure(Coloring&);
  void decreasing_procedure(Coloring &ass);
  
  void Run_procedure(Coloring&);

  virtual void Run(Coloring &initial_ass);
  virtual void TestRun(Coloring &initial_ass);

  void setColExplorer(ColNeighborhoodExplorer* e) { p_ne = e;}
  void update(Coloring &a) { p_sm->updateColoringData(a);}
  ColoringManager *getColoringManager() {return p_sm;};
  ColNeighborhoodExplorer *p_ne;

protected:
  void InitializeRun() ;
  void TerminateRun() {};
  void ComputeMoveCost();
  void UpdateIterationCounter() {number_of_iterations++;} 
  bool MaxIterationExpired() {return number_of_iterations>max_iteration;}
  
  // input
  Graph* p_in; 
  // helpers
  ColoringManager* p_sm;  


  // state data
  Random *rg;
  Control *p_ctrl;

  unsigned long idle_iterations;
  unsigned long number_of_iterations;

  unsigned long max_iteration; 
  unsigned long max_idle_iteration;

  // std::ostream* pos;
};
 


class ColoringRunnerUpHill:public ColoringRunner
{
 public:
  ColoringRunnerUpHill(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~ColoringRunnerUpHill() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
};


class ColoringTabuSearch:public ColoringRunner
{
 public: // FIXME: da riportare anche nel .cpp
  ColoringTabuSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, TabuList *tl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in), p_tl(tl)
    {};
   
  virtual ~ColoringTabuSearch() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
   TabuList *p_tl;
};


//tabu search without aspiration criteria
class ColoringTabuSearchWithoutAsp:public ColoringRunner
{
 public:
  ColoringTabuSearchWithoutAsp(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~ColoringTabuSearchWithoutAsp() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
};





class SimulatedAnnealing:public ColoringRunner
{
 public:
  SimulatedAnnealing(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~SimulatedAnnealing() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
  bool SACriterion(fvalue delta, double Temp);
};




class GuidedLocalSearch:public ColoringRunner
{
 public:
  GuidedLocalSearch(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~GuidedLocalSearch() {};
  
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);

   protected:
};

class GuidedLocalSearchScaled:public ColoringRunner
{
 public:
  GuidedLocalSearchScaled(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~GuidedLocalSearchScaled() {};
  
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);

   protected:
};




class ILS:public ColoringRunner
{
 public:
  ILS(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~ILS() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
  void LocalSearch(Coloring &initial_state, TabuList &tl, 
		   unsigned long int max_idle_iterations_ls);
};

class ILS_adj:public ILS
{
 public:
  ILS_adj(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ILS(s,e,rnd,ctrl,in) {};
   
  virtual ~ILS_adj() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);

};




class MinConflicts:public ColoringRunner
{
 public:
  MinConflicts(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, TabuList *tl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in), p_tl(tl) {};
   
  virtual ~MinConflicts() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
  TabuList* p_tl;
};

class NoveltyRunner:public ColoringRunner
{
 public:
  NoveltyRunner(ColoringManager* s, ColNeighborhoodExplorer* e,
		     Random *rnd, Control *ctrl, Graph* in = NULL):
    ColoringRunner(s,e,rnd,ctrl,in) {};
   
  virtual ~NoveltyRunner() {};
  void Run(Coloring &initial_ass);
  void TestRun(Coloring &initial_ass);
 protected:
};



#endif


