#include <fstream>
#include "Runners.h"

inline bool SimulatedAnnealing::SACriterion(fvalue delta, double Temp)
{
  return (rg->next() < (exp((double)(-delta)/Temp)));
}

void SimulatedAnnealing::TestRun(Coloring &initial_ass)
{
    InitializeRun();
  Coloring current_state(initial_ass);
  ColMove mv;
  long int changes, trials;
  long int cutoff = (long int) ((double) p_ctrl->cutoff * p_in->Vertices() 
				* initial_ass.CurrColors());
  long int size  = (long int) ((double) p_ctrl->sizefactor * p_in->Vertices()
			       * initial_ass.CurrColors());

  p_ctrl->getOutputStream()<<"------ Sa parameters --------------\n";
  p_ctrl->getOutputStream()<<"temperature               \t"<<p_ctrl->temperature<<endl;
  p_ctrl->getOutputStream()<<"cutoff                    \t"<<p_ctrl->cutoff 	   <<endl;
  p_ctrl->getOutputStream()<<"cutoff                    \t"<<cutoff<<endl;
  p_ctrl->getOutputStream()<<"minpercent                \t"<<p_ctrl->minpercent <<endl;
  p_ctrl->getOutputStream()<<"freeze_lim                \t"<<p_ctrl->freeze_lim <<endl;
  p_ctrl->getOutputStream()<<"tempfactor                \t"<<p_ctrl->tempfactor <<endl;
  p_ctrl->getOutputStream()<<"sizefactor                \t"<<p_ctrl->sizefactor <<endl;
  p_ctrl->getOutputStream()<<"size                      \t"<<size<<endl;
  p_ctrl->getOutputStream()<<"-----------------------------------\n";

  
  long int freezecount = 0;
  ofstream os( "tmp/sa_profile.dat" );
  double Temp = p_ctrl->temperature; 
  while ((freezecount < p_ctrl->freeze_lim) //&& (current_state.fitness != 0) 
	 && !(MaxIterationExpired()) && 
	 (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
      
      changes = trials = 0;
      bool impr = false;
      while ((trials < size) && (changes < cutoff) && !(MaxIterationExpired()) && 
	     (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
	{
	  trials++;
	  UpdateIterationCounter();
	  p_ne->NextMove(current_state,mv,number_of_iterations);

	  if (mv.delta <= 0)
	    {
	      p_ne->MakeMove(current_state,mv);
	      changes++;
	      //cout<<current_state.CurrColors()<<endl;
	      //  	  if (current_state.fitness != 0)
	      //  	    current_state.feasible = false;
	      //  	  else
	      //  	    current_state.feasible = true;
	      if (current_state.fitness < initial_ass.fitness)
		{
		  
		  p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
		  idle_iterations = 0;
		   impr = true; 
		if (current_state.CurrColors() <= initial_ass.CurrColors())
		 {
		   initial_ass = current_state;
		 }
		}
	    }
	  else if (SACriterion(mv.delta,Temp))
	    {
	      p_ne->MakeMove(current_state,mv);
	      changes++;
	    }
	  idle_iterations++;

	}
      
      Temp = (double) p_ctrl->tempfactor * Temp;
           
      if (impr)
	freezecount = 0;
      if (((double) changes/trials) < p_ctrl->minpercent)
	freezecount++;
      (os)<<number_of_iterations<<" "<<Temp<<" "<<current_state.fitness<<" "
	  <<current_state.CurrColors()<<" "<<((double) changes/trials) <<endl; 

    }
  os.close();
      
  TerminateRun();
}


void SimulatedAnnealing::Run(Coloring &initial_ass)
{
  InitializeRun();
  Coloring current_state(initial_ass);
  ColMove mv;
  long int changes, trials;
long int cutoff = (long int) ((double) p_ctrl->cutoff * p_in->Vertices()
			      * initial_ass.CurrColors());
  long int size  = (long int) ((double) p_ctrl->sizefactor * p_in->Vertices()
			       * initial_ass.CurrColors());

  p_ctrl->getOutputStream()<<"------ Sa parameters --------------\n";
  p_ctrl->getOutputStream()<<"temperature               \t"<<p_ctrl->temperature<<endl;
  p_ctrl->getOutputStream()<<"cutoff                    \t"<<p_ctrl->cutoff 	   <<endl;
  p_ctrl->getOutputStream()<<"cutoff                    \t"<<cutoff<<endl;
  p_ctrl->getOutputStream()<<"minpercent                \t"<<p_ctrl->minpercent <<endl;
  p_ctrl->getOutputStream()<<"freeze_lim                \t"<<p_ctrl->freeze_lim <<endl;
  p_ctrl->getOutputStream()<<"tempfactor                \t"<<p_ctrl->tempfactor <<endl;
  p_ctrl->getOutputStream()<<"sizefactor                \t"<<p_ctrl->sizefactor <<endl;
  p_ctrl->getOutputStream()<<"size                      \t"<<size<<endl;
  p_ctrl->getOutputStream()<<"-----------------------------------\n";

  
  long int freezecount = 0;
  double Temp = p_ctrl->temperature; 
  while ((freezecount < p_ctrl->freeze_lim) //&& (current_state.fitness != 0) 
	 && !(MaxIterationExpired()) && 
	 (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
      
      changes = trials = 0;
      bool impr = false;      
      while ((trials < size) && (changes < cutoff) && !(MaxIterationExpired()) && 
	     (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
	{
	  trials++;

	  UpdateIterationCounter();
	  p_ne->NextMove(current_state,mv,number_of_iterations);

	  if (mv.delta <= 0)
	    {
	      p_ne->MakeMove(current_state,mv);
	      changes++;
	  
	      //  	  if (current_state.fitness != 0)
	      //  	    current_state.feasible = false;
	      //  	  else
	      //  	    current_state.feasible = true;
	      if (current_state.fitness < initial_ass.fitness)
		{
		  impr = true; 
		  p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
		  idle_iterations = 0;
		  
		if (current_state.CurrColors() <= initial_ass.CurrColors())
		 {
		   initial_ass = current_state;
		 }
		}
	    }
	  else if (SACriterion(mv.delta,Temp))
	    {
	      p_ne->MakeMove(current_state,mv);
	      changes++;
	    }
	  idle_iterations++;
	}
      
      Temp = (double) p_ctrl->tempfactor * Temp;
      if (impr)
	freezecount = 0;
      if (((double) changes/trials) < p_ctrl->minpercent)
	freezecount++;

    }
      
  TerminateRun();
}
