#include "Runners.h"
#include "Explorer.h"


//for MinConflicts!!!!
void ColOneExMinConfExplorer::BestMove(const Coloring& coloring, ColMove& rc, 
				       TabuList &t_l, unsigned long iter,
				       fvalue best_so_far) 
 {    
   //const Graph &g = *p_in; // g is an alias for the input Graph
   fvalue min = MAX_FVALUE;
   VertexSet::iterator pt;  
   vector < pair<Vertex, Color> > candidates;
   fvalue delta;

   rc.seq.clear();
   rc.seq.resize(2);
   rc.type= 1;

   //select one node in conflict randomly
   int max = rg->ran0N(coloring.conflicts_set.size()-1);
   pt = coloring.conflicts_set.begin();
   for (int i = 0; 
	i < max; i++, pt++);

   //select for the node the best color
   for (Color j=0 ; j < coloring.CurrColors() ; j++) {
     //if (coloring.col[*pt]==j)
       {
	 if (coloring.col[*pt]==j)
	   delta = 0;
	 else
	   delta = coloring.Delta(*pt,j);
	 if (delta == min) {
	   if ((!t_l.isTabu(*pt,j,iter)) 
	       || (coloring.fitness + delta < best_so_far))
	     {
	       pair<Vertex, Color> initPair;
	       initPair.first = *pt ;
	       initPair.second = j ;
	       candidates.push_back(initPair);
	     }
	 }
	 else if (delta < min)  {
	   if ((!t_l.isTabu(*pt,j,iter)) 
	       || (coloring.fitness + delta < best_so_far))
	     {
	       candidates.clear();
	       pair<Vertex, Color> initPair;
	       initPair.first = *pt ;
	       initPair.second = j ;
	       candidates.push_back(initPair);
	       min = delta;
	     }
	 }
       }
   }
   
   //select randomly among the best colors for the node
   if (candidates.size()>0)
      {
       int rnd = rg->ran0N(candidates.size()-1);
       rc.seq[0] = candidates[rnd].first;
       rc.seq[1] = candidates[rnd].second;
       rc.delta = min;
       //rc.seq.push_back(coloring.col[rc.seq[0]]);
      }
   else
     {
       // do {
	 rc.seq[0]=rg->ran0N(p_in->Vertices()-1);
	 rc.seq[1]=rg->ran0N(coloring.CurrColors()-1);
	 //} while (rc.seq[1]==coloring.col[rc.seq[0]]);
	 if (rc.seq[1]== (Vertex) coloring.col[rc.seq[0]])
	   rc.delta = 0;
	 else
	   rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
     }
 }
 



//for Novelty+!!!! revised version after Thomas comments
void ColOneExNoveltyExplorer::BestMove(const Coloring& coloring, ColMove& rc, 
 				       TabuList &t_l, unsigned long iter,
				       fvalue best_so_far) 
{    
  //const Graph &g = *p_in; // g is an alias for the input Graph
  fvalue min = MAX_FVALUE;
  unsigned long int min_age = LONG_MAX;
  fvalue second_min = MAX_FVALUE;
  unsigned long rnd;

  VertexSet::iterator pt;  
  vector < pair<Vertex, Color> > candidates;
  vector < pair<Vertex, Color> > second_candidates;
  fvalue delta;

  rc.seq.clear();
  rc.seq.resize(2);
  rc.type= 1;

 
  double wp = 0.0;
  double prob = t_l.prob;
   
  if (rg->next() < (wp))
    {
      do {
	rc.seq[0]=rg->ran0N(p_in->Vertices()-1);
	rc.seq[1]=rg->ran0N(coloring.CurrColors()-1);
      } while (rc.seq[1]==(Vertex) coloring.col[rc.seq[0]]);
      rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
      //cout<<"*";
      return;
    }
  else
    {
      int max = rg->ran0N(coloring.conflicts_set.size()-1);
      pt = coloring.conflicts_set.begin();
      for (int i = 0; 
	   i < max; i++, pt++);
      for (Color j=0 ; j < coloring.CurrColors() ; j++) {
	if ( j != coloring.col[*pt])
	  {
	    if (t_l.IterTabu(*pt,j) < min_age)
	      min_age = t_l.IterTabu(*pt,j);
	   
	    delta = coloring.Delta(*pt,j);
	    //cout<<"."<<delta;
	    if (delta <= min)
	      {
		if (delta < min)
		  {
		    second_candidates = candidates;
		    second_min = min;
		    candidates.clear();
		    candidates.push_back(pair<Vertex, Color>(*pt, j));
		  }
		else
		  {
		    candidates.push_back(pair<Vertex, Color>(*pt, j));
		  }
		
		min = delta;
	      }
	  else if (delta <= second_min)
	    {
	  	if (delta < second_min)
	  	  {
	  	    second_candidates.clear();
	  	    second_candidates.push_back(pair<Vertex, Color>(*pt, j));
	  	  }
	  	else
	  	  {
	  	    second_candidates.push_back(pair<Vertex, Color>(*pt, j));
	  	  }
	  	second_min = delta;
	    }
	  }	 
      }
    }


  //cout<<">"<<min<<" "<<candidates.size()<<" "
  //  <<second_min<<" "<<second_candidates.size()<<endl;
  if (candidates.size() == 1)
    {
      if (t_l.IterTabu(candidates[0].first,candidates[0].second) == min_age)
	{
	  if (rg->next()<1-prob)
	    {
	      rc.seq[0] = candidates[0].first;
	      rc.seq[1] = candidates[0].second;
	      rc.delta = min;
	      //cout<<"+"<<endl;
	      //cout<<rc<<endl;
	      return;
	    }
	  else
	    {
	      //cout<<second_candidates.size()<<endl;
	      int rnd = rg->ran0N(second_candidates.size()-1);
	      rc.seq[0] = second_candidates[rnd].first;
	      rc.seq[1] = second_candidates[rnd].second;
	      rc.delta = second_min;//coloring.Delta(rc.seq[0],rc.seq[1]);
	      //cout<<"-"<<endl;
	      return;
	    }
	}
      else
	{
	  rc.seq[0] = candidates[0].first;
	  rc.seq[1] = candidates[0].second;
	  rc.delta = min;
	  //cout<<"++"<<endl;
	  return;
	}
    }
  else if (candidates.size()>0)
    {
      do 
	{
	  rnd = rg->ran0N(candidates.size()-1);
	} while ((t_l.IterTabu(candidates[rnd].first,candidates[rnd].second) 
		  == min_age) && (rg->next()<prob));
      rc.seq[0] = candidates[rnd].first;
      rc.seq[1] = candidates[rnd].second;
      rc.delta = min;
      //cout<<"--"<<endl;
      return;
      //rc.seq.push_back(coloring.col[rc.seq[0]]);
    }
  else
    {
      cerr<<"Something strange in Novelty+ BestMove selection\n";
      do {
	rc.seq[0]=rg->ran0N(p_in->Vertices()-1);
	rc.seq[1]=rg->ran0N(coloring.CurrColors()-1);
      } while (rc.seq[1]== (Vertex) coloring.col[rc.seq[0]]);
      rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
    }
}


/* Version used for the paper
//for Novelty+!!!!
void ColOneExNoveltyExplorer::BestMove(const Coloring& coloring, ColMove& rc, 
 				       TabuList &t_l, unsigned long iter,
				       fvalue best_so_far) 
{    
  //const Graph &g = *p_in; // g is an alias for the input Graph
  fvalue min = MAX_FVALUE;
  unsigned long int min_age = LONG_MAX;
  unsigned long int max_age = 0;
  unsigned long rnd;

  VertexSet::iterator pt;  
  vector < pair<Vertex, Color> > candidates;
  vector < pair<Vertex, Color> > oldest_candidates;
  fvalue delta;

  rc.seq.clear();
  rc.seq.resize(2);
  rc.type= 1;

 
  double wp = 0.01;
  double prob = t_l.prob;
   
  if (rg->next() < (wp))
    {
      do {
	rc.seq[0]=rg->ran0N(p_in->Vertices()-1);
	rc.seq[1]=rg->ran0N(coloring.CurrColors()-1);
      } while (rc.seq[1]==coloring.col[rc.seq[0]]);
      rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
      //cout<<"*";
      return;
    }
  else
    {
      int max = rg->ran0N(coloring.conflicts_set.size()-1);
      pt = coloring.conflicts_set.begin();
      for (int i = 0; 
	   i < max; i++, pt++);
      for (Color j=0 ; j < coloring.CurrColors() ; j++) {
	if ( j != coloring.col[*pt])
	  {
	    if (t_l.IterTabu(*pt,j) < min_age)
	      min_age = t_l.IterTabu(*pt,j);

	    if (t_l.IterTabu(*pt,j) == max_age)
	      {
		max_age = t_l.IterTabu(*pt,j);
		oldest_candidates.push_back(pair<Vertex, Color>(*pt,j));
	      }
	    else if (t_l.IterTabu(*pt,j) > max_age)
	      {
		oldest_candidates.clear();
		max_age = t_l.IterTabu(*pt,j);
		oldest_candidates.push_back(pair<Vertex, Color>(*pt,j));
	      }
	   
	    delta = coloring.Delta(*pt,j);
	    if (delta == min) {
	      pair<Vertex, Color> initPair;
	      initPair.first = *pt ;
	      initPair.second = j ;
	      candidates.push_back(initPair);
	    }
	    if (delta < min)  {
	      candidates.clear();
	      pair<Vertex, Color> initPair;
	      initPair.first = *pt ;
	      initPair.second = j ;
	      candidates.push_back(initPair);
	      min = delta;
	    }
	  }	 
      }
    }

  if (candidates.size() == 1)
    {
      if (t_l.IterTabu(candidates[0].first,candidates[0].second) == min_age)
	{
	  if (rg->next()<1-prob)
	    {
	      rc.seq[0] = candidates[0].first;
	      rc.seq[1] = candidates[0].second;
	      rc.delta = min;
	      //cout<<"+";
	      return;
	    }
	  else
	    {
	      int rnd = rg->ran0N(candidates.size()-1);
	      rc.seq[0] = oldest_candidates[rnd].first;
	      rc.seq[1] = oldest_candidates[rnd].second;
	      rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
	      //cout<<"-";
	      return;
	    }
	}
      else
	{
	  rc.seq[0] = candidates[0].first;
	  rc.seq[1] = candidates[0].second;
	  rc.delta = min;
	  //cout<<"+";
	  return;
	}
    }
  else if (candidates.size()>0)
    {
      do 
	{
	  rnd = rg->ran0N(candidates.size()-1);
	} while ((t_l.IterTabu(candidates[rnd].first,candidates[rnd].second) 
		  == min_age) && (rg->next()<prob));
      rc.seq[0] = candidates[rnd].first;
      rc.seq[1] = candidates[rnd].second;
      rc.delta = min;
      //cout<<"-";
      return;
      //rc.seq.push_back(coloring.col[rc.seq[0]]);
    }
  else
    {
      cerr<<"Somthing strange in Novelty+ BestMove selection\n";
      do {
	rc.seq[0]=rg->ran0N(p_in->Vertices()-1);
	rc.seq[1]=rg->ran0N(coloring.CurrColors()-1);
      } while (rc.seq[1]==coloring.col[rc.seq[0]]);
      rc.delta = coloring.Delta(rc.seq[0],rc.seq[1]);
    }
}
*/

void MinConflicts::Run(Coloring &initial_ass)
{
    InitializeRun();
    Coloring current_state(initial_ass);
    ColMove mv;
    bool tl_created = false;

    if (p_tl == NULL) {
        p_tl = new TabuListMinConf(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);
        tl_created = true;
    } else
      p_tl->Init(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

    while ((current_state.fitness != 0) && !(MaxIterationExpired()) &&
            (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
        UpdateIterationCounter();
        p_ne->BestMove(current_state,mv,*p_tl,
                       number_of_iterations,initial_ass.fitness);
        //        Coloring test_state(current_state);

        //        if (!(test_state == current_state))
        //  	{
        //  	  cerr<<"before"<<endl;
        //  	  exit(1);
        //  	}
        p_tl->UpdateTabuList(current_state,mv,number_of_iterations);
        p_ne->MakeMove(current_state,mv);


        //      test_state.col[mv.seq[0]] = mv.seq[1];

        //current_state.fitness += mv.delta;

        //        p_sm->updateColoringData(test_state);

        //        cout<<mv<<endl;
        //        if (!(test_state == current_state))
        //  	{
        //  	  exit(1);
        //  	}

        //        cout << current_state;
        if (current_state.fitness != 0)
            current_state.feasible = false;
        else
            current_state.feasible = true;
        if (current_state.fitness < initial_ass.fitness)
        {
            initial_ass = current_state;
            p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
            idle_iterations = 0;
        }
        idle_iterations++;

    }
    //  cout<<(current_state.fitness != 0)<<" "<<!(MaxIterationExpired())<<" "<<
    //      (idle_iterations < max_idle_iteration)<<" "<<p_ctrl->timeLeft()<<endl;
    //    cout<<NumberOfIterations()<<" "<<max_iteration<<endl;
    TerminateRun();
    if (tl_created) {
        delete p_tl;
        p_tl = NULL;
    }
}



void MinConflicts::TestRun(Coloring &initial_ass)
{
  InitializeRun();
  Coloring current_state(initial_ass);
  ColMove mv;
  
  TabuListMinConf tl(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);

  while ((current_state.fitness != 0) && !(MaxIterationExpired()) && 
	 (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {

      UpdateIterationCounter();
      p_ne->BestMove(current_state,mv,tl,
		     number_of_iterations,initial_ass.fitness);
      Coloring test_state(current_state);
      //cout << test_state<<endl;
        if (!(test_state == current_state))
  	{
  	  cerr<<"before"<<endl;
  	  exit(1);
  	}
	cout<<mv<<endl;;
       tl.UpdateTabuList(current_state,mv,number_of_iterations); 
       p_ne->MakeMove(current_state,mv);
       p_ne->MakeMove(test_state,mv);
	  //test_state.col[mv.seq[0]] = mv.seq[1];

	  //	  current_state.fitness += mv.delta;	  
	  
	  p_sm->updateColoringData(test_state);
	  //cout << test_state<<endl;
	    if (!(test_state == current_state))
  	    {
  	      exit(1);
  	    }
	    //cerr<<mv.delta<<" "<<current_state.fitness<<" "<<number_of_iterations<<endl;
	    //cout << current_state;	 
	  
	    if (current_state.fitness != 0)
	      current_state.feasible = false;
	    else
	      current_state.feasible = true;
	    
	    if (current_state.fitness < initial_ass.fitness)
	      {
		initial_ass = current_state;
		p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
		idle_iterations = 0;

	      }
      idle_iterations++;
    
    }
  TerminateRun();
}

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------


void NoveltyRunner::Run(Coloring &initial_ass)
{
  InitializeRun();
  Coloring current_state(initial_ass);
  ColMove mv;
  
  HistoryNovelty tl(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);
  tl.prob = p_ctrl->prob;
  while ((current_state.fitness != 0) && !(MaxIterationExpired()) && 
	 (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {
      UpdateIterationCounter();
      p_ne->BestMove(current_state,mv,tl,
		     number_of_iterations,initial_ass.fitness);

      tl.UpdateTabuList(current_state,mv,number_of_iterations);
      p_ne->MakeMove(current_state,mv);
  
      if (current_state.fitness != 0)
	current_state.feasible = false;
      else
	current_state.feasible = true;
      if (current_state.fitness < initial_ass.fitness)
	{
	  initial_ass = current_state;
	  p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
	  idle_iterations = 0;
	}
      idle_iterations++;
     
    }
  TerminateRun();
}


void NoveltyRunner::TestRun(Coloring &initial_ass)
{
  InitializeRun();
  Coloring current_state(initial_ass);
  ColMove mv;
  
  HistoryNovelty tl(p_in->Vertices(),current_state.CurrColors(),rg,p_ctrl);
  tl.prob = p_ctrl->prob;
  while ((current_state.fitness != 0) && !(MaxIterationExpired()) && 
	 (idle_iterations < max_idle_iteration) && p_ctrl->timeLeft())
    {

      UpdateIterationCounter();
      p_ne->BestMove(current_state,mv,tl,
		     number_of_iterations,initial_ass.fitness);
      Coloring test_state(current_state);
      //cout << test_state<<endl;
        if (!(test_state == current_state))
  	{
  	  cerr<<"before"<<endl;
  	  exit(1);
  	}
	cout<<mv<<endl;;
       tl.UpdateTabuList(current_state,mv,number_of_iterations); 
       p_ne->MakeMove(current_state,mv);
       p_ne->MakeMove(test_state,mv);
	  //test_state.col[mv.seq[0]] = mv.seq[1];

	  //	  current_state.fitness += mv.delta;	  
	  
	  p_sm->updateColoringData(test_state);
	  //cout << test_state<<endl;
	    if (!(test_state == current_state))
  	    {
  	      exit(1);
  	    }
	    //cerr<<mv.delta<<" "<<current_state.fitness<<" "<<number_of_iterations<<endl;
	    //cout << current_state;	 
	  
	    if (current_state.fitness != 0)
	      current_state.feasible = false;
	    else
	      current_state.feasible = true;
	    
	    if (current_state.fitness < initial_ass.fitness)
	      {
		initial_ass = current_state;
		p_ctrl->setCurrentCost(current_state.fitness,current_state.CurrColors(),number_of_iterations);
		idle_iterations = 0;

	      }
      idle_iterations++;
    
    }
  TerminateRun();
}

