/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <algorithm>
#include <cmath>

#include <iostream>
#include <fstream>
#include "Solution.h"
//#include <string>
//#include <sstream>

using namespace std;
//using std::ostringstream; using std::string;


std::ostream& operator<<(std::ostream& os, const Coloring& coloring)
{
  const Graph& g = *coloring.p_in; // g is an alias for the input Graph
  
  if (false)
    {
      int viol = 0;
      
      // write the coloring
      for (Vertex v = 0; v < g.Vertices(); v++)
	{
	  os << v << "[" << coloring.col[v] << "], ";
	  for (Vertex w = v + 1; w < g.Vertices(); w++)
	    if (g.Edge(v,w) && coloring.col[v] == coloring.col[w])
	      viol++;
	}
      
      os << endl <<" Colors: " << coloring.k
	 << " Conflicting edges: " << viol
	 << ", Conflicting nodes: " << coloring.conflicts_set.size()
	 << " Fitness: "<< coloring.fitness<< endl;
      //copy(coloring.conflicts.begin(), coloring.conflicts.end(), 
      //ostream_iterator<Vertex>(os, " "));
      //copy(coloring.adj_x_color.begin(), coloring.adj_x_color.end(), ostream_iterator<Vertex>(os, " "));
    }
  else
    for (Vertex v = 0; v < g.Vertices(); v++)
      {
	os<<coloring.col[v]+1<<endl;
      }
  return os;
}

std::ostream& operator<<(std::ostream& os, const ColMove& mv)
{
  os <<"Type: " << mv.type;
  if (mv.type != 1)
    {
      os <<" Length: "<<mv.seq.size() << " Seq: ";
      for(vector<Vertex>::const_iterator pt = mv.seq.begin();
	  pt != mv.seq.end(); pt++)
	{
	  os<<*pt<<" ";
	}
    }
  else
    {
      os <<" Length: "<<mv.seq.size()-1 << " Seq: ";
      for(int pt = 0; pt < 2; pt++)
	{
	  os<<mv.seq[pt]<<" ";
	}
    }
  os <<"Cost: "<<mv.delta;
      
  return os;
}


Coloring::Coloring(Random* rnd, Graph* pin) 
{
  rg = rnd;
  // initialize the delta map
  p_in = pin;
  col.resize(pin->Vertices()); 
   // initialize the delta map
  
//    impGraph = new fvalue *[p_in->Vertices()];
//    for(Vertex i=0; i<p_in->Vertices(); i++)
//      impGraph[i] = new fvalue [p_in->Vertices()];
  
  // finally clear the conflict set
  conflicts_set.clear(); 
  conflicts.resize(pin->Vertices(),0);
  conflict_edges.clear();
  k = INT_MAX;
  feasible = false;
}


Coloring& Coloring::operator=(const Coloring &a)
{
  if (this == &a) return *this;
  /* if (this == NULL)
    {
      try {
	this = new Coloring(a.rg, a.p_in)
      } catch (bad alloc xa)
	{
	  cerr<<"Bad alloc in copy constructor Solution"<<endl;
	}
	}*/
  p_in = a.p_in;
  rg = a.rg;
  
  k = a.k;
  fitness = a.fitness;

  //col.resize(p_in->Vertices()); 
  //  col.clear();
  col = a.col;
  // delta = a.delta;

   
 //   for(Vertex i=0; i<p_in->Vertices(); i++)
//      {
//        for (Vertex j=0; j<p_in->Vertices(); j++)
//  	impGraph[i][j] = a.impGraph[i][j];
//      }
 
  //adj_x_color.resize(k*p_in->Vertices());
  adj_x_color = a.adj_x_color;
  conflicts = a.conflicts;
  conflicts_set = a.conflicts_set;
  conflict_edges = a.conflict_edges;

  partitions = a.partitions;
  feasible = a.feasible;
  loop_on_partition = a.loop_on_partition;
 

    pen_delta = a.pen_delta;
  pen_adj_x_color = a.pen_adj_x_color;
  return *this;
}


Coloring::Coloring(const Coloring &a)
{
  p_in = a.p_in;
  rg = a.rg;
  
  k = a.k;
  fitness = a.fitness;
  //col.resize(p_in->Vertices()); 
  col = a.col;
  //delta = a.delta;
  //adj_x_color.resize(k*p_in->Vertices());
  adj_x_color = a.adj_x_color;
  
  //     impGraph = new fvalue *[p_in->Vertices()];
  //    for(Vertex i=0; i<p_in->Vertices(); i++)
  //      {
  //        impGraph[i] = new fvalue [p_in->Vertices()];
  //        for (Vertex j=0; j<p_in->Vertices(); j++)
  //  	impGraph[i][j] = a.impGraph[i][j];
  //  	}
  
  conflicts = a.conflicts;
  conflict_edges = a.conflict_edges;
  conflicts_set = a.conflicts_set;
  partitions = a.partitions;
  loop_on_partition = a.loop_on_partition;
  feasible = a.feasible;
  
    pen_delta = a.pen_delta;
  pen_adj_x_color = a.pen_adj_x_color;
  
}

Coloring::~Coloring()
{
  //  if (impGraph)
//      {
//        for(Vertex i=0; i < p_in->Vertices(); i++)
//  	delete [] impGraph[i];
//        delete [] impGraph;    
//      }
}




bool Coloring::operator==(const Coloring &a)
{
  
  if (p_in != a.p_in)
    {
      cerr << "input data!"<<endl;
      return false;
    }
  if (rg != a.rg)
    { 
      cerr << "random data!"<<endl;
      return false;
    }
  
  if (k != a.k)
    {
      cerr << "k!"<<endl;
      return false;
    }
  if (conflicts != a.conflicts)
    {
      cerr << "conflicts data! "<<endl;
      return false;
    }
 
  if (conflict_edges != a.conflict_edges)
    {
      cerr << "conflicting edges data! "<<endl;
      set<pair<Vertex, Vertex> >::iterator pt;
      for (pt = conflict_edges.begin();
	   pt != conflict_edges.end(); pt++)
      {
	  cout<<"["<<(*pt).first<<","<<(*pt).second<<"] ";
      }
      cout<<endl;
      for (pt = a.conflict_edges.begin();
	   pt != a.conflict_edges.end(); pt++)
      {
	  cout<<"["<<(*pt).first<<","<<(*pt).second<<"] ";
      }
      cout<<endl;
      return false;
      }
 
  if (fitness != a.fitness)
    {
      cerr << "fitness data!"<<endl;
      cerr << fitness << " " << a.fitness<< endl;
      return false;
    }

 
  if (feasible != a.feasible)
    {
      cerr << "feasible data!";
      
      return false;
    }
 
  if (loop_on_partition != a.loop_on_partition)
    {
      cerr << "loop_on_partition data!"<<endl;
      return false;
    }
 
  if (col != a.col)
    {
      cerr << "col data!"<<endl;
      return false;
    }
 //   if (delta != a.delta)
//      {
//        for (Vertex i=0; i<p_in->Vertices(); i++)
//  	for (Color j=0; j<a.k; j++)
//         if (Delta(i,j) != a.Delta(i,j))
//  	 {
//  	   cerr << "delta data! "<<i<<" "<<j<<" "<<Delta(i,j)
//  		<<" "<<a.Delta(i,j)<<endl;
//  	   return false;
//  	 }
  //   }
 
  if (adj_x_color != a.adj_x_color)
    {
      cerr << "adj_color data!"<<endl;
      return false;
    }
  
 //    for (Vertex i=0; i<p_in->Vertices(); i++)
//       for (Vertex j=0; j<p_in->Vertices(); j++)
//         if (impGraph[i][j] != a.impGraph[i][j])
//  	 {
//  	   cerr << "ImpGraph data! "<<i<<" "<<j<<" "<<impGraph[i][j]
//  		<<" "<<a.impGraph[i][j]<<endl;
//  	   return false;
//  	 }
 
  for (Color i=0; i<a.k; i++)
    if (partitions[i] != a.partitions[i])
      {
	cerr << "partitions data!"<<endl;
	return false;
      }
	
  if (conflicts_set != a.conflicts_set)
    {
      cerr << "conflicts set data!"<<endl;
      return false;
    }
  



  /*  for (Vertex i = 0; i< p_in->Vertices(); i++)
    {
      for (Color j = 0; j< a.CurrColors(); j++)
	{
	  if (pen_adj_x_color[i*k+j] - a.pen_adj_x_color[i*k+j] > 0.00001 )
	    {
	      printf("%.5f %.5f\n",pen_adj_x_color[i*k+j], a.pen_adj_x_color[i*k+j]);
	      cerr << "pen_adj_x_color data!"<<endl;
	      return false;
	    } 
	}
	}*/

//   for (Vertex i = 0; i< p_in->Vertices(); i++)
//     for (Color j = 0; j< k; j++)
//       if (pen_delta[i][j] - a.pen_delta[i][j] >  0.00001)
// 	{
// 	  printf("%.5f %.5f\n",pen_delta[i][j],a.pen_delta[i][j]);
// 	  cerr << "pen_delta data!"<<endl;
// 	  return false;
// 	}

  
//    for (Vertex i = 0; i< p_in->Vertices(); i++)
//      for (Color j = 0; j< k; j++)
//        if (col[i]!=j)
// 	 if (PenDelta(i,j) - a.pen_delta[i][j] >  0.00001)
// 	   {
// 	     printf("%d %d %.5f %.5f\n",i,j, PenDelta(i,j), pen_delta[i][j]);
// 	     return false;
// 	   }
 
  return true;
}



void Coloring::SetInput(Graph* pin) 
{
  p_in = pin; 
  col.resize(pin->Vertices()); 
   // initialize the delta map
  
//    for(Vertex i=0; i<pin->Vertices(); i++)
//      delete [] impGraph[i];
//    delete impGraph;

//    impGraph = new fvalue *[pin->Vertices()];
//    for(Vertex i=0; i<pin->Vertices(); i++)
//      impGraph[i] = new fvalue [pin->Vertices()];
  
  // finally clear the conflict set
  conflicts_set.clear(); 
  conflict_edges.clear(); 
  conflicts.resize(pin->Vertices(),0);
}

void Coloring::SetColors(Color num_col)
{
  this->k = num_col;
  //partitions.clear();
  //partitions.resize(k);
  //delta.clear();
  //delta.resize(k*p_in->Vertices());
  adj_x_color.clear();
  adj_x_color.resize(k*p_in->Vertices());
}

// Coloring Methods
  
//  inline fvalue& Coloring::Delta(const Vertex& v, const Color& c) 
//  {
//    return delta[v*k+c];
//  }

/*
const fvalue &Coloring::Delta(const Vertex& v, const Color& c) const 
{
  return delta[v*k+c];
}*/

Vertex Coloring::choose_conflict() const
{
  VertexSet::const_iterator ci = conflicts_set.begin();
  unsigned int n = rg->ran0N(conflicts_set.size()-1);
  for (unsigned int i = 0; i < n; i++, ci++)
    ;
  return *ci; 
}


ColMove::ColMove()
{
  seq.clear();
  type = 0;
  delta = 0;
}

ColMove::ColMove(const ColMove& a)
{
  seq = a.seq;
  type = a.type;
  delta = a.delta;
}

ColMove& ColMove::operator=(const ColMove& a)
{
  seq = a.seq;
  type = a.type;
  delta = a.delta;
  return *this;
}


bool ColMove::operator==(const ColMove &a)
{ 
  if (type != a.type)
    return false;
  if (seq != a.seq)
    return false;
  if (delta != a.delta)
    return false;
  return true;
}


ColoringManager::ColoringManager(Random *rnd, Graph* pin = NULL, Control *ctrl = NULL)
{
  p_in = pin;
  rg = rnd;
  p_ctrl=ctrl;
}

 
void ColoringManager::updateColoringData(Coloring &ass) const 
{
  const Graph &g = *p_in; // g is an alias for the input Graph

  // first build the group list
  // and the edge list
 
 /*sceglie il ciclo piu conveniente su colore o nodi adiacenti*/  
  //approx_col = 2*N/ass.k;
  //approx_adj = N*(p_in->AvDensity());
 
  if ( (float)(g.Vertices()/ass.CurrColors()) < (float)(g.Vertices()*(p_in->Density())) )
    ass.loop_on_partition = true;
  else
    ass.loop_on_partition = false;
  
  ass.fitness = 0;
  ass.conflicts.resize(g.Vertices());
  ass.conflicts_set.clear();
  ass.conflict_edges.clear();
  ass.partitions.resize(ass.CurrColors());
  for (Color c = 0; c < ass.CurrColors(); c++)
    {
      ass.partitions[c].clear();
    }
  // rebuild the conflicting list
  // and the delta values
  // Update the improvement graph
  // this is of the order N^2!!
  ass.adj_x_color.clear();
  ass.adj_x_color.resize(ass.CurrColors()*g.Vertices(),0);
  for (Vertex i=0; i<g.Vertices(); i++)
    {
      ass.conflicts[i]=0;
      ass.partitions[ass.col[i]].insert(i);
      VertexSet::const_iterator pt;
      for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++)
	{
	  ass.AdjColor(i,ass.col[*pt])++;
	}
    }
  for (Vertex i=0; i<g.Vertices(); i++)
  {
//  	for (Color c = 0; c < ass.k; c++)
//  	{
//  	  if (ass.col[i]!=c)
//  	    ass.Delta(i, c)= ass.AdjColor(i,c) - ass.AdjColor(i,ass.col[i]);
//  	  else
//  	    ass.Delta(i,c) = MAX_FVALUE;
//  	}
    for (Vertex j=i; j<g.Vertices(); j++)
	{
	   if (g.Edge(i,j) && (ass.col[i] == ass.col[j])) 
	    {
	      ass.fitness++;
	      ass.conflicts_set.insert(i);
	      ass.conflicts[i]++;
	      ass.conflicts_set.insert(j);
	      ass.conflicts[j]++;
	      ass.conflict_edges.insert(pair<Vertex,Vertex> (i,j));
	    }
//  	  if (ass.col[i]!=ass.col[j])
//  	    {
//  	      ass.impGraph[i][j]=0;
//  	      ass.impGraph[j][i]=0;
	      
//  	      ass.impGraph[i][j] = ass.AdjColor(i,ass.col[j]) 
//  		- ass.AdjColor(j,ass.col[j]) - g.Edge(i,j);
//  	      ass.impGraph[j][i] = ass.AdjColor(j,ass.col[i]) 
//  		- ass.AdjColor(i,ass.col[i]) - g.Edge(i,j);
//  	    }
//  	  else 
//  	    {
//  	      ass.impGraph[i][j] = MAX_FVALUE;
//  	      ass.impGraph[j][i] = MAX_FVALUE;
//  	    }

	} 
    }
  if (ass.fitness != 0)
    ass.feasible = false;
  else
    ass.feasible = true;
}



void ColManPenaltyFunction::updateColoringData(Coloring &ass) const 
{
  if ( (float)(p_in->Vertices()/ass.CurrColors()) < 
       (float)(p_in->Vertices()*(p_in->Density())) )
    ass.loop_on_partition = true;
  else
    ass.loop_on_partition = false;
  
  ass.fitness = 0;
  ass.conflicts.resize(p_in->Vertices());
  ass.conflicts_set.clear();
  ass.partitions.resize(ass.CurrColors());
  for (Color c = 0; c < ass.CurrColors(); c++)
    {
      ass.partitions[c].clear();
    }
  // rebuild the conflicting list
  // and the delta values
  // Update the improvement graph
  // this is of the order N^2!!
  ass.adj_x_color.clear();
  //ass.adj_x_color.resize(ass.CurrColors()*g.Vertices(),0);
  for (Vertex i=0; i<p_in->Vertices(); i++)
    {
      ass.conflicts[i]=0;
      ass.partitions[ass.col[i]].insert(i);
    }

  for (Color i=0; i < ass.CurrColors(); i++)
    {
      ass.fitness -= ass.partitions[i].size() * ass.partitions[i].size();
    }

  //ass.feasible = true;
}


fvalue ColoringManager::computeFitness(Coloring &ass)
  {
    ass.fitness = 0;
    for (Vertex v = 0; v < p_in->Vertices(); v++)
    {
      for (Vertex w = v+1; w < p_in->Vertices(); w++)
	{
	  if (p_in->Edge(v,w) && (ass.col[v] == ass.col[w])) 
	    {
	      ass.fitness++;
	    }
	}
    }
    return ass.fitness;
  }



fvalue ColManPenaltyFunction::computeFitness(Coloring &ass)
  {
    ass.fitness = 0;
    for (Color c = 0; c < ass.CurrColors(); c++)
    {
      ass.fitness += ass.partitions[c].size()*ass.partitions[c].size();
    }
    return ass.fitness;
  }


/*
void ColoringManager::updateDeltaMatrixOneEx(Coloring &ass, Vertex var, 
					     Color old_value, Color value)
{

  ass.conflicts[var] += ass.Delta(var,value);
  if (ass.conflicts[var] == 0)
    ass.conflicts_set.erase(var);
  else if (ass.Delta(var,value) > 0)
    ass.conflicts_set.insert(var);

  ass.Delta(var,old_value) = - ass.Delta(var,value);
  for (Color j=0;j< ass.CurrColors();j++)
    {
      if (j!=old_value && j!=value)
	 {
	   ass.Delta(var,j) -= ass.Delta(var,value);
	 }
    }
  ass.Delta(var,value) = 0;

  VertexSet::const_iterator pt;
  for (pt = p_in->adjacency_set[var].begin();
       pt != p_in->adjacency_set[var].end(); pt++)
    {
      if (ass.col[*pt] != old_value)
	ass.Delta(*pt,old_value)--;
      if (ass.col[*pt] != value)
	ass.Delta(*pt,value)++;
      
      if (ass.col[*pt] == value) 
	{
	  ass.conflicts[*pt]++;
	  ass.conflicts_set.insert(*pt);
	  for (Color j = 0; j < ass.CurrColors(); j++) 
	    {
	      if (j != value)
		ass.Delta(*pt,j)--;
	    }
	}
      
      if (ass.col[*pt] == old_value) 
	{
	  ass.conflicts[*pt]--;
	  if (ass.conflicts[*pt]==0)
	    ass.conflicts_set.erase(*pt);
	  for (Color j = 0; j < ass.CurrColors(); j++) 
	    {
	      if (j != old_value)
		ass.Delta(*pt,j)++;
		//delta_matrix[var][j]++;
	    }
	}
    }
}
*/

/*
void ColoringManager::updateImpGraphOneEx(Coloring &ass, Vertex var, 
					     Color old_col, Color new_col)
{
  fvalue nodes_adj_var_in_old_col = 0;
  fvalue nodes_adj_var_in_new_col = 0;
  fvalue nodes_adj_pt_in_old_col = 0;
  fvalue nodes_adj_pt_in_new_col = 0;

  VertexSet::const_iterator pt;
  VertexSet::const_iterator pt_inner;
 
  Vertex N = p_in->Vertices();
 
  //-------------------- 1 - adjacents nodes influences--------------------
  for(pt = p_in->adjacency_set[var].begin();
      pt != p_in->adjacency_set[var].end(); pt++)
    {
      //------- a) -------
      if (ass.col[*pt]==old_col)
	{
	  nodes_adj_var_in_old_col++;
	  for (Vertex i=0;i < N;i++)
	    {
	      if ((ass.col[i] != old_col) && (i != var) )
		ass.impGraph[i][*pt]++;
	    }
	  //------- b) ------------update j E Si -> k E Sj-----------------------
	  for (pt_inner = ass.partitions[new_col].begin();
	       pt_inner != ass.partitions[new_col].end(); pt_inner++)
	    {
	      if (*pt_inner != var)
		{
 		  ass.impGraph[*pt][*pt_inner]++;
		}
	    }
	}
      else
	{
	  //------- d) -------
	  for (pt_inner = ass.partitions[old_col].begin();
	       pt_inner != ass.partitions[old_col].end(); pt_inner++)
	    {
	      ass.impGraph[*pt][*pt_inner]--;
	    }
	  //------- e) -------
	  if (ass.col[*pt] != new_col)
	    {
	      for (pt_inner = ass.partitions[new_col].begin();
		   pt_inner != ass.partitions[new_col].end(); pt_inner++)
		{
		  if (*pt_inner != var)
		    {
		      ass.impGraph[*pt][*pt_inner]++;		   
		    }	    
		}
	    }
	}
      if (ass.col[*pt] == new_col)
	nodes_adj_var_in_new_col++;
    }

  for (pt = ass.partitions[new_col].begin();
       pt != ass.partitions[new_col].end(); pt++)
    {
      //------- c) ------------update edges i/Sj -> j E Sj / var-----------------
      if (*pt != var)
	{
	  if (p_in->Edge(var,*pt))
	  {
	    for (Vertex i = 0; i < N ;i++)
	      {
		if (ass.col[i] != new_col)
		  {
		    ass.impGraph[i][*pt]--;
		    }
		}
	    }
	}
      //-------------------- 2 - remove edges --------------------
      ass.impGraph[*pt][var] = MAX_FVALUE;
      ass.impGraph[var][*pt] = MAX_FVALUE;
    }
//-------------------- 3 - new edges added --------------------
   for (pt = ass.partitions[old_col].begin();
       pt != ass.partitions[old_col].end(); pt++)
    {
      nodes_adj_pt_in_old_col = 0;
      nodes_adj_pt_in_new_col = 0;
      //select cheeper loop 
      if (ass.loop_on_partition)
	{
	  for (pt_inner = ass.partitions[old_col].begin();
	       pt_inner != ass.partitions[old_col].end(); pt_inner++)
	    {
	      if (p_in->Edge(*pt_inner,*pt))
		nodes_adj_pt_in_old_col++;
	    }
	  for (pt_inner = ass.partitions[new_col].begin();
	       pt_inner != ass.partitions[new_col].end(); pt_inner++)
	    {
	      if (p_in->Edge(*pt_inner,*pt))
		{
		  if (*pt_inner != var)
		    nodes_adj_pt_in_new_col++;
		}
	    }
	}
      else
	{
	  for(pt_inner = p_in->adjacency_set[*pt].begin();
	      pt_inner != p_in->adjacency_set[*pt].end(); pt_inner++)
	    {
	      if (ass.col[*pt_inner] == old_col)
		nodes_adj_pt_in_old_col++;
	      else if (ass.col[*pt_inner] == new_col)
		{
		  if (*pt_inner != var)
		    nodes_adj_pt_in_new_col++;
		}
	    }
	}


      ass.impGraph[*pt][var] = nodes_adj_pt_in_new_col - nodes_adj_var_in_new_col;
      ass.impGraph[var][*pt] = - nodes_adj_pt_in_old_col + nodes_adj_var_in_old_col;
      if (p_in->Edge(*pt,var))
	ass.impGraph[var][*pt]--;
      
//        if (B[pt->node][var]!=0)
//  	{
//  	  Ass.ImpGraph->aCost[pt->node][var]--;
//  	  Ass.ImpGraph->aCost[var][pt->node]--;
//  	  }
    }
//--------------------- 4 - update edges i/Si->var---------------------------
   for (Vertex i=0; i < N; i++)
    {
      if ((ass.col[i] != new_col) && (ass.col[i] != old_col))
	{
	  nodes_adj_pt_in_new_col = 0;
	  for (pt = ass.partitions[new_col].begin();
	       pt != ass.partitions[new_col].end(); pt++)
	    {	      
	      if (p_in->Edge(*pt,i))
		{
		  if (*pt != var)
		    nodes_adj_pt_in_new_col++;
		}
	    }
	  ass.impGraph[i][var] =  nodes_adj_pt_in_new_col - nodes_adj_var_in_new_col;
	  
//  	  if (B[i][var]!=0)
//  	    {
//  	      Ass.ImpGraph->aCost[i][var]--;
	  //     }
	}
    }
}
*/

void ColoringManager::updateDeltaDataOneEx(Coloring &ass, Vertex var, 
					   Color old_col, Color new_col)
{
    // cout<<"-->"<<var<<"["<<old_col<<","<<new_col<<"]"<<endl;
  ass.conflicts[var] = ass.AdjColor(var,new_col);
  //+=ass.Delta(var,new_col);
  
  if (ass.conflicts[var] == 0)
    ass.conflicts_set.erase(var);
  else if (ass.conflicts[var])
    ass.conflicts_set.insert(var);

//    ass.Delta(var,old_col) = - ass.Delta(var,new_col);
  
//     for (Color j=0;j< ass.k;j++)
//        {
//          if (j!=old_col && j!=new_col)
//    	{
//    	  ass.Delta(var,j) -= ass.Delta(var,new_col);
//    	}
//        }
//     ass.Delta(var,new_col) = MAX_FVALUE;

  VertexSet::const_iterator pt;
  for (pt = p_in->adjacency_set[var].begin();
       pt != p_in->adjacency_set[var].end(); pt++)
    {
      ass.AdjColor((*pt),old_col)--;
      ass.AdjColor((*pt),new_col)++;
//	
//	ass.Delta((*pt),old_col)--;
//	ass.Delta((*pt),new_col)++;
//	
      if (ass.col[(*pt)]==new_col)
  	{
	  //	  ass.Delta((*pt),old_col)--;
  	  ass.conflicts[(*pt)]++;
  	  ass.conflicts_set.insert(*pt);
	  if (var < *pt)
	  {
	      ass.conflict_edges.insert(pair<Vertex,Vertex> (var,*pt));
	  }
	  else 
	  {
	      ass.conflict_edges.insert(pair<Vertex,Vertex> (*pt,var));
	  }
	  //ass.Delta((*pt),new_col)=MAX_FVALUE;
	  //ass.Delta((*pt),old_col)--;
	}
      if (ass.col[(*pt)]==old_col)
	{
	  ass.conflicts[(*pt)]--;
	  if (ass.conflicts[(*pt)]==0)
	    ass.conflicts_set.erase(*pt);

	  //set<pair<Vertex, Vertex > >::iterator exp;
	  if (var < *pt)
	      {
		  //	  cout<<var<<" "<<*pt<<" ";
		  //exp = ass.conflict_edges.find(pair<Vertex,Vertex> (var,*pt));
		  //if (exp==ass.conflict_edges.end())
		  //{
		  //   cout<<"no "<<endl; 
		  //}
		  //else
		  //  cout<<"["<<(*exp).first<<","<<(*exp).second<<"]"<<endl;
		  ass.conflict_edges.erase(pair<Vertex,Vertex> (var,*pt));
	      }
	  else 
	  {
	      //cout<<*pt<<" "<<var<<" ";
	      //exp = ass.conflict_edges.find(pair<Vertex,Vertex> (*pt, var));
	      // if (exp==ass.conflict_edges.end())
	      //	  {
	      //	      cout<<"no "<<endl; 
	      //	  }
	      //else
	      //   cout<<"["<<(*exp).first<<","<<(*exp).second<<"]"<<endl;
	      ass.conflict_edges.erase(pair<Vertex,Vertex> (*pt,var));
	  }

	  //ass.Delta((*pt),old_col)=MAX_FVALUE;
	  //ass.Delta((*pt),new_col)++;
	}
    }
  
 
 //    ass.Delta(var,old_col) = - ass.Delta(var,new_col);
  
//     for (Color j=0;j< ass.k;j++)
//        {
//          if (j!=old_col && j!=new_col)
//    	{
//    	  ass.Delta(var,j) -= ass.Delta(var,new_col);
//    	}
//        }
//     ass.Delta(var,new_col) = MAX_FVALUE;
//
//  //VertexSet::const_iterator pt;
//  for (pt = p_in->adjacency_set[var].begin();
//	 pt != p_in->adjacency_set[var].end(); pt++)
//    {
//	if (ass.col[*pt] != old_col)
//	  ass.Delta(*pt,old_col)--;
//	if (ass.col[*pt] != new_col)
//	  ass.Delta(*pt,new_col)++;
//    
//	if (ass.col[*pt] == new_col) 
//	  {
//	    //  ass.conflicts[*pt]++;
//	    //ass.conflicts_set.insert(*pt);
//	    for (Color j = 0; j < ass.k; j++) 
//	      {
//		if (j != new_col)
// 		  ass.Delta(*pt,j)--;
//	      }
//	  }
//    
//	if (ass.col[*pt] == old_col) 
//	  {
//	    //ass.conflicts[*pt]--;
//	    //if (ass.conflicts[*pt]==0)
//	    //  ass.conflicts_set.erase(*pt);
//	    for (Color j = 0; j < ass.k; j++) 
//	      {
//		if (j != old_col)
//		  ass.Delta(*pt,j)++;
//		/* delta_matrix[var][j]++; */
//	      }
//	  }
//    }
// 

//      fvalue nodes_adj_var_in_old_col = 0;
//      fvalue nodes_adj_var_in_new_col = 0;
//      fvalue nodes_adj_pt_in_old_col = 0;
//      fvalue nodes_adj_pt_in_new_col = 0;

//      //  VertexSet::const_iterator pt;
//      VertexSet::const_iterator pt_inner;

//      Vertex N = p_in->Vertices();

//      /*-------------------- 1 - adjacents nodes influences--------------------*/
//      for(pt = p_in->adjacency_set[var].begin();
//  	pt != p_in->adjacency_set[var].end(); pt++)
//        {
//  	/*------- a) -------*/
//  	if (ass.col[*pt]==old_col)
//  	{
//  	  nodes_adj_var_in_old_col++;
//  	  for (Vertex i=0;i < N;i++)
//  	    {
//  	      if ((ass.col[i] != old_col) && (i != var) )
//  		ass.impGraph[i][*pt]++;
//  	    }
//  	  /*------- b) ------------update j E Si -> k E Sj------------------------*/
//  	  for (pt_inner = ass.partitions[new_col].begin();
//  	       pt_inner != ass.partitions[new_col].end(); pt_inner++)
//  	    {
//  	      if (*pt_inner != var)
//  		{
//  		  ass.impGraph[*pt][*pt_inner]++;
//  		}
//  	    }
//  	}
//  	else
//  	{
//  	  /*------- d) -------*/
//  	  for (pt_inner = ass.partitions[old_col].begin();
//  	       pt_inner != ass.partitions[old_col].end(); pt_inner++)
//  	    {
//  	      ass.impGraph[*pt][*pt_inner]--;
//  	    }
//  	  /*------- e) -------*/
//  	  if (ass.col[*pt] != new_col)
//  	    {
//  	      for (pt_inner = ass.partitions[new_col].begin();
//  		   pt_inner != ass.partitions[new_col].end(); pt_inner++)
//  		{
//  		  if (*pt_inner != var)
//  		    {
//  		      ass.impGraph[*pt][*pt_inner]++;		   
//  		    }	    
//  		}
//  	    }
//  	}
//  	if (ass.col[*pt] == new_col)
//  	nodes_adj_var_in_new_col++;
//        }

//      for (pt = ass.partitions[new_col].begin();
//  	 pt != ass.partitions[new_col].end(); pt++)
//        {
//  	/*------- c) ------------update edges i/Sj -> j E Sj / var--------------------*/
//  	if (*pt != var)
//  	{
//  	  if (p_in->Edge(var,*pt))
//  	    {
//  	      for (Vertex i = 0; i < N ;i++)
//  		{
//  		  if (ass.col[i] != new_col)
//  		    {
//  		      ass.impGraph[i][*pt]--;
//  		    }
//  		}
//  	    }
//  	}
//  	/*-------------------- 2 - remove edges --------------------*/
//  	ass.impGraph[*pt][var] = MAX_FVALUE;
//  	ass.impGraph[var][*pt] = MAX_FVALUE;
//        }
//      /*-------------------- 3 - new edges added --------------------*/
//      for (pt = ass.partitions[old_col].begin();
//  	 pt != ass.partitions[old_col].end(); pt++)
//        {
//  	nodes_adj_pt_in_old_col = 0;
//  	nodes_adj_pt_in_new_col = 0;
//  	//select cheeper loop 
//  	if (ass.loop_on_partition)
//  	{
//  	  for (pt_inner = ass.partitions[old_col].begin();
//  	       pt_inner != ass.partitions[old_col].end(); pt_inner++)
//  	    {
//  	      if (p_in->Edge(*pt_inner,*pt))
//  		nodes_adj_pt_in_old_col++;
//  	    }
//  	  for (pt_inner = ass.partitions[new_col].begin();
//  	       pt_inner != ass.partitions[new_col].end(); pt_inner++)
//  	    {
//  	      if (p_in->Edge(*pt_inner,*pt))
//  		{
//  		  if (*pt_inner != var)
//  		    nodes_adj_pt_in_new_col++;
//  		}
//  	    }
//  	}
//  	else
//  	{
//  	  for(pt_inner = p_in->adjacency_set[*pt].begin();
//  	      pt_inner != p_in->adjacency_set[*pt].end(); pt_inner++)
//  	    {
//  	      if (ass.col[*pt_inner] == old_col)
//  		nodes_adj_pt_in_old_col++;
//  	      else if (ass.col[*pt_inner] == new_col)
//  		{
//  		  if (*pt_inner != var)
//  		    nodes_adj_pt_in_new_col++;
//  		}
//  	    }
//  	}


//  	ass.impGraph[*pt][var] = nodes_adj_pt_in_new_col - nodes_adj_var_in_new_col;
//  	ass.impGraph[var][*pt] = - nodes_adj_pt_in_old_col + nodes_adj_var_in_old_col;
//  	if (p_in->Edge(*pt,var))
//  	ass.impGraph[var][*pt]--;
    
//  	/*if (B[pt->node][var]!=0)
//  	{
//  	Ass.ImpGraph->aCost[pt->node][var]--;
//  	Ass.ImpGraph->aCost[var][pt->node]--;
//  	}*/
//        }
//      /*--------------------- 4 - update edges i/Si->var---------------------------*/
//      for (Vertex i=0; i < N; i++)
//        {
//  	if ((ass.col[i] != new_col) && (ass.col[i] != old_col))
//  	{
//  	  nodes_adj_pt_in_new_col = 0;
//  	  for (pt = ass.partitions[new_col].begin();
//  	       pt != ass.partitions[new_col].end(); pt++)
//  	    {	      
//  	      if (p_in->Edge(*pt,i))
//  		{
//  		  if (*pt != var)
//  		    nodes_adj_pt_in_new_col++;
//  		}
//  	    }
//  	  ass.impGraph[i][var] =  nodes_adj_pt_in_new_col - nodes_adj_var_in_new_col;
	  
//  	  /*if (B[i][var]!=0)
//  	    {
//  	    Ass.ImpGraph->aCost[i][var]--;
//  	    }*/
//  	}
//        }
}





void ColoringManager::decrease_colors(Coloring &ass)
{
  Color removed_col = (Color) rg->ran0N(ass.CurrColors()-1);

  for (Vertex i=0; i < p_in->Vertices(); i++)
    {
      if (ass.col[i]>removed_col)
	{
	  ass.col[i]=ass.col[i]-1;
	}
      else if (ass.col[i]==removed_col)
	{
	  ass.col[i]=rg->ran0N(ass.CurrColors()-2);
	}
    }
  ass.SetColors(ass.CurrColors()-1);
  updateColoringData(ass);
}






//######################################################################
//######################################################################
//                        ColManWithPenalties
//######################################################################
//######################################################################


ColManWithPenalties::ColManWithPenalties(Random *rnd, Graph* pin = NULL,Control *ctrl=NULL):
    ColoringManager(rnd,pin,ctrl) 
{
  penalties.resize(p_in->Vertices());
  for (unsigned int i = 0; i < p_in->Vertices(); i++)
    {
      penalties[i].resize(p_in->Vertices(),0);
    }
}




void ColManWithPenalties::reinit_data()
{
  for (unsigned int i = 0; i < p_in->Vertices(); i++)
    {
      for (unsigned int j = 0; j < p_in->Vertices(); j++)
	{
	  penalties[i][j] = 0;
	}
    }
}

void ColManWithPenalties::scale_penalties()
{
  for (unsigned int i = 0; i < p_in->Vertices(); i++)
    {
      for (unsigned int j = 0; j < p_in->Vertices(); j++)
	{
	  penalties[i][j] *= (float) 4/5;
	}
    }
}

 
void ColManWithPenalties::updateColoringData(Coloring &ass) const
{

  const Graph &g = *p_in; // g is an alias for the input Graph
  
  float lambda = p_ctrl->lambda;
  // first build the group list
  // and the edge list
 
  /*sceglie il ciclo piu conveniente su colore o nodi adiacenti*/  
  //approx_col = 2*N/ass.k;
  //approx_adj = N*(p_in->AvDensity());
 
  if ( (float)(g.Vertices()/ass.CurrColors()) < (float)(g.Vertices()*(p_in->Density())) )
    ass.loop_on_partition = true;
  else
    ass.loop_on_partition = false;
  
  ass.fitness = 0;
  ass.conflicts.resize(g.Vertices());
  ass.conflicts_set.clear();
  //ass.activation_bit.resize(g.Vertices())
  ass.partitions.resize(ass.CurrColors());
  //  ass.pen_delta.resize(g.Vertices());
  for (Color c = 0; c < ass.CurrColors(); c++)
    {
      ass.partitions[c].clear();
    }
  ass.pen_adj_x_color.clear();
  ass.pen_adj_x_color.resize(ass.CurrColors()*g.Vertices(),0);
  ass.adj_x_color.clear();
  ass.adj_x_color.resize(ass.CurrColors()*g.Vertices(),0);
  //cout << "here"<<endl;
  for (Vertex i=0; i<g.Vertices(); i++)
    {
      //activation_bit[i] = false;
      //  ass.pen_delta[i].resize(ass.CurrColors(),0);
      ass.conflicts[i]=0;
      ass.partitions[ass.col[i]].insert(i);
      VertexSet::const_iterator pt;
      for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++)
	{
	  ass.PenAdjColor(i,ass.col[*pt]) += (1 + (lambda * penalties[i][*pt]));
	  ass.AdjColor(i,ass.col[*pt])++;
	}
    }
  //------------------------------------------------------------
//    for (Vertex i=0; i<g.Vertices(); i++)
//     {
//       VertexSet::const_iterator pt;
//       for (Color c = 0; c < ass.CurrColors(); c++)
// 	{ 
// 	  ass.pen_delta[i][c]=0;
// 	  if (c!=ass.col[i])
// 	    for (pt = g.adjacency_set[i].begin(); pt != g.adjacency_set[i].end(); pt++)
// 	      {
// 		if (ass.col[*pt]==c)
// 		  ass.pen_delta[i][c] += (1 + (lambda * penalties[i][*pt]));  
// 		if (ass.col[*pt]==ass.col[i])
// 		  ass.pen_delta[i][c] -= (1 + (lambda * penalties[i][*pt]));  
		
// 	      } 

// 	}
//     }
  //------------------------------------------------------------
  //cout << "here2"<<endl;
  for (Vertex i=0; i<g.Vertices(); i++)
    {
     
      for (Vertex j=i; j<g.Vertices(); j++)
	{
	  if (g.Edge(i,j) && (ass.col[i] == ass.col[j])) 
	    {
	      ass.fitness++;
	      ass.conflicts_set.insert(i);
	      ass.conflicts[i]++;
	      ass.conflicts_set.insert(j);
	      ass.conflicts[j]++;
	    }
	} 
    }
  if (ass.fitness != 0)
    ass.feasible = false;
  else
    ass.feasible = true; 
}



void ColManWithPenalties::updateDeltaDataOneEx(Coloring &ass, Vertex var, 
					       Color old_col, Color new_col)
{
  float lambda = p_ctrl->lambda;
  ass.conflicts[var] = ass.AdjColor(var,new_col);
  if (ass.conflicts[var] == 0)
    ass.conflicts_set.erase(var);
  else if (ass.conflicts[var])
    ass.conflicts_set.insert(var);


 //------------------------------------------------------------
//   ass.pen_delta[var][old_col] = - ass.pen_delta[var][new_col];
//   for (Color c=0; c < ass.CurrColors(); c++)
//     {
//       if (c!=old_col && c!=new_col)
// 	{
// 	  ass.pen_delta[var][c]= ass.pen_delta[var][c]-ass.pen_delta[var][new_col];
// 	}
//     }
//   ass.pen_delta[var][new_col] = 0;
 //------------------------------------------------------------


  VertexSet::const_iterator pt;
  for (pt = p_in->adjacency_set[var].begin();
       pt != p_in->adjacency_set[var].end(); pt++)
    {
      ass.AdjColor((*pt),old_col)--;
      ass.AdjColor((*pt),new_col)++;
      ass.PenAdjColor((*pt),old_col) -= (1 + (lambda * penalties[var][*pt]));
      ass.PenAdjColor((*pt),new_col) += (1 + (lambda * penalties[var][*pt]));
      if (ass.col[*pt]==new_col)
  	{
  	  ass.conflicts[*pt]++;
  	  ass.conflicts_set.insert(*pt);
	  //------------------------------------------------------------
// 	  for (Color c=0; c < ass.CurrColors(); c++)
// 	    {
// 	      if (c != new_col)
// 		ass.pen_delta[*pt][c] -= (1 + (lambda * penalties[var][*pt]));
// 	    }
	  //------------------------------------------------------------
	}
      if (ass.col[(*pt)]==old_col)
	{
	  ass.conflicts[(*pt)]--;
	  if (ass.conflicts[(*pt)]==0)
	    ass.conflicts_set.erase(*pt);
	  //------------------------------------------------------------
// 	  for (Color c=0; c < ass.CurrColors(); c++)
// 	    {
// 	      if (c != old_col)
// 		ass.pen_delta[*pt][c] += (1 + (lambda * penalties[var][*pt]));
// 	    }
	  //------------------------------------------------------------
	}
      //------------------------------------------------------------
//       if (ass.col[(*pt)]!=old_col)
// 	ass.pen_delta[*pt][old_col] -= (1 + (lambda * penalties[var][*pt]));
//       if (ass.col[(*pt)]!=new_col)
// 	ass.pen_delta[*pt][new_col] += (1 + (lambda * penalties[var][*pt]));
      //------------------------------------------------------------
    }

 
}
 
 
void ColManWithPenalties::update_data(Coloring &col)
{
  float util;
  float max;
  vector< pair<Vertex, Vertex> > max_util_features;
  
  vector< pair<Vertex, Vertex> >::iterator p;
  max = -1;
  
  vector< pair<Vertex, Vertex> >::iterator pt;
  for (pt = p_in->edges.begin(); pt != p_in->edges.end(); pt++)
    {
      if (col.col[(*pt).first]==col.col[(*pt).second])
	{
	  util = ((float) 1)/(1+penalties[(*pt).first][(*pt).second]);
	  if (util == max) {
	    max_util_features.push_back(*pt);
	  }
	  else if (util > max)  {
	    max_util_features.clear();
	    max_util_features.push_back(*pt);
	    max = util;
	  }
	}
    }
  
  for (p = max_util_features.begin(); p != max_util_features.end(); p++)
    {
      
      penalties[(*p).second][(*p).first]++;
      penalties[(*p).first][(*p).second]++;

      if (col.col[(*p).first] == col.col[(*p).second])
	{
	  col.PenAdjColor((*p).first,col.col[(*p).first]) += p_ctrl->lambda;
	  col.PenAdjColor((*p).second,col.col[(*p).second]) += p_ctrl->lambda;
	}
      else
	{
	  col.PenAdjColor((*p).second,col.col[(*p).first]) += p_ctrl->lambda;
	  col.PenAdjColor((*p).first,col.col[(*p).second]) += p_ctrl->lambda; 
	}

      //------------------------------------------------------------
//       if (col.col[(*p).first] == col.col[(*p).second])
// 	{
// 	  for (Color c = 0; c < col.CurrColors(); c++)
// 	    {
// 	      if (c!=col.col[(*p).first])
// 		{
// 		  col.pen_delta[(*p).first][c] -= (p_ctrl->lambda); 
// 		  col.pen_delta[(*p).second][c] -= (p_ctrl->lambda); 
// 		}
// 	    }
// 	}
//       else
// 	{
// 	  col.pen_delta[(*p).second][col.col[(*p).first]] += p_ctrl->lambda;
// 	  col.pen_delta[(*p).first][col.col[(*p).second]] += p_ctrl->lambda; 
// 	}
    }
}

void ColManWithPenalties::print_data()
{
 ofstream pen_os("tmp/gls-pen_profile.dat");
      for (Vertex i = 0; i< p_in->Vertices(); i++)
	for (Vertex j = 0; j< p_in->Vertices(); j++)
	  pen_os<<penalties[i][j]<<" ";
      pen_os<<endl;
      pen_os.close();
}

void Coloring::print_pen_delta()
{
 ofstream pen_os("tmp/gls-pen_profile.dat");
      for (Vertex i = 0; i< p_in->Vertices(); i++)
	for (Color j = 0; j< k; j++)
	  pen_os<<PenDelta(i,j)<<" ";
      pen_os<<endl;
      pen_os.close();
}

