/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef SOLUTION_H
#define SOLUTION_H
#include <vector>
#include "Problem.h"
#include "Control.h"
#include "Random.h"

using namespace std;

class Coloring 
{
  friend std::ostream& operator<<(std::ostream&, const Coloring&);
 public:   
  Coloring(Random* rnd, Graph* pin);  
  Coloring() : p_in(0), rg(0) {};
  virtual ~Coloring();
  Coloring(const Coloring &);

  void SetInput(Graph* pin);
  void SetColors(Color num_col);
  const Color CurrColors() const {return k;};
  Coloring& operator=(const Coloring &a);

  //    Color& operator[](const Vertex& );
  //Color operator[](const Vertex& ) const;
  //fvalue& Delta(const Vertex&, const Color&);
  //const fvalue &Delta(const Vertex&, const Color&) const;
  //fvalue Delta(const Vertex&, const Color&) const;
  //fvalue ImpGraph(const Vertex& v, const Vertex& w) const;
  //
  //fvalue& AdjColor(const Vertex& v, const Color& c) ;
  //const fvalue &AdjColor(const Vertex& v, const Color& c) const;
    
     
   inline fvalue Delta(const Vertex& v, const Color& c) const 
    {
      if (c==col[v])
	return MAX_FVALUE;
      else
	{
	  return adj_x_color[v*k+c]-adj_x_color[v*k+col[v]];
	}
      //return delta[v*k+c];
    }

  inline fvalue ImpGraph(const Vertex& v, const Vertex& w) const 
    {
      //  return impGraph[v][w];
      // cerr<<v<<" "<<w<<endl;
      if (v==w)
	return MAX_FVALUE;
      else if (w>=p_in->Vertices() && v>=p_in->Vertices())
	{
	  return 0;
	}
      else if (w>=p_in->Vertices())//v<p_in->Vertices()
	{
	  return adj_x_color[v*k+w-p_in->Vertices()];
	}
      else if (v>=p_in->Vertices())//v>p_in->Vertices()
	{
	  return 0-adj_x_color[w*k+col[w]];
	}
      else if (col[v]==col[w])
	return MAX_FVALUE;
      else
	{
	  //cout<< "["<<adj_x_color[v*k+col[w]]<<" "<<adj_x_color[w*k+col[w]]<<" "<<p_in->Edge(v,w)<<"]"<<endl;      
	  return adj_x_color[v*k+col[w]]-adj_x_color[w*k+col[w]]-p_in->Edge(v,w);
	}
      //return delta[v*k+c];
    }

  inline Color Class(const Vertex v) const 
    {
      int l=v-p_in->Vertices();
      if (l >= 0)
	return l;
      else
	return col[v];
    }

  inline fvalue& AdjColor(const Vertex& v, const Color& c) 
    {
      return adj_x_color[v*k+c];
    }

  inline const fvalue &AdjColor(const Vertex& v, const Color& c) const 
    {
      return adj_x_color[v*k+c];
    }

  bool operator==(const Coloring &a);
 
  Vertex choose_conflict() const;
    

  bool InConflicts(const Vertex& v) const
    { return conflicts_set.find(v) != conflicts_set.end(); }


  inline float PenDelta(const Vertex& v, const Color& c) const 
    {
      if (c==col[v])
	return FLT_MAX;
      else
	{
	  return pen_adj_x_color[v*k+c]-pen_adj_x_color[v*k+col[v]];
	}
      //return delta[v*k+c];
    }

  inline float& PenAdjColor(const Vertex& v, const Color& c) 
    {
      return pen_adj_x_color[v*k+c];
    }
  
  inline const float &PenAdjColor(const Vertex& v, const Color& c) const 
    {
      return pen_adj_x_color[v*k+c];
    }
  
   vector<float> pen_adj_x_color;
     
  ColorMap col;  // the map between nodes of the graphs and colors
  fvalue fitness;

  //fvalue **impGraph;
  // for speeding up the computation
  vector<Vertex> conflicts; // the set of conflicting nodes
  VertexSet conflicts_set;    
  set<pair<Vertex, Vertex> > conflict_edges;

  vector<VertexSet> partitions;
  //vector<int> clashes;
 
  bool feasible;
  bool loop_on_partition;
  vector<fvalue> adj_x_color;
  void print_pen_delta();
  vector< vector<float> > pen_delta;
  // vector<bool> activation_bit;
   Color k; // the actual number of colors used 
protected:
 
  Graph *p_in; 
  Random* rg;
  //vector<fvalue> delta;	// this is a map V -> (C -> R)
    
};






class ColMove
{
 public:
  friend std::ostream& operator<<(std::ostream&, const ColMove&);
  ColMove();
  ColMove(const ColMove &a);
  ColMove& operator=(const ColMove &a);
  bool operator==(const ColMove &a);
  bool operator<(const ColMove &a) const {return seq[0]<a.seq[0];};
  vector<Vertex> seq;
  unsigned short int type; //1-one-ex (0-node 1-new-col, 2-cycle, 
  //3-path, 4-Kempe 5-pen_move 6-2-swap(pair-exchange)
  fvalue delta;
};

class TabuList;

class ColoringManager
{
 public:
  ColoringManager(Random *rnd, Graph* pin,Control *ctrl);
  virtual ~ColoringManager() {};
  //void UpdateColoringData(vector<Vertex>) const;
  virtual fvalue computeFitness(Coloring &ass);
  virtual void updateColoringData(Coloring&) const;    
  //    void updateColoringData1(Coloring&) const;    
  // void updateColoringData2(Coloring&) const;    
    
  void randomAssignment(Coloring&, Color num_col);
  void constructiveGreedyBounded(Coloring&, Color num_col);
  void constructiveGreedy(Coloring&);
  void constructiveRecursiveLargestFirst(Coloring&);
  void constructiveRecursiveLargestFirstReverse(Coloring&);
  void constructiveRecursiveLargestFirstStatic(Coloring&);
  void constructiveRecursiveLargestFirstBounded(Coloring&, Color num_col);
  void constructiveDSATUR(Coloring& ass);
  void constructiveDSATURBounded(Coloring& ass, Color num_col);

  void constructiveDSATURPartial(Coloring& ass, Color K);
  void constructiveDSATURPartialBounded(Coloring& ass, Color num_col);
  void readStartingSolution(Coloring& ass, Color num_col);
 
  void constructiveGreedyPartialBounded(Coloring&, Color);
  void constructiveGreedyPartial(Coloring&, Color);
    
  void decrease_colors(Coloring &ass);
  virtual void updateDeltaDataOneEx(Coloring &ass, Vertex var, Color old_value, 
				    Color value);
  Vertex distanceLap(Coloring &a, Coloring &b);
  //void updateDeltaMatrixOneEx(Coloring &ass, Vertex var, Color old_value, Color value);
  //void updateImpGraphOneEx(Coloring &ass, Vertex var, Color old_value, Color value);
  //void updateColoringDataOneEx (Coloring &ass, Vertex var, Color old_value, Color value);
  /*void updateDeltaMatrix1ex (Coloring&);
    
  void updateImpGraph1ex (Coloring&);
    
  void updateDeltaMatrixCycle (Coloring&,CYCLE cycle, LONGINT cycle_length, char ** B, 
  set **adj); 
  void updateImpGraphCycle (Coloring&, CYCLE cycle, LONGINT cycle_length, char ** B, 
  ImprovingGraph *ImpGraph, set **adj);      

  void updateDeltaMatrix ( LONGINT value, LONGINT old_value, LONGINT var, char ** B);  
  */
  Vertex perturbate_seq(Coloring &ass,Color to_remove, TabuList &tl, long int iter);
  Vertex perturbate_dsat(Coloring &ass,Color to_remove, TabuList &tl, long int iter);
  Vertex perturbate_best_moves(Coloring &ass,Color to_remove, 
			       TabuList &tl, long int iter);
  Vertex perturbate_adj(Coloring &ass,Color to_remove, 
			TabuList &tl, long int iter);

  virtual void update_data( Coloring &col) {};
  virtual void reinit_data() {};
    virtual void scale_penalties() {};
  virtual void print_data() {};
  void color(Coloring &ass, Vertex var, Color level, Vertex &changed);
  void uncolor(Coloring &ass, Vertex var, Color level);
  double p_smooth;
 protected:
  void SetInput(Graph* pin) { p_in = pin; }
  Graph *p_in;
  Random* rg;
  Control *p_ctrl;  
}; 



class ColManPenaltyFunction:public ColoringManager
{
 public:
  ColManPenaltyFunction(Random *rnd, Graph* pin,Control *ctrl):
    ColoringManager(rnd,pin,ctrl) {};
  ~ColManPenaltyFunction() {};
  void updateColoringData(Coloring&) const;    
  fvalue computeFitness(Coloring &ass);
}; 
 


class ColManWithPenalties:public ColoringManager
{
 public:
  ColManWithPenalties(Random *rnd, Graph* pin, Control *ctrl);
  ~ColManWithPenalties() {};
  //fvalue computeFitness(Coloring &ass);
  void updateColoringData(Coloring&) const;    
  void updateDeltaDataOneEx(Coloring &ass, Vertex var, Color old_col, 
			    Color new_col);
  void update_data( Coloring &col);
  void reinit_data();
  void print_data();
	void scale_penalties();
 protected:
  vector< vector<float> > penalties;  
}; 



struct lessMove
{
  bool operator() (const ColMove &a, const ColMove &b) const
  {
    return a<b;
  }
};

class TabuList
{
 public:
  TabuList(Vertex v, Color c, Random *rnd, Control *ctrl);
  TabuList(Control *ctrl);
  TabuList() {};
  virtual ~TabuList() {};
  virtual void Init(Vertex v, Color c, Random *rnd, Control *ctrl);
  virtual bool isTabu(const Vertex& v, const Color& c, unsigned long iter)
    {
      return tabu_list[v*colors+c]>=iter;}
  virtual unsigned long int &IterTabu(const Vertex& v, const Color& c) 
    {return tabu_list[v*colors+c];}
  virtual const unsigned long int IterTabu(const Vertex& v, const Color& c) const 
    { return tabu_list[v*colors+c];}
  virtual bool isTabu(const Vertex& v, unsigned long iter)
    { cerr<<"wrong access 1"<<endl; return 0;}
  virtual unsigned long int &IterTabu(const Vertex& v)
    { cerr<<"wrong access 2"<<endl; return tabu_list[0];}
  virtual const unsigned long int IterTabu(const Vertex& v) const
    { cerr<<"wrong access 3"<<endl; return 0;}
 
  virtual void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
  double prob;

  virtual bool areEqual(ColMove mv) {return false;};
  virtual const unsigned long int IterTabuMatrix(const Vertex& v, const Vertex& c) const 
    {cerr<<"wrong access 4"<<endl; return 0;};
  virtual void cleanTable() {};

  virtual   void UpdateTabuList(ColMove&, ColMove&) {};
  virtual  set<ColMove, lessMove> &List(const ColMove&) { cerr<<"&List"<<endl; 
  return tabu_matrix_list[0][0];};
  vector< vector < set<ColMove, lessMove> > > tabu_matrix_list;
 protected:
 unsigned long ComputeLength(const fvalue& conflicts);
  
  Color colors;
  Vertex vertices;
  vector<unsigned long int> tabu_list;
  Random *rg;
  Control *p_ctrl;
 
};



class TabuListVLSN:public TabuList
{
 public:
  TabuListVLSN(Control *ctrl):TabuList(ctrl) {};
  TabuListVLSN(Vertex v,  Color col,Random *rnd, Control *ctrl);
  TabuListVLSN() {};
  virtual ~TabuListVLSN() {};
  void Init(Vertex v, Color c, Random *rnd, Control *ctrl);
  bool isTabu(const Vertex& v1,  const Color& c,unsigned long iter);
  void cleanTable();
  void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
 private:
  Vertex vertices;
};


class TabuListVertices:public TabuList
{
 public:
  TabuListVertices(Vertex v, Random *rnd, Control *ctrl);
  TabuListVertices() {};
  virtual ~TabuListVertices() {};
  bool isTabu(const Vertex& v, unsigned long iter)
    { return tabu_list[v]>=iter;}
  unsigned long int &IterTabu(const Vertex& v)
    {return tabu_list[v];}
  const unsigned long int IterTabu(const Vertex& v) const
    {return tabu_list[v];}
  void cleanTable();
  void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
};




class TabuListPart:public TabuList
{
 public:
  TabuListPart(Vertex v, Color c, Random *rnd, Control *ctrl, Coloring *state);
  virtual ~TabuListPart() {};
  bool isTabu(const Vertex& v, const Color& c, unsigned long iter);

  // bool isTabu(const Vertex v, const Vertex w, unsigned long iter)
  //  {return tabu_list[v*vertices+w]>=iter;}

  unsigned long int &IterTabu(const Vertex& v, const Vertex& w) 
    {return tabu_list[v*vertices+w];}
  const unsigned long int IterTabu(const Vertex& v, const Vertex& w) const 
    { return tabu_list[v*vertices+w];}
  void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
 protected:
  Vertex vertices;
  Coloring *p_st;
};


class TabuListMinConf:public TabuList
{
 public:
  TabuListMinConf(Vertex v, Color c, Random *rnd, Control *ctrl):
    TabuList(v,c,rnd,ctrl) {};
  virtual ~TabuListMinConf() {};
  void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
};

class HistoryNovelty:public TabuList
{
 public:
  HistoryNovelty(Vertex v, Color c, Random *rnd, Control *ctrl):
    TabuList(v,c,rnd,ctrl) {};
  virtual ~HistoryNovelty() {};
  void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);

};


class TabuListDoubleRunner:public TabuList
{
 public:
  TabuListDoubleRunner(Vertex v, Random *rnd, Control *ctrl);
  virtual ~TabuListDoubleRunner() {};
  virtual bool isTabu(const Vertex& u, const Vertex& v, unsigned long iter)
    {
      return tabu_matrix[u][v]>=iter;
    }
  const unsigned long int IterTabuMatrix(const Vertex& u, const Vertex& v) const 
    { 
      return tabu_matrix[u][v];}
  virtual void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
  bool areEqual(ColMove mv)
    {
      return (tabu_matrix[mv.seq[0]][mv.seq[2]] == tabu_matrix[mv.seq[1]][mv.seq[3]]
	      && tabu_matrix[mv.seq[0]][mv.seq[2]]!=0);
    }
  virtual void cleanTable();
 protected:
  vector< vector <unsigned long int> > tabu_matrix;
};

class TabuListPerMoves:public TabuList
{
 public:
  TabuListPerMoves(Vertex v, Random *rnd, Control *ctrl);
  virtual ~TabuListPerMoves() {};
  void UpdateTabuList(ColMove& mv, ColMove& prec_mv);
  set<ColMove, lessMove> &List(const ColMove&); 
  void cleanTable();
  //vector< vector < set<ColMove, lessMove> > > tabu_matrix_list;
};

class TabuListNew: public TabuList
{
public:
  TabuListNew(Control* ctrl) : TabuList(ctrl) {};
  virtual ~TabuListNew() {};
  virtual void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
};

class TabuListNewOld: public TabuList
{
public:
  TabuListNewOld(Control* ctrl) : TabuList(ctrl) {};
  virtual ~TabuListNewOld()  {};
  virtual void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
};

class TabuListVertex: public TabuList
{
public:
  TabuListVertex(Control* ctrl) : TabuList(ctrl) {};
  virtual ~TabuListVertex()  {};
  virtual void UpdateTabuList(const Coloring& st, const ColMove& mv, unsigned long iter);
};

#endif
