
/* GCP version 6.0 Copyright (C) 2003 Marco Chiarandini  
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de

 This file is part of GCP.

 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include"Solver.h"



void ConstructiveHeuristics::Solve(Coloring &current)
{
  assert(p_ctrl->init == 1 || p_ctrl->init == 3 || p_ctrl->init == 5 
	 || p_ctrl->init == 9 || p_ctrl->init == 11); 
  p_runner->initialization_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
}



void RandomRestart::Solve(Coloring &current)
{
  unsigned int restart = 0;
  while (restart < p_ctrl->max_restart && p_ctrl->timeLeft())
    {
      p_runner->initialization_procedure(current);
      p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
      if (!current.feasible)
	//p_runner->decreasing_procedure(current);
	p_runner->Run_procedure(current);
      restart++;
    }
 
}

void Metaheuristic_Fixed::Solve(Coloring &current)
{
  //cout<<p_ctrl->getTime()<<endl;

  p_runner->initialization_procedure(current);

  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
//  cout<<p_runner->p_in->Vertices();

  if (p_ctrl->init==0)
    return;
  if (!current.feasible)
  {
    p_runner->Run_procedure(current);
  }
}


void Metaheuristic_Descend::Solve(Coloring &initial)
 {
  
  Coloring current(initial);
  //cout<<p_ctrl->getTime()<<endl;
  total_iterations = 0;
  p_runner->initialization_procedure(current);
  
  //p_runner->TryReduceGraph(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);  
//cout<<p_ctrl->getTime()<<endl;
  while (!MaxIterationExpired() && p_ctrl->timeLeft())
    {
      if (current.feasible)
	{
	  initial = current;
	  p_runner->decreasing_procedure(current);
	  //p_runner->TryReduceGraph(current);	  
	  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
	}
      p_runner->SetMaxIteration(p_ctrl->max_iter-total_iterations);
      p_runner->Run_procedure(current);
      total_iterations = total_iterations + p_runner->NumberOfIterations();
      p_ctrl->setPartialIter(total_iterations);
      if (current.fitness!=0)
	  break;
    }
  p_ctrl->setPartialIter(total_iterations);
}

//----------------------------------------------------------------------
  /**
     Sets the i-th runner managed by the solver to the passed parameter.

     @param r a pointer to a compatible runner
     @param i the runner position to set with r
  */
void ColoringMultiRunnerSolver::setRunner(ColoringRunner *r, unsigned int i) 
{ 
  assert(i < runners.size());
  runners[i] = r; 
}


/**
     Adds the given runner to the list of the managed runners.
     
     @param r a pointer to a compatible runner to add
  */
void ColoringMultiRunnerSolver::addRunner(ColoringRunner *r)
{ runners.push_back(r); }

/**
     Clears the vector of runners managed by the solver.
  */
void ColoringMultiRunnerSolver::ClearRunners()
{ runners.clear(); }


void ColoringMultiRunnerSolver::Solve(Coloring &initial)
{
 Coloring current(initial);
 //cout<<p_ctrl->getTime()<<endl;
  runners[0]->initialization_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
  //cout<<p_ctrl->getTime()<<endl;
   total_iterations = 0;

  fvalue past_fit;
  bool improved = true;
  while (improved && !MaxIterationExpired() && p_ctrl->timeLeft())
  {
      improved = false;
      unsigned int i =(i+1)%2;
      cout << "Runner "<<i<<endl;
      runners[i]->update(current);
      if (current.feasible)
	runners[i]->decreasing_procedure(current);
      past_fit = current.fitness;
      runners[i]->SetMaxIteration(p_ctrl->max_iter-total_iterations);
      if (i==1)
	{
	  runners[i]->SetMaxIdleIteration(100);
	  runners[i]->Run_procedure(current);
	}
      else 
	{
	  runners[i]->SetMaxIdleIteration(1000);
	  runners[i]->Run_procedure(current);
	}
      total_iterations = total_iterations + runners[i]->NumberOfIterations();
      p_ctrl->setPartialIter(total_iterations);
      // if (past_fit > current.fitness)
      cout<<current.fitness<<endl;
      improved = true;
  }
}

/*
void ColoringMultiRunnerSolver::Solve(Coloring &initial)
{
 Coloring current(initial);
 //cout<<p_ctrl->getTime()<<endl;
  runners[0]->initialization_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
  //cout<<p_ctrl->getTime()<<endl;
   total_iterations = 0;

  fvalue past_fit;
  bool improved = true;
  while (improved)
  {
      improved = false;
      //for (unsigned int i = 0; i < runners.size(); i++)
      {
	  unsigned int i =(i+1)%2;
	  cout << "Runner "<<i<<endl;
	  runners[0]->update(current);
	  if (current.feasible)
	      runners[0]->decreasing_procedure(current);
	  past_fit = current.fitness;
	  runners[0]->SetMaxIteration(1000);
	  runners[0]->Run_procedure(current);
	  if (current.fitness!=0)
	    runners[1]->Run_procedure(current);
	  total_iterations = total_iterations + runners[i]->NumberOfIterations();
	  p_ctrl->setPartialIter(total_iterations);
	  // if (past_fit > current.fitness)
	  cout<<current.fitness<<endl;
	    improved = true;
      }
      }
}
*/

//------------------------------------------------------------

//HEA decreasing
void Metaheuristic_HEA::Solve(Coloring &current)
{
  SIZE = 10;
  //  cout<<p_ctrl->getTime()<<endl;
  if ((p_ctrl->init % 2) == 0)
    p_ctrl->init -= 1;
  total_iterations = 0;
  p_runner->initialization_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
  p_ctrl->k = current.CurrColors()-1;
  if ((p_ctrl->init % 2) != 0)
    p_ctrl->init += 1;
  
  while (!MaxIterationExpired() && p_ctrl->timeLeft())
    {
      HEA(current);
      p_ctrl->setPartialIter(total_iterations);
      if (current.feasible)
	p_ctrl->k--;
    }
  p_ctrl->setPartialIter(total_iterations);
  
}


//HEA dsatur but first sol given by (rlf).
void Metaheuristic_HEA_rlf::Solve(Coloring &initial)
{
  Coloring current(initial);
  SIZE = 10;
  //  cout<<p_ctrl->getTime()<<endl;
  total_iterations = 0;
  p_ctrl->init = 3;
  p_runner->initialization_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
  
  p_ctrl->init = 6;
  
  while (!MaxIterationExpired() && p_ctrl->timeLeft())
    {
      unsigned int worst_fit_parent = 0;
      unsigned int count_cross = 0;
      unsigned int count_restart = 0;

      if (current.feasible)
	{
	  initial=current;
	  p_runner->decreasing_procedure(current);
	  p_ctrl->k = current.CurrColors();
	}
      
      Coloring *pop_solutions[SIZE];
      Coloring *p_son = NULL;
       
      p_runner->SetMaxIteration(p_ctrl->ls_iter);      
   	for (unsigned int i = 0; i < SIZE; i++)
	{
	  //cout<<" parents "<<i<<endl;
	  pop_solutions[i] = new Coloring(p_ctrl->rg, gr);
	  if (i==0)
	    {
	      *(pop_solutions[i])=current;
	    }
	  else 
	    {
	      p_runner->initialization_procedure(*pop_solutions[i]);
	    }
	  p_runner->Run_procedure(*pop_solutions[i]);
	  total_iterations = total_iterations + p_runner->NumberOfIterations();
	  p_ctrl->setPartialIter(total_iterations);
	
	    if (current.fitness > pop_solutions[i]->fitness)
	    {
	      current = *pop_solutions[i];
	      //cout<<current.feasible<<"---------------------------------------"<<endl;	
	      if (current.feasible)
		{
		  for (unsigned int j = 0; j <= i; j++)
		    {
		      delete pop_solutions[j];
		    }
		  //p_ctrl->setHEACrossovers(count_cross,count_restart);
		  break;
		  //	  p_runner->decreasing_procedure(current);
		  //p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
		}
	    }
	}
      //cout<<"choose parents "<<total_iterations<<endl;

       
      while (!current.feasible && !MaxIterationExpired() && p_ctrl->timeLeft())
	{  
	  //choose parents
	  int a,b;
	   
	  a = p_ctrl->rg->ran0N(SIZE-1);
	  do {
	    b = p_ctrl->rg->ran0N(SIZE-1);
	  } while (b==a);
	   
	  if (pop_solutions[a]->fitness >= pop_solutions[b]->fitness)
	    worst_fit_parent = a;
	  else
	    worst_fit_parent = b;

	  //crossover
	  //cout<<"crossover"<<endl;
	  p_son = crossOver(pop_solutions[a], pop_solutions[b]);
	  count_cross++;

	  //local search
	  //cout<<"local search "<<total_iterations<<endl;
	  p_runner->Run_procedure(*p_son);
	  total_iterations = total_iterations + p_runner->NumberOfIterations();
	  p_ctrl->setPartialIter(total_iterations);
	  if (current.fitness > p_son->fitness)
	    {
	      current = *p_son;
	      if (current.feasible)
		{
		  delete p_son; 
		  //deallocate memory
		  for (unsigned int i = 0; i < SIZE; i++)
		    {
		      delete pop_solutions[i];
		    }
		  break;
		}
	    }
      
	  //remove old parent
	  //cout<<"remove old parent "<<total_iterations<<endl;
	  delete pop_solutions[worst_fit_parent];
	  pop_solutions[worst_fit_parent] = p_son;
      
	  //cout<<"distance "<<averageDistanceSolutions(pop_solutions)<<endl;
	  if (averageDistanceSolutions(pop_solutions) < 20)
	    {
	     
	      count_restart++;
	      p_ctrl->k=current.CurrColors();
	      *(pop_solutions[0]) = current;
	      for (unsigned int i = 1; i < SIZE; i++)
		{
		  //pop_solutions[i] = new Coloring(p_ctrl->rg, gr);
		  p_runner->initialization_procedure(*pop_solutions[i]);
		  p_runner->SetMaxIteration(p_ctrl->ls_iter);      
		  p_runner->Run_procedure(*pop_solutions[i]);
		  total_iterations = total_iterations + p_runner->NumberOfIterations();
		  p_ctrl->setPartialIter(total_iterations);
		  if (current.fitness > pop_solutions[i]->fitness)
		    {
		      current = *pop_solutions[i];
		      //deallocate memory
		      if (current.feasible)
			{
			  //deallocate memory
			  for (unsigned int i = 0; i < SIZE; i++)
			    {
			      delete pop_solutions[i];
			    }
			  //p_ctrl->setHEACrossovers(count_cross,count_restart);
			  break;
			}
		    }
		}
	    }
	  
	}
      p_ctrl->setHEACrossovers(count_cross,count_restart);
      //p_ctrl->setPartialIter(total_iterations);
    }
  
}

//for Hea on a fixed color use this
void Metaheuristic_HEA_fixed::Solve(Coloring &current)
{
  SIZE = 10;
 
  total_iterations = 0;
  p_runner->initialization_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
  
  while (!MaxIterationExpired() && p_ctrl->timeLeft())
    {
      HEA(current);
      p_ctrl->setPartialIter(total_iterations);
      if (current.feasible)
	return;
    }
}


void Metaheuristic_HEA::HEA(Coloring &current)
{
 unsigned int worst_fit_parent = 0;
 unsigned int count_cross = 0;
 unsigned int count_restart = 0;

  Coloring *pop_solutions[SIZE];
  Coloring *p_son = NULL;
  //  int best_sol = 0;
//    int best_fit_sol = LONG_INT;

  current.fitness =  MAX_FVALUE;
  current.feasible =  false;

  p_runner->SetMaxIteration(p_ctrl->ls_iter);      
  for (unsigned int i = 0; i < SIZE; i++)
    {
      //cout<<" parents "<<i<<endl;
      pop_solutions[i] = new Coloring(p_ctrl->rg, gr);
      p_runner->initialization_procedure(*pop_solutions[i]);
      p_runner->Run_procedure(*pop_solutions[i]);
      total_iterations = total_iterations + p_runner->NumberOfIterations();
      p_ctrl->setPartialIter(total_iterations);
      if (current.fitness > pop_solutions[i]->fitness)
	{
	
	  current = *pop_solutions[i];
	  //cout<<current.feasible<<"---------------------------------------"<<endl;	
	  if (current.feasible)
	    {
	      for (unsigned int j = 0; j <= i; j++)
		{
		  delete pop_solutions[j];
		}
	      p_ctrl->setHEACrossovers(count_cross,count_restart);
	      return;
	      //	  p_runner->decreasing_procedure(current);
	      //p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
	    }
	}
    }
  //cout<<"choose parents "<<total_iterations<<endl;


  while (!MaxIterationExpired() && p_ctrl->timeLeft())
    {  
      //choose parents
      int a,b;
  
      a = p_ctrl->rg->ran0N(SIZE-1);
      do {
	b = p_ctrl->rg->ran0N(SIZE-1);
      } while (b==a);
  
      if (pop_solutions[a]->fitness >= pop_solutions[b]->fitness)
	worst_fit_parent = a;
      else
	worst_fit_parent = b;

      //crossover
      //cout<<"crossover"<<endl;
      p_son = crossOver(pop_solutions[a], pop_solutions[b]);
      count_cross++;

      //local search
      //cout<<"local search "<<total_iterations<<endl;
      p_runner->Run_procedure(*p_son);
      total_iterations = total_iterations + p_runner->NumberOfIterations();
      p_ctrl->setPartialIter(total_iterations);
      if (current.fitness > p_son->fitness)
	{
	  current = *p_son;
	  if (current.feasible)
	    {
	      delete p_son; 
	      break;
	    }
	}
      
      //remove old parent
      //cout<<"remove old parent "<<total_iterations<<endl;
      delete pop_solutions[worst_fit_parent];
      pop_solutions[worst_fit_parent] = p_son;
      
      //cout<<"distance "<<averageDistanceSolutions(pop_solutions)<<endl;
      if (averageDistanceSolutions(pop_solutions) < 20)
	{
	  count_restart++;
	  p_ctrl->k=current.CurrColors();
	  *(pop_solutions[0]) = current;
	  for (unsigned int i = 1; i < SIZE; i++)
	    {
	      //pop_solutions[i] = new Coloring(p_ctrl->rg, gr);
	      p_runner->initialization_procedure(*pop_solutions[i]);
	      p_runner->SetMaxIteration(p_ctrl->ls_iter);      
	      p_runner->Run_procedure(*pop_solutions[i]);
	      total_iterations = total_iterations + p_runner->NumberOfIterations();
	      p_ctrl->setPartialIter(total_iterations);
	      if (current.fitness > pop_solutions[i]->fitness)
		{
		  current = *pop_solutions[i];
		  //deallocate memory
		  if (current.feasible)
		    {
		      for (unsigned int i = 0; i < SIZE; i++)
			{
			  delete pop_solutions[i];
			}
		      p_ctrl->setHEACrossovers(count_cross,count_restart);
		      return;
		    }
		}
	    }
	}
	  
    }

  //deallocate memory
  for (unsigned int i = 0; i < SIZE; i++)
    {
      delete pop_solutions[i];
    }
  p_ctrl->setHEACrossovers(count_cross,count_restart);

}

Coloring *Metaheuristic_HEA::crossOver(Coloring *par_a, Coloring *par_b)
{
  int ran_ass =0;
  unsigned int parent, other_parent = 0;
  Color l;
  int max;
  int par;
  Color in_p2;

  Color k = par_a->CurrColors();

  vector<VertexSet> partitions[2];

  partitions[0] = par_a->partitions;
  partitions[1] = par_b->partitions;

  Coloring *pt_son = new Coloring(p_ctrl->rg,gr);
  pt_son->SetColors(k);
  
  for (l=0;l<k;l++)
    {
      parent = (l % 2);
      max = -1;
      par = -1;

      for (Color m=0;m<k;m++)
	{
	  if ((partitions[parent][m].size() != 0) 
	      && ((int) partitions[parent][m].size() > max))
	    {
	      par = m;
	      max = partitions[parent][m].size();
	    }
	}
      if (max > 0)
	{
	  VertexSet::iterator pt;
	  pt = partitions[parent][par].begin();
	  while (pt!=partitions[parent][par].end())
	    {
	      pt_son->col[*pt] = l;
	      /*	      printf("\np: %ld elim: %ld ",l,pt->node);*/
	      assert((*pt)>=0 && (*pt)<gr->Vertices());
	      
	      other_parent = (parent+1)%2;
	      if (other_parent == 0)
		in_p2 = par_a->col[*pt];	      
	      else
		in_p2 = par_b->col[*pt];	      

	      partitions[other_parent][in_p2].erase(*pt);
	      pt++;
	    }
	  partitions[parent][par].clear();
	}
    }
  
  VertexSet::iterator pt;
  for (l=0;l<k;l++)
    {
      for (pt = partitions[0][l].begin();
	   pt != partitions[0][l].end();
	   pt++)
	{
	  ran_ass++;
	  pt_son->col[*pt] = p_ctrl->rg->ran0N(k-1);
	}
      for (pt = partitions[1][l].begin();
	   pt != partitions[1][l].end();
	   pt++)
	{
	  ran_ass++;
	  pt_son->col[*pt] = p_ctrl->rg->ran0N(k-1);
	}
    }
  p_sm->updateColoringData(*pt_son);
  return pt_son;
}



double Metaheuristic_HEA::averageDistanceSolutions(Coloring *p_pop[])
{
  Vertex sum = 0, distance, min;
  double average = 0;
  unsigned int count;
  

  count = 0;  
  for (unsigned int out=0;out<SIZE;out++)
    {
      //mean = 0;
      min = INT_MAX;
      
      for (unsigned int in=out+1;in<SIZE;in++)
	{
	  distance = p_sm->distanceLap((*p_pop[out]), (*p_pop[in]));
	  
	  sum += distance;
	  if (distance < min)
	    min = distance;
	  count++;
	}
      //mean = ((REAL) mean)/count;
    }

  average = ((double) (2*sum)/(SIZE*(SIZE-1)));
  //cout<<count<<" "<<((SIZE*(SIZE-1)/2))<<" "<<average<<endl;
  return average;
}


//------------------------------------------------------------
//------------------------------------------------------------

void XRLFSolver::Solve(Coloring &current)
{
  //cout<<p_ctrl->getTime()<<endl;
  total_iterations = 0;
  p_runner->Run_procedure(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);
}


//------------------------------------------------------------
//------------------------------------------------------------

void PenaltyFunction::Solve(Coloring &current)
{
  //cout<<p_ctrl->getTime()<<endl;
  total_iterations = 0;
  p_runner->initialization_procedure(current);
  current.feasible = true;
  p_runner->p_ne->initAuxiliar(current.CurrColors());
  //  p_runner->TryReduceGraph(current);
  p_ctrl->setCurrentCost(current.fitness,current.CurrColors(),0);  
//cout<<p_ctrl->getTime()<<endl;
  p_runner->SetMaxIteration(p_ctrl->max_iter-total_iterations);

  current.partitions.resize(current.CurrColors()+1);
  p_runner->Run_procedure(current);
  total_iterations = total_iterations + p_runner->NumberOfIterations();
  p_ctrl->setPartialIter(total_iterations);
}
