/* GCP version 3.5 Copyright (C) 2003 Marco Chiarandini
   Technische Universitaet Darmstadt
   email: machud@intellektik.informatik.tu-darmstadt.de
 
 This file is part of GCP.
 
 GCP is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 Foobar is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with GCP; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <cmath>
#include <cfloat>
#include <string>
#include<fstream>

//#include <unistd.h>
#include <csignal>

#include <sys/time.h>
#include <sys/resource.h>


#include "Timer.h"
#include "Random.h"
#include "Utilities.h"
#include "Problem.h"
#include "Solution.h"
#include "Explorer.h"
#include "Runners.h"
#include "Solver.h"
#include "Settings.h"


#define METAOUTPUT

using namespace std;


void got_timeout(int s)
{
  cerr<<"Interrupted. Time Limit exceeded!\n";
  exit(EXIT_FAILURE);
}


long int *array;

int main( int argc, char** argv )
{

  //Create a control object from the command line arguments.
  Random random(0);

  Control control(&random, argc, argv );

  Timer halt;
  halt.setAlarm((int) (control.getNumberOfTries()*control.getTimeLimit()+1),got_timeout);

  control.resetTime();

  Graph graph;
  graph.Load(control.getInputStream(),control.print_statistics);

#ifdef METAOUTPUT
  //control.informationsReport();
#endif

  graph.detectDisconnectivity();

  //Coloring *coloring = new Coloring(&random, &graph);
  while (control.triesLeft())
  {
    ColoringManager *coloringManager = new ColoringManager(&random, &graph, &control);
    ColOneExExplorer coee(coloringManager, &random, &graph);

    ColVLSNImprovExplorer vlsn_impr(coloringManager, &random, 50, &graph);
    ColVLSNExplorer vlsn(coloringManager, &random, 50, &graph);
    KempeChains kc(coloringManager, &random, &graph);
    ColOneExExplorer *coeewp = NULL;
    ColOneExMinConfExplorer *coemce = NULL;
    ColOneExNoveltyExplorer *coene = NULL;

    ColoringRunner *coloringRunner = NULL;
    TabuList *tabuList = NULL;

    switch(control.invtype)
    {
        case 0:
        break;
        case INV_OLD:
        tabuList = new TabuList(&control);
        control.getOutputStream() << "InverseMove: OLD\n";
        break;
        case INV_NEW:
        tabuList = new TabuListNew(&control);
        control.getOutputStream() << "InverseMove: NEW\n";
        break;
        case INV_NEWOLD:
        tabuList = new TabuListNewOld(&control);
        control.getOutputStream() << "InverseMove: NEW-OLD\n";
        break;
        case INV_VERTEX:
        tabuList = new TabuListVertex(&control);
        control.getOutputStream() << "InverseMove: VERTEX\n";
        break;
        case VLSN:
        tabuList = new TabuListVLSN(&control);
        control.getOutputStream() << "VLSN InverseMove: OLD\n";
        break;
        default:
        cerr << "Error: option invtype " << control.invtype << " not supported" << endl;
        exit(-1);
    }

    switch (control.Neighborhood)
    {
        case 0:
        coloringRunner = new ColoringRunner(coloringManager, &coee, &random, &control, &graph);
        break;

        case 1:
        coloringRunner = new ColoringTabuSearch(coloringManager, &coee, &random,
                                                &control, tabuList, &graph);
        control.getOutputStream() << "Runner: RESTRICTED\n";
        break;
        //----------------------------------------------------------------------
        case 25:
        coloringRunner = new ColoringRunner(coloringManager,
                                            &vlsn_impr, &random, &control, &graph);
        break;
        case 26:
        coloringRunner = new ColoringRunnerUpHill(coloringManager,
                         &vlsn, &random, &control, &graph);
        break;
        case 27:
        coloringRunner = new ColoringTabuSearch(coloringManager,
                                                &vlsn, &random, &control, tabuList, &graph);
        break;
        case 45:
        delete coloringManager;
        coloringManager = new ColManPenaltyFunction(&random, &graph, &control);
        coloringRunner = new SimulatedAnnealing(coloringManager, &kc, &random,
                                                &control, &graph);
        break;
        //----------------------------------------------------------------------
        case 50:
        delete coloringManager;
        coloringManager = new ColManWithPenalties(&random, &graph, &control);
        coeewp = new ColOneExExplorer(coloringManager, &random, &graph);
        coloringRunner = new GuidedLocalSearch(coloringManager, coeewp, &random,
                                               &control, &graph);
        break;
        case 60:
        coloringRunner = new ILS(coloringManager, &coee, &random,
                                 &control, &graph);
        break;
        case 71:
        coemce = new ColOneExMinConfExplorer(coloringManager, &random, &graph);
        coloringRunner = new MinConflicts(coloringManager, coemce, &random,
                                          &control, tabuList, &graph);
        control.getOutputStream() << "Runner: MINCONFLICT\n";
        break;
        default:
        cerr<<"No runner + explorer selected\n";
        break;
    }

    if (coloringRunner==NULL)
      coloringRunner = new ColoringRunner(coloringManager,&random,&control,&graph);

    vector<ColoringRunner* > runners;

    ColoringSolver *coloringSolver = NULL;// = new ColoringSolver(&control, coloringRunner);
    switch (control.solver)
    {
        case 0:
        coloringRunner->SetMaxIdleIteration(LONG_MAX);//100 * graph.Vertices() * coloring.k);
        coloringRunner->SetMaxIteration(control.max_iter);//LONG_MAX);
        coloringSolver = new RandomRestart(&control, coloringRunner);
        break;
        case 1:
        if (control.max_iter == 0)
          control.max_iter =  10000 * graph.Vertices();

        coloringRunner->SetMaxIdleIteration(LONG_MAX);//100 * graph.Vertices() * coloring.k);
        coloringRunner->SetMaxIteration(control.max_iter);//LONG_MAX);
        coloringSolver = new Metaheuristic_Fixed(&control, coloringRunner);
        break;
        case 2:
        //control.max_iter = 10000 * graph.Vertices();
        if (control.max_iter == 0)
          control.max_iter =  10000 * graph.Vertices();

        coloringRunner->SetMaxIdleIteration(LONG_MAX);//100 * graph.Vertices() * coloring.k);
        coloringSolver = new Metaheuristic_Descend(&control, coloringRunner);
        break;
        case 4: //HEA
        //control.max_iter = 10000 * graph.Vertices();
        coloringRunner->SetMaxIdleIteration(LONG_MAX);//100 * graph.Vertices() * coloring.k);

        coloringSolver = new Metaheuristic_HEA(&control,
                                               coloringRunner,
                                               &graph, coloringManager);
        break;

        case 6: //PenaltyFunction
        if (control.max_iter == 0)
          control.max_iter =  10000 * graph.Vertices();
        coloringRunner->SetMaxIdleIteration(LONG_MAX);//100 * graph.Vertices() * coloring.k);
        coloringSolver = new PenaltyFunction(&control,
                                             coloringRunner);
        break;
        case 12: //HEA with first solution determined by RLF and the others by dsatur
        //control.max_iter = 10000 * graph.Vertices();
        if (control.max_iter == 0)
          control.max_iter =  10000 * graph.Vertices();
        coloringRunner->SetMaxIdleIteration(LONG_MAX);//100 * graph.Vertices() * coloring.k);

        coloringSolver = new Metaheuristic_HEA_rlf(&control,
                         coloringRunner,
                         &graph, coloringManager);
        break;
        case 100:
        coloringSolver = new ConstructiveHeuristics(&control, coloringRunner);
        break;
        default:
        cerr<<"Wrong solver setting\n";
        exit(1);
    }


    //------------------------------------------------------------
    control.beginTry();
    Coloring *coloring = new Coloring(&random, &graph);
    coloringSolver->Solve(*coloring);
    control.endTry();
    //------------------------------------------------------------

    if (control.verify)
    {
      ofstream ooo(control.os_verify.c_str());
      //ooo<<control.real_size_solved<<endl;
      ooo<<(*coloring);//ooo<<original_coloring;
      ooo.close();
    }

    delete coloring;
    delete coloringSolver;
    delete coloringRunner;
    delete coene;
    delete coemce;
    delete coeewp;
    delete coloringManager;

  }
  return 1;

}

