#include <iostream>
#include <sstream>
#include <stdexcept>
#include "Control.h"

using namespace std;
using std::ostringstream; using std::string;



%%

i file "f inputfile (DIMACS format necessary)"	reqarg
o out	"f outputfile "				reqarg
n num	" number of tests"				reqarg
t time	" maximum time"				reqarg
m max	" number of iterations for each trial (optarg)"	optarg
s seed	" seed"					reqarg

  init " way of initializing a solution: "		reqarg
  neigh  " the neighborhood"				reqarg
  solver " Solver: "                                  reqarg
  test " run test version (1) or not (0) (optarg)"		optarg
  stat " prints instance statistics and reduction (optarg)"	optarg
  red  " apply reduction to the graph (optarg)"		optarg
  verify  " write the solution in tmp-sol.col"	reqarg
f filestart	" name of the file where a starting solution is recorded."	reqarg
k initk	" initial coloring"						reqarg
K finalK   " final coloring"							reqarg	
  lowerbound " best lower bound known. It improves the preprocessing"	reqarg

  decr " kind of decreasing:"						reqarg
  
 tltype " type of tabu length for tabu search "						reqarg
 invtype " type of inverse move for tabu search "						reqarg
  alpha " alpha parameter for adaptive tabu length"						reqarg
  tl " fixed tabu lenght "						reqarg
  tl1 " random min tabu lenght "						reqarg
  tl2 " random max tabu lenght "						reqarg
 
  restart " # of restart for Sovler 0"                                  reqarg
 
  lsiter " used if solver [4,5] to limit its iterations"		reqarg

  lambda " factor to weight the influence of penalties in GLS"		reqarg	
  sidewalks " number of sidewalks moves in guided local search "	reqarg	
  
  gamma " factor to determine the number of colors to remove in ILS"		reqarg
  pert  " which perturbation to use: 0 sequential, 1 dsatur, 2 rebuild"	reqarg
  ilsprob " acceptance criterion in ILS (0=Random Walk 1=AlwaysBest)"	reqarg
  prob " probility p in Novelty+"						reqarg

  setlim " size of graph to do exhaustive search for indep sets in XRLF "	reqarg
  candnum " number of candiates to sample in XRLF "				reqarg
  trialnum " number of RLF trials in XRLF "					reqarg
  exactlim " size of graph to be solved exactly "				reqarg

  temperature                    reqarg
  cutoff                         reqarg
  minpercent                     reqarg
  freezelim                      reqarg
  tempfactor                     reqarg  
  sizefactor                     reqarg

h help  "prints help information"					return
  morehelp  "prints some more help information"					return
  sahelp  "prints some more on sa help information"				return
  
%%

void Control::ComParser(int argc, char** argv ) 
  {
    struct options OPT;
    
    const char *progname;

  progname = argv [0] != 0 && *(argv [0]) != '\0'
    ? argv [0]
    : "gcp";

  try
    {
      int i = parse_options (&OPT, argc, argv);

      if (OPT.opt_help)
	{
	  //Usage();
	   std::cout << "Usage: " << progname
                             << " [OPTION]... [ARGUMENT]...\n\n"
                             << "Options:\n" STR_HELP;
	  exit(1);
	}

      if(OPT.opt_sahelp)
	{	
	  SAUsage();
	  exit(1);
	}
	  
       if(OPT.opt_morehelp)
	{	
	  Usage();
	  exit(1);
	}

      if(i <= 1)
	{
	  //Usage();
	   std::cout << "Usage: " << progname
                             << " [OPTION]... [ARGUMENT]...\n\n"
                             << "Options:\n" STR_HELP;
	   exit(1);
	}

      if (OPT.opt_file)
	{
	  is = OPT.arg_file;
	}
      else
	{
	  cerr<<"Instance file missing!\n"<<endl;
	  exit(1);
	}

      if(OPT.opt_out)
	{
	  os = new ofstream(OPT.arg_out);
	} else {
	  cerr << "Warning: No output file given, writing to stdout" << endl;
	  os = &cout;
	} 
	
      if(OPT.opt_filestart)
	{
	ss = (OPT.arg_filestart);
	}
      else {
	cerr << "No starting solution file given" << endl;
      }
      
      if(OPT.opt_num)
	{
	  maxTry = atoi(OPT.arg_num);
	} else {
	  cerr << "Warning: Number of tries is set to default (1)" << endl;
	}

      if(OPT.opt_time)
	{
	  timeLimit = atof(OPT.arg_time);
	} else {
	  cerr << "Warning: Time limit is set to default DBL_MAX" << endl;
	}

      if(OPT.opt_max)
	{
	  if (OPT.arg_max != 0)
	    max_iter = atoi(OPT.arg_max);
	  else
	    {
	      max_iter = 0;
	      cerr << "Warning: Max iter is set to 10000 |V|"<<endl;
	    }
	} else {
	  cerr << "Warning: Max iter is set to very large number (LONG_MAX)"<<endl;
	}

       if(OPT.opt_seed)
	{
	  seed = atoi(OPT.arg_seed);
	  rg->set_seed( seed );
	} else {
	  cerr << "Warning: " << seed << " used as default random seed" << endl;
	}
	
      if(OPT.opt_init) {
	init = atoi(OPT.arg_init);
      } else {
	cerr << "Warning: init not set" << endl;
      }

      if(OPT.opt_neigh)
	{
	  Neighborhood = atoi(OPT.arg_neigh);
	} else {
	  cerr << "Warning: neigh not set" << endl;
	}

      if(OPT.opt_solver)
	{
	  solver = atoi(OPT.arg_solver);
	} else {
	  cerr << "Warning: solver not set" << endl;
	}

      if (OPT.opt_test)
	{
	  test_version = 1;
	  cout<<"Running Test Version"<<endl;
	} 

      if(OPT.opt_stat)
	{
	  print_statistics = true;
	}

      if(OPT.opt_red)
	{
	  reduction = true;
	} else
	  reduction = false;
      

      if(OPT.opt_verify)
	{
	  os_verify = OPT.arg_verify;
	  verify = true;
	}


      if(OPT.opt_initk) 
	{
	  k = atoi(OPT.arg_initk);
	} else {
	  cerr << "Warning: No initial color set" << endl;
	}

      if (OPT.opt_finalk)
	{  
	  finalK = atoi(OPT.arg_finalk);
	} else {
	  cerr << "Warning: No initial color set" << endl;
	}
	 

      if(OPT.opt_lowerbound)  {
	lower_bound = atoi(OPT.arg_lowerbound);
      } else {
	cerr << "Warning: No lower bound set" << endl;
      }

      if(OPT.opt_decr)
	{
	  decr = atoi(OPT.arg_decr);
	  //cout<<proc<<endl;
	} else {
	  cerr << "Warning: decr not set" << endl;
	}

      if(OPT.opt_alpha)
	{
	  alpha = atof(OPT.arg_alpha);
	} else {
	}
	
	  if( OPT.opt_tltype)
	{
	  tltype =  atoi( OPT.arg_tltype); 
	} else {
	  cerr << "Warning: tltype parameter not set" << endl;
	}
	
	  if( OPT.opt_tl)
	{
	  tl =  atoi( OPT.arg_tl); 
	} else {
	  cerr << "Warning: tl parameter not set" << endl;
	}
	
	  if(OPT.opt_tl1)
	{
	  tl1 = atoi(OPT.arg_tl1);
	} else {
	}
	
	  if(OPT.opt_tl2)
	{
	  tl2 = atoi(OPT.arg_tl2);
	} else {
	}
	
	  if(OPT.opt_invtype)
	{
	  invtype = atoi(OPT.arg_invtype);
	} else {
	}

      if(OPT.opt_restart)
	{
	  max_restart = atoi(OPT.arg_restart);
	} else {
	}

      if(OPT.opt_lsiter)
	{
	  ls_iter = atoi(OPT.arg_lsiter);
	} else {
	  cerr << "Warning: lsiter not set" << endl;
	}

      if(OPT.opt_lambda)
	{
	  lambda = atof(OPT.arg_lambda);
	} else {
	  cerr << "Warning: lambda parameter not set" << endl;
	}

       if(OPT.opt_sidewalks)
	{
	  sidewalks = atoi(OPT.arg_sidewalks);
	} else {
	  cerr << "Warning: lambda parameter not set" << endl;
	}

      if(OPT.opt_gamma)
	{
	  gamma =  atof(OPT.arg_gamma);
	} else {
	  cerr << "Warning: gamma parameter not set" << endl;
	}

      if(OPT.opt_pert)
	{
	  pert =  atoi(OPT.arg_pert);
	} else {
	  cerr << "Warning: pert parameter not set" << endl;
	}



      if(OPT.opt_ilsprob)
	{
	  ilsprob =  atof(OPT.arg_ilsprob);
	} else {
	  cerr << "Warning: prob parameter not set" << endl;
	}

    
	
      if(OPT.opt_prob)
	{
	  prob = atof(OPT.arg_prob);
	} else {
	  cerr << "Warning: prob parameter not set" << endl;
	}

       if(OPT.opt_candnum)
	{
	  cand_num = atoi(OPT.arg_candnum);
	} else {
	  cerr << "Warning: candnum parameter not set; cand_num = 50" << endl;
	}

       if(OPT.opt_setlim)
	{
	  set_lim = atoi(OPT.arg_setlim);
	} else {
	  cerr << "Warning: setlim parameter not set; set_lim = 63" << endl;
	}

      if(OPT.opt_trialnum)
	{
	  trial_num =  atoi(OPT.arg_trialnum);
	} else {
	  cerr << "Warning: trial_num parameter not set; trial_num = 160" << endl;
	}

      if(OPT.opt_exactlim)
	{
	  exact_lim = atoi(OPT.arg_exactlim);
	} else {
	  cerr << "Warning: exact_lim parameter not set; exact_lim = 70" << endl;
	}





      if(OPT.opt_temperature)
	{
	  temperature = atoi(OPT.arg_temperature);
	} else {
	  cerr << "Warning: sa parameter not set" << endl;
	}
	
      if(OPT.opt_cutoff)
	{
	  cutoff = atof(OPT.arg_cutoff);
	} else {
	  cerr << "Warning: sa parameter not set" << endl;
	}
	
      if(OPT.opt_minpercent)
	{
	  minpercent = atof(OPT.arg_minpercent);
	    } else {
	      cerr << "Warning: sa parameter not set" << endl;
	    }

      if(OPT.opt_freezelim)
	{
	  freeze_lim = atoi(OPT.arg_freezelim);
	  //cout<<proc<<endl;
	} else {
	  cerr << "Warning: sa parameter not set" << endl;
	}

      if(OPT.opt_tempfactor)
	{
	  tempfactor = atof(OPT.arg_tempfactor);
	  //cout<<proc<<endl;
	} else {
	  cerr << "Warning: sa parameter not set" << endl;
	}
	
      if(OPT.opt_sizefactor)
	{
	  sizefactor = atof(OPT.arg_sizefactor);
	} else {
	  cerr << "Warning: sa parameter not set" << endl;
	}
	
  }
  catch (std::runtime_error e)
    {
      std::cerr << progname << ": " << e.what () << "\n"
	"Try `" << progname
		<< " --help' for more information.\n";
      exit(1);
    }
  }
